{  getparam -- pascal unit for parsing command line
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{
parameter files:

#       comment (until end of line)
"..."   everything between is a string
""      character "
}


{$mode objfpc}
unit getparam;

{$define showhelponerr}         { shows help message on error (type 1,2)}
{ define strengthvareval}       { strict type evaluation }


interface
const maxparam=64;

const tpstr       = 0;
      tpsmallint  = 1;
      tpword      = 2;
      tpbyte      = 3;
      tpshortint  = 4;
      tplint      = 5;
      tplongint   = 5;
      tpdword     = 6;
      tpboolean   = 7;
      tpastr      = 8;
      tpreal      = 9;
      tpsingle    = 10;
      tpdouble    = 11;
      tpextended  = 12;
      tpaastr     = 13;

      ranothing   = 0;
      ravdreq     = 1;  { variable must be defined }
      rapdreq     = 2;  { option/parameter must be defined }

      aasize      = 128;

type taastr=record            { ansistring array }
        aa    : array of ansistring;
        size  : longint;
     end;
     paastr = ^taastr;

     pparamdef=^tparamdef;
     tparamdef=record
        pt       : pointer;     { pointer to variable }
        default  : pointer;     { pointer to default value }
        tp       : byte;        { type of option }
        str      : string[15];  { option string }
        pd       : boolean;     { option defined }
        vd       : boolean;     { variable defined }
        des      : shortstring; { option description }
        ra       : byte;        { index of raprc in ratable }
        fmt      : string[35];  { format string, e.g. <nn.n> }
     end;

     traproc=procedure(var p:tparamdef);

     tparams=object
        ignorecase : boolean;   { ignore case }
        overwrite  : boolean;   { allow multi define }
        helpf      : boolean;   { show help }
        params     : array [1..maxparam] of tparamdef;   { parameter }
        opts       : array [1..maxparam] of tparamdef;   { options }
        pc,oc      : longint;
        tpsizes    : array[0..255] of longint;
        raprocs    : array[0..255] of traproc;
        paramusage : ansistring;
        constructor init;
        destructor done;
        procedure run;
        function defparam(const str,des:shortstring; pi,pj:pointer;tp,ra:byte):longint;
        function defparam(const str,des:shortstring; pi:pointer;tp,ra:byte):longint;
        function defparam(const str,des:shortstring; var d:double):longint;
        function defparam(const str,des:shortstring; var s:single):longint;
        function defparam(const str,des:shortstring; var b:boolean):longint;
        function defparam(const str,des:shortstring; var d:longint):longint;
        function defparam(const str,des:shortstring; var b:byte):longint;
        function defparam(const str,des:shortstring; var w:word):longint;
        function defparam(const str,des:shortstring; var s:shortstring):longint;
        function defparam(const str,des:shortstring; var s:ansistring):longint;
        function defparam(const str,des:shortstring; var a:taastr):longint;
        procedure deffmt(str:shortstring; const fmt:shortstring);
        procedure param2ptp(var p:tparamdef);
        function paramfmt(tp:longint):shortstring;
        function getnextparam:ansistring;
        function getnextparam2:ansistring;
        function lastparam:boolean;
        procedure showhelp;
        function optid(s:shortstring):longint;
        function vardef(p:pointer):boolean;
        procedure paramerr(const s:ansistring);
        function paramdef(const s:shortstring):boolean;
        procedure copytp(q:pointer;var p:tparamdef);
        procedure copytp(sa:ansistring;var p:tparamdef);
        procedure setraproc(i:byte;p:traproc);
        function condupstr(const s:shortstring):shortstring;
     private
        pn           : longint; { number of parameters in command line }
        pc2,oc2      : longint; { second parameter / option count }
        op           : boolean; { true if option }
        pf           : boolean; { read from parameter file }
        fp           : ansistring;  { parameter from file }
        tf           : text;    { parameter file }
        procedure getfp;
        procedure paramerr(b:byte;const s:ansistring);
     end;

var paramstr2   : function(i:longint):ansistring;
    paramcount2 : longint;

const longdescription : ansistring = '';

implementation

const paramerrnum=234;
{********************************************************************}
{******** raprocs ***************************************************}
{********************************************************************}
procedure rapnothing(var p:tparamdef);
begin
end;

procedure rapvdreq(var p:tparamdef);
begin
    if not p.vd then
    begin
        writeln(stderr,'insufficient parameters');
        halt(paramerrnum);
    end;
end;

procedure rappdreq(var p:tparamdef);
begin
    if not p.pd then
    begin
        writeln(stderr,p.str+': not defined');
        halt(paramerrnum);
    end;
end;

{********************************************************************}
{******** tparams ***************************************************}
{********************************************************************}
{******** init ******************************************************}
constructor tparams.init;
var i:longint;
begin
    ignorecase:=true;
    overwrite:=true;
    helpf:=false;
    pc:=0;
    pn:=0;
    oc:=0;
    pf:=false;
    for i:=1 to maxparam do
    begin
        params[i].pt:=nil;
        opts[i].pt:=nil;
    end;
    for i:=0 to 255 do
    begin
        tpsizes[i]:=0;
        raprocs[i]:=@rapnothing;
    end;
    tpsizes[tpstr]:=sizeof(shortstring);
    tpsizes[tpsmallint]:=sizeof(longint);
    tpsizes[tpword]:=sizeof(word);
    tpsizes[tpbyte]:=sizeof(byte);
    tpsizes[tpshortint]:=sizeof(shortint);
    tpsizes[tplint]:=sizeof(longint);
    tpsizes[tpdword]:=sizeof(dword);
    tpsizes[tpboolean]:=sizeof(boolean);
    tpsizes[tpreal]:=sizeof(real);
    tpsizes[tpsingle]:=sizeof(single);
    tpsizes[tpdouble]:=sizeof(double);
    tpsizes[tpextended]:=sizeof(extended);
    tpsizes[tpastr]:=sizeof(ansistring);
    tpsizes[tpaastr]:=sizeof(taastr);

    setraproc(ranothing,@rapnothing);
    setraproc(ravdreq,@rapvdreq);
    setraproc(rapdreq,@rappdreq);

    paramusage:='';
end;

{******** done ******************************************************}
destructor tparams.done;
begin
end;

{******** getnextparam **********************************************}
function tparams.getnextparam:ansistring;
begin
    getnextparam:=getnextparam2;
    if pf then getfp
    else inc(pn);
end;

{******** getnextparam2 *********************************************}
function tparams.getnextparam2:ansistring;
var s  : ansistring;
begin
    getnextparam2:='';
    if pf then s:=fp
    else begin
        if pn+1>paramcount2 then paramerr(2,'illegal parameter line (something is expected)');
        s:=paramstr2(pn+1);
        if s[1]='@' then
        begin
            inc(pn);
            pf:=true;
            delete(s,1,1);
{$I-}
            assign(tf,s);
            reset(tf);
{$I+}
            if ioresult<>0 then paramerr(2,'cant find parameter file: '+s);
            getfp;
            s:=fp;
        end;
    end;
    getnextparam2:=s;
end;

{******** getfp *****************************************************}
procedure tparams.getfp;
var ch  : char;
    s   : shortstring;
begin
    fp:='';
    if eof(tf) then
    begin
        close(tf);
        if pn<paramcount2 then fp:=paramstr2(pn+1);
        pf:=false;
    end
    else begin
        repeat
            read(tf,ch);
        until (ch>#32) or eof(tf);
        while (ch='#') and not eof(tf) do
        begin
            readln(tf,s);
            if not eof(tf) then
            begin
                repeat
                    read(tf,ch);
                until (ch>#32) or eof(tf);
            end;
        end;
        if eof(tf) then
        begin
            case ch of
                '"'           : paramerr(2,'error in parameter file '+paramstr2(pn)+': missing "');
                #33,#36..#255 : fp:=ch;
                else begin
                    close(tf);
                    if pn<paramcount2 then fp:=paramstr2(pn+1);
                    pf:=false;
                end;
            end;
        end
        else begin
            s:='';
            repeat
                case ch of
                '"'  :
                    begin
                        read(tf,ch);
                        if ch='"' then
                        begin
                            s:=s+ch;
                            if not eof(tf) then read(tf,ch)
                            else ch:=#0;
                        end
                        else begin
                            repeat
                                s:=s+ch;
                                if not eof(tf) then read(tf,ch);
                                while (ch<>'"') do
                                begin
                                    s:=s+ch;
                                    if eof(tf) then paramerr(2,'error in parameter file '+paramstr2(pn)+': missing "');
                                    read(tf,ch);
                                end;
                                if eof(tf) then ch:=#0
                                else read(tf,ch);
                                until ch<>'"';
                            end;
                    end;
                else
                    begin
                        s:=s+ch;
                        read(tf,ch);
                    end;
                end;
                if eof(tf) then
                    case ch of
                    '"' :
                        paramerr(2,'error in parameter file '+paramstr2(pn)+': missing "');
                    #33,#36..#255 :
                        s:=s+ch;
                    end;
            until (ch<=#32) or (eof(tf)) or (ch='#');
            fp:=s;
            if not eof(tf) and (ch='#') then readln(tf,s);
        end;
    end;
end;

{******** lastparam *************************************************}
function tparams.lastparam:boolean;
begin
    if not pf and (pn=paramcount2) then lastparam:=true
    else lastparam:=false;
end;

{******** param2ptp *************************************************}
procedure tparams.param2ptp(var p:tparamdef);
var l   : longint;
    i   : longint;
    s   : shortstring;
    sa  : ansistring;
    e,f : extended;
    re  : real;
    si  : single;
    dou : double;
    d   : dword;
begin
    with p do
    begin
        if pt=nil then paramerr(0,'no variable given');
        case tp of
            tpstr:
                begin
                    s:=getnextparam;
                    copytp(@s,p);
                end;
            tpastr, tpaastr:
                begin
                    sa:=getnextparam;
                    copytp(sa,p);
                end;
            tpsmallint :
                begin
                    s:=getnextparam;
                    val(s,l,i);
                    if i<>0 then paramerr(1,s+' not an integer');
                    if (l<-32768) or (l>32767) then paramerr(1,s+' not in range (-32768..32767)');
                    copytp(@l,p);
                end;
            tpword :
                begin
                    s:=getnextparam;
                    val(s,l,i);
                    if i<>0 then paramerr(1,s+' not an integer');
                    if (l<0) or (l>65535) then paramerr(1,s+' not in range (0..65535)');
                    copytp(@l,p);
                end;
            tpbyte :
                begin
                    s:=getnextparam;
                    val(s,l,i);
                    if i<>0 then paramerr(1,s+' not an integer');
                    if (l<0) or (l>255) then paramerr(1,s+' not in range (0..255)');
                    copytp(@l,p);
                end;
            tpshortint :
                begin
                    s:=getnextparam;
                    val(s,l,i);
                    if i<>0 then paramerr(1,s+' not an integer');
                    if (l<128) or (l>127) then paramerr(1,s+' not in range (-128..127)');
                    copytp(@l,p);
                end;
            tplint :
                begin
                    s:=getnextparam;
                    val(s,l,i);
                    if i<>0 then paramerr(1,s+' not an integer');
                    copytp(@l,p);
                end;
            tpdword :
                begin
                    s:=getnextparam;
                    val(s,d,i);
                    if i<>0 then paramerr(1,s+' not an integer');
                    copytp(@d,p);
                end;
            tpboolean :
                begin
                    i:=1;
                    copytp(@i,p);
                end;
            tpreal, tpsingle, tpdouble, tpextended :
                begin
                    s:=getnextparam;
                    f:=1.0;
                    if s[length(s)]='%' then
                    begin
                        f:=0.01;
                        delete(s,length(s),1);
                    end;
                    val(s,e,i);
                    if i<>0 then paramerr(1,s+' not a number');
                    case tp of
                        tpreal :
                            begin
                                re:=e*f;
                                copytp(@re,p);
                            end;
                        tpsingle :
                            begin
                                si:=e*f;
                                copytp(@si,p);
                            end;
                        tpdouble :
                            begin
                                dou:=e*f;
                                copytp(@dou,p);
                            end;
                        tpextended :
                            begin
                                e:=e*f;
                                copytp(@e,p);
                            end;
                    end;
                end;
            else paramerr(0,'unknown parameter type');
        end;
    end;
end;

{******** param2ptp *************************************************}
function tparams.paramfmt(tp:longint):shortstring;
begin
    case tp of
        tpstr, tpastr, tpaastr :
            paramfmt:='<string>';
        tpdword, tpsmallint, tpword, tpbyte, tpshortint, tplongint :
            paramfmt:='<nnn>';
        tpreal, tpsingle, tpdouble, tpextended :
            paramfmt:='<nn.n>';
        else paramfmt:='';
    end;
end;

{******** run *******************************************************}
procedure tparams.run;
var i        : longint;
    s        : shortstring;

    procedure defvar(var p:tparamdef);
    var i:longint;
    begin
{$ifdef strengthvareval}
        for i:=1 to oc do
            if (p.pt=opts[i].pt) and (p.tp=opts[i].tp) then opts[i].vd:=true;
        for i:=1 to pc do
            if (p.pt=params[i].pt) and (p.tp=params[i].tp) then params[i].vd:=true;
{$else}
        for i:=1 to oc do
            if p.pt=opts[i].pt then opts[i].vd:=true;
        for i:=1 to pc do
            if p.pt=params[i].pt then params[i].vd:=true;
{$endif}
    end;

    procedure getdefault(var p:tparamdef;it:longint);
    var i : longint;
        b : boolean;
    begin
        with p do
            if (default<>nil) and (it<200) and not vd then
            begin
                i:=1;
                b:=true;
                while (i<=oc) and not vd do
                begin
{$ifdef strengthvareval}
                    if (opts[i].pt=default) and (opts[i].tp=tp) then
{$else}
                    if opts[i].pt=default then
{$endif}
                    begin
                        b:=false;
                        if not opts[i].vd and (opts[i].default<>nil) then getdefault(opts[i],it+1);
                        if opts[i].vd then
                        begin
                            defvar(p);
                            copytp(opts[i].pt,p);
                        end;
                    end;
                    inc(i);
                end;
                i:=1;
                while (i<=pc) and not vd do
                begin
{$ifdef strengthvareval}
                    if (params[i].pt=default) and (params[i].tp=tp) then
{$else}
                    if params[i].pt=default then
{$endif}
                    begin
                        b:=false;
                        if not params[i].vd and (params[i].default<>nil) then getdefault(params[i],it+1);
                        if params[i].vd then
                        begin
                            defvar(p);
                            copytp(params[i].pt,p);
                        end;
                    end;
                    inc(i);
                end;
                if b then
                begin
                    defvar(p);
                    copytp(default,p);
                end;
            end;
    end;

begin
    pc2:=0;
    while not lastparam do
    begin
        s:=getnextparam2;
        if s<>'' then
        begin
            if (s[1]='-') and (length(s)>1) then
            begin
                op:=true;
                s:=getnextparam;
                oc2:=optid(s);
                if oc2=0 then paramerr(2,s+': unknown option');
                if not overwrite and (opts[oc2].vd) then paramerr(1,'illegal used (maybe defined twice or incompatible with an other argument)');
                param2ptp(opts[oc2]);
                opts[oc2].pd:=true;
                defvar(opts[oc2]);
            end
            else begin
                op:=false;
                if params[pc2].tp<>tpaastr then inc(pc2);
                if pc2>pc then paramerr(2,'to much parameters');
                param2ptp(params[pc2]);
                params[pc2].pd:=true;
                defvar(params[pc2]);
            end;
        end;
    end;

    if helpf then
    begin
        showhelp;
        halt(0);
    end;
    for i:=1 to oc do
        if not opts[i].vd and (opts[i].default<>nil) then getdefault(opts[i],0);
    for i:=1 to pc do
        if not params[i].vd and (params[i].default<>nil) then getdefault(params[i],0);
    for i:=1 to oc do
        raprocs[opts[i].ra](opts[i]);
    for i:=1 to pc do
        raprocs[params[i].ra](params[i]);
end;

{******** defparam **************************************************}
function tparams.defparam(const str,des:shortstring;pi,pj:pointer;tp,ra:byte):longint;
var i : longint;
    b : boolean;
begin
    defparam:=0;        // makes the compiler happy
    if pi=nil then paramerr(0,str+': pointer to variable must not be nil');
    if str[1]='<' then
    begin
        b:=false;
        for i:=1 to pc do
            if params[i].str=str then b:=true;
        if b then paramerr(0,str+': already defined');
        inc(pc);
        defparam:=pc;
        params[pc].pt:=pi;
        params[pc].default:=pj;
        params[pc].str:=str;
        params[pc].des:=des;
        params[pc].tp:=tp;
        params[pc].pd:=false;
        params[pc].vd:=false;
        params[pc].ra:=ra;
        params[pc].fmt:=str;
    end
    else if str[1]='-' then
    begin
         b:=false;
         for i:=1 to oc do
            if opts[i].str=str then b:=true;
        if b then paramerr(0,str+': already defined');
        inc(oc);
        defparam:=oc;
        opts[oc].pt:=pi;
        opts[oc].default:=pj;
        opts[oc].str:=str;
        opts[oc].des:=des;
        opts[oc].tp:=tp;
        opts[oc].pd:=false;
        opts[oc].vd:=false;
        opts[oc].ra:=ra;
        opts[oc].fmt:=paramfmt(tp);
    end
    else paramerr(0,str+': neither parameter (  <...> ) nor option ( -... )');
end;

function tparams.defparam(const str,des:shortstring;pi:pointer;tp,ra:byte):longint;
begin
    result:=defparam(str,des,pi,nil,tp,ra);
end;

function tparams.defparam(const str,des:shortstring;var d:double):longint;
begin
    defparam:=defparam(str,des,@d,tpdouble,0);
end;

function tparams.defparam(const str,des:shortstring;var s:single):longint;
begin
    defparam:=defparam(str,des,@s,tpsingle,0);
end;

function tparams.defparam(const str,des:shortstring;var b:boolean):longint;
begin
    defparam:=defparam(str,des,@b,tpboolean,0);
end;

function tparams.defparam(const str,des:shortstring;var d:longint):longint;
begin
    defparam:=defparam(str,des,@d,tplongint,0);
end;

function tparams.defparam(const str,des:shortstring;var b:byte):longint;
begin
    defparam:=defparam(str,des,@b,tpbyte,0);
end;

function tparams.defparam(const str,des:shortstring;var w:word):longint;
begin
    defparam:=defparam(str,des,@w,tpword,0);
end;

function tparams.defparam(const str,des:shortstring;var s:shortstring):longint;
begin
    defparam:=defparam(str,des,@s,tpstr,0);
end;

function tparams.defparam(const str,des:shortstring;var s:ansistring):longint;
begin
    defparam:=defparam(str,des,@s,tpastr,0);
end;

function tparams.defparam(const str,des:shortstring;var a:taastr):longint;
begin
    defparam:=defparam(str,des,@a,tpaastr,0);
end;


{******** showhelp **************************************************}
procedure tparams.deffmt(str:shortstring; const fmt:shortstring);
var i:longint;
begin
    str:=condupstr(str);
    for i:=1 to oc do
    if condupstr(opts[i].str)=str then
    begin
        opts[i].fmt:=fmt;
        exit;
    end;
    paramerr(0,str+': not defined');
end;

{******** showhelp **************************************************}
procedure tparams.showhelp;
var i,j,k:longint;
begin
    writeln(stderr);
    if longdescription<>'' then write(stderr,longdescription);
    write(stderr,'Usage: '+paramstr2(0)+' [<options>] [@<parameter file>] ');
    if paramusage='' then
    begin
        for i:=1 to pc do
            write(stderr,' '+params[i].str);
        writeln(stderr);
    end
    else writeln(stderr,paramusage);
    writeln(stderr);
    writeln(stderr,'Parameters:');
    for i:=1 to pc do
        with params[i] do
        begin
            write(stderr,'  '+str+' ');
            for j:=9 downto length(str) do
                write(stderr,#32);
            j:=0;
            while j<length(des) do
            begin
                j+=1;
                if (j<length(des)) and (copy(des,j,2)='\\') then
                begin
                    write(stderr,#13#10+'            ');
                    j+=1;
                end
                else write(stderr,des[j]);
            end;
            writeln(stderr);
        end;
    writeln(stderr);
    writeln(stderr,'Options:');

    k:=0;
    for i:=1 to oc do
        with opts[i] do
            if length(str)+length(fmt)>k then k:=length(str)+length(fmt);
    k+=4;

    for i:=1 to oc do
        with opts[i] do
        begin
            write(stderr,'  ',str,' ',fmt);
            for j:=length(str)+length(fmt)+4 to k do
                write(stderr,#32);
            j:=0;
            while j<length(des) do
            begin
                j+=1;
                if (j<length(des)) and (copy(des,j,2)='\\') then
                begin
                    write(stderr,#13#10,' ':k);
                    j+=1;
                end
                else write(stderr,des[j]);
            end;
            writeln(stderr);
        end;
    writeln(stderr);
end;

{******** optid *****************************************************}
function tparams.optid(s:shortstring):longint;
var i,j:longint;
begin
    j:=0;
    s:=condupstr(s);
    for i:=1 to oc do
    if condupstr(opts[i].str)=s then
    begin
        j:=i;
        break;
     end;
    optid:=j;
end;

{******** vardef ****************************************************}
function tparams.vardef(p:pointer):boolean;
var i:longint;
begin
    vardef:=false;
    for i:=1 to oc do
        if (p=opts[i].pt) and (opts[i].vd) then vardef:=true;
    for i:=1 to pc do
        if (p=params[i].pt) and (params[i].vd) then vardef:=true;
end;

{******** pramdef ***************************************************}
function tparams.paramdef(const s:shortstring):boolean;
begin
    paramdef:=opts[optid(s)].pd;
end;

{******** paramerr **************************************************}
procedure tparams.paramerr(b:byte;const s:ansistring);
{
    0  error in source code
    1  user error; shows help, which param/opt
    2  user error; shows only help
}
begin
    case b of
        0:
            write(stderr,'error in source code: ');
        1:
            begin
{$ifdef showhelponerr}
                showhelp;
{$endif}
                if op then write(stderr,opts[oc2].str+': ')
                else write(stderr,params[pc2].str+': ');
            end;
{$ifdef showhelponerr}
        2 : showhelp;
{$endif}
    end;
    writeln(stderr,s);
    halt(paramerrnum);
end;

procedure tparams.paramerr(const s:ansistring);
begin
    paramerr(2,s);
end;

{******** copytp ****************************************************}
// non-reference counted types
procedure tparams.copytp(q:pointer;var p:tparamdef);
begin
    if (p.tp<>tpastr) and (p.tp<>tpaastr) then move(q^,p.pt^,tpsizes[p.tp]);
end;

// reference counted types
procedure tparams.copytp(sa:ansistring;var p:tparamdef);
begin
    if p.tp=tpastr then pansistring(p.pt)^:=sa
    else if p.tp=tpaastr then
        with paastr(p.pt)^ do
        begin
            if size<0 then size:=0;
            if size>high(aa) then setlength(aa,size+aasize);
            aa[size]:=sa;
            inc(size);
        end;
end;

{******** setraproc *************************************************}
procedure tparams.setraproc(i:byte;p:traproc);
begin
    raprocs[i]:=p;
end;


{******** condupstr *************************************************}
function tparams.condupstr(const s:shortstring):shortstring;
var i:longint;
begin
    condupstr:=s;
    if ignorecase then
        for i:=1 to length(s) do
            condupstr[i]:=upcase(s[i]);
end;

begin
    paramstr2:=@paramstr;
    paramcount2:=paramcount;
end.
