{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name colorize}
{@description Colorize image using color map}

{$I directives.pas}
uses simage,getparam;
var par                 : tparams;
    cc                  : CSImg_Config;
    fc,fi,fd            : TSImg_FPImage;
    fnc,fni,fno,fnl,fnf : ansistring;
    fnu,fnd             : ansistring;
    mi                  : CSImg_MetaInfo;
    ql,qc,ml,mu         : double;
    me,mb,ms            : double;
    lx,ly,imap,nl,lm    : longint;
    minv                : boolean;

    bmap                : array of simgfp;
    map                 : TSImg_ColorMap;
    i                   : longint;
    r,s                 : double;

begin
    fc.pt:=nil;
    fi.pt:=nil;
    with par do
    begin
        init;
        paramusage:='<color image> [<infile> [<outfile>]]';
        fni:='';        defparam('<infile>','Luminance image (default: stdin)',fni);
        fno:='';        defparam('<outfile>','Destination image (default: stdout)',fno);
        fnf:='';        defparam('-f','Color field used to determine color',fnf); deffmt('-f','<infile>');
        fnu:='';        defparam('-fu','Color field determined by compression ratio where this image is the uncompressed component fu.\\'+
                                       'If -fc is not specified the luminance image is used as compressed component fc.\\'+
                                       'Color field is 0 if fc<=0 or fu<=fc, otherwise 1-fc/fu.', fnu); deffmt('-fu','<infile>');
        fnc:='';        defparam('-fc','Compressed component for color field, see -fu.',fnc);  deffmt('-fc','<infile>');
        fnd:='';        defparam('-fd','Color de-saturation image. Range id 0 (no de-saturation) to 1 (full de-saturation).\\'+
                                       'If -fu is given de-saturation is calculated by fu<0 ? 1 : fd<0 ? 0 : min(1,fd/fu).',fnd);
        ql:=-1;         defparam('-ql','Black level quantile for luminance image',ql);
        qc:=-1;         defparam('-qc','Black level quantile for image given by -fc,',qc);
                        defparam('-o','Output image',fno); deffmt('-o','<outfile>');

        imap:=3;        defparam('-m','Base color map. 0: blue to white, 1: red to white 2: black to blue to white\\'+
                                      '3: optimized blue to white, 4: optimized black to blue to white. 5: optimized heat map. Default: 3',imap);
        ml:=-1;         defparam('-ml','Quantile of color field where color map starts. By default fixed value 0 is used.',ml);
        mu:=-1;         defparam('-mu','Quantile of color field where color map ends. By default fixed value 1 is used.',mu);
        minv:=false;    defparam('-mi','Invert color map.',minv);
        me:=0.9;        defparam('-me','Color map histogram equalization. Valid: 0..1, Default: 0.9.',me);
        mb:=0;          defparam('-mb','Color map histogram balance. <0 overweight smaller values, >0 overweight larger values\\'+
                                       'No effect at -me 0. Valid: -1..1. Default: 0',mb);
        ms:=1.0;        defparam('-ms','Saturation. Valid: 0..1, Feasible: 0.5..1, Default: 1',ms);

        fnl:='';        defparam('-l','Legend output image',fnl); deffmt('-l','<outfile>');
        lx:=800;        defparam('-lx','Legend width. Default: 800',lx);
        ly:=50;         defparam('-ly','Legend height. Default: 50',ly);
        nl:=5;          defparam('-ln','Number of labels. <2 to disable labels. Default: 5.',nl);
        lm:=4;          defparam('-lm','Label mode: >0; number of valid digits, 0: percent, <0 relative to maximum in percent. Default: 4',lm);
                        defparam('-h','Help',helpf);
        run;

        cc:=CSImg_Config.create(SImg_Config_Default);
        cc.debug(cc.name+': values: fni='+fni+' fno='+fno+' fnf='+fnf+' fnu='+fnu+' fnc='+fnc+' fnd='+fnd+' ql='+real2str(ql)+' qc='+real2str(qc)+' map='+int2str(imap)+' ml='+real2str(ml)+' mu='+real2str(mu)+' mi='+boolean2str(minv)+
            ' me='+real2str(me)+' mb='+real2str(mb)+' ms='+real2str(ms)+' fnl='+fnl+' lx='+int2str(lx)+' ly='+int2str(ly)+' nl='+int2str(nl)+' lm='+int2str(lm));

        if (fnf<>'') and (fnu<>'') then paramerr('-f and -fu can''t be used together');
        if (fnf='') and (fnu='') then paramerr('Either -f or -fu must be given');
        if (fnc<>'') and (fnu='') then paramerr('-fc can''t be used without -fu');

        { color desaturation field }
        fd.pt:=nil;
        if fnd<>'' then
        begin
            cc.Info(cc.name+': Reading color de-saturation information '+fnd);
            SImg_ReadImageFile(cc, fnd, fd);
            if fd.nc<>1 then cc.error('Color de-saturation image must be a 1-color image');
        end;

        { color field }
        if fnf<>'' then
        begin
            cc.Info(cc.name+': Reading color field '+fnf);
            SImg_ReadImageFile(cc, fnf, fc);
            if fc.nc<>1 then cc.error(cc.name+': Color field must be a 1-color image');
            if (fd.pt<>nil) and ((fc.nx<>fd.nx) or (fc.ny<>fd.ny)) then cc.error(cc.name+': Size mismatch between color field and de-saturation image');
        end else
        begin
            cc.Info(cc.name+': Reading uncompressed image '+fnu);
            SImg_ReadImageFile(cc, fnu, fc);
            if fc.nc<>1 then cc.error('Uncompressed image must be a 1-color image');
            if (fd.pt<>nil) and ((fc.nx<>fd.nx) or (fc.ny<>fd.ny)) then cc.error(cc.name+': Size mismatch between uncompressed image and color de-saturation image');
            if fnc<>'' then
            begin
                cc.Info(cc.name+': Reading compressed image '+fnc);
                SImg_ReadImageFile(cc, fnc, fi);
                if fi.nc<>1 then cc.error('Compressed image must be a 1-color image');
                if (fi.nx<>fc.nx) or (fi.ny<>fc.ny) then cc.error('Size mismatch between compressed and uncompressed image');
                if qc>=simg_mina then
                begin
                    r:=SImg_GetQuantile(cc,fi,qc);
                    for i:=0 to fi.size-1 do
                    begin
                        s:=fi.pt[i]-r;
                        if s<0 then s:=0;
                        fi.pt[i]:=r;
                    end;
                end;
                SImg_ColorFieldRatio(cc, fc,fi,fd, fc);
                SImg_Free(cc, fi);
            end;
        end;

        { luminance image }
        fi.pt:=nil;
        mi:=CSImg_MetaInfo.create();
        cc.Info(cc.name+': Reading luminance image '+fni);
        SImg_ReadImageFile(cc, fni, fi, mi);
        if fi.nc<>1 then cc.error('Luminance image must be a 1-color image');
        if (fi.nx<>fc.nx) or (fi.ny<>fc.ny)  then cc.error('Size mismatch between luminance image and '+select(fnf<>'','color field','uncompressed image'));
        mi.deleteRaw;
        if ql>=simg_mina then
        begin
            r:=SImg_GetQuantile(cc,fi,ql);
            for i:=0 to fi.size-1 do
            begin
                s:=fi.pt[i]-r;
                if s<0 then s:=0;
                fi.pt[i]:=r;
            end;
        end;
        if (fnu<>'') and (fnc='') then SImg_ColorFieldRatio(cc, fc,fi,fd, fc);
//      SImg_WriteImageFile(cc, 'colorize-fc.fits', fc);

        { color map }
        if imap<=0 then bmap:=bmap.create(
                0,  1,  0, 0, 1,
                1,  1,  0, 1, 1,
                2,  1,  0, 1, 0,
                3,  1,  1, 1, 0,
                4,  1,  1, 0, 0,
                5,  1,  1, 0, 1,
                6,  1,  1, 1, 1
            )
        else if imap=1 then bmap:=bmap.create(
                0,  1,  1, 0, 0,
                1,  1,  1, 1, 0,
                2,  1,  0, 1, 0,
                3,  1,  0, 1, 1,
                4,  1,  0, 0, 1,
                5,  1,  1, 0, 1,
                6,  1,  1, 1, 1
            )
        else if imap=2 then bmap:=bmap.create(
               -1,    0,  0, 0, 0,
                0,  0.5,  0, 0, 1,
                1, 0.75,  0, 1, 1,
                2, 0.88,  0, 1, 0,
                3, 0.94,  1, 1, 0,
                4, 0.97,  1, 0, 0,
                5, 0.99,  1, 0, 1,
                6,    1,  1, 1, 1
            )
        else if imap=3 then bmap:=bmap.create(
                0.2,  1,  0.35, 0.3, 1,
                1,  1,  0, 1, 1,
                2,  1,  0.2, 1, 0.2,
                3,  1,  1, 1, 0,
                4,  1,  1, 0, 0,
                5,  1,  1, 0, 1,
                6,  1,  1, 1, 1
            )
        else if imap=4 then bmap:=bmap.create(
               -1,    0,  0, 0, 0,
                0.2,  0.5,  0.3, 0.2, 1,
                1, 0.75,  0, 1, 1,
                2, 0.88,  0.2, 1, 0.2,
                3, 0.94,  1, 1, 0,
                4, 0.97,  1, 0, 0,
                5, 0.99,  1, 0, 1,
                6,    1,  1, 1, 1
            )
        else bmap:=bmap.create(
                0.0,  1,  1, 0, 0.5,
                0.25, 1,  1, 0, 0,
                1.4,  1,  1, 1, 0,
                2.0,  1,  0.8, 0.9, 1
            );

        cc.Info(cc.name+': Calculate color map');
        SImg_TransformColorMap(cc, fi,fc, bmap, ml,mu, minv, me,mb,ms, map);
        cc.info('Color field range: '+real2str(map.l)+' .. '+real2str(map.u));

        { colorize + write }
        cc.Info(cc.name+': Colorize image');
        SImg_Colorize(cc, fi,fc,fd, fi, map,true);
        cc.Info(cc.name+': Write output image '+fno);
        SImg_WriteImageFile(cc, fno, fi, mi);
        SImg_Free(cc,fc);
        SImg_Free(cc,fi);
        SImg_Free(cc,fd);

        { legend }
        if fnl<>'' then
        begin
            cc.Info(cc.name+': Drawing and writing legend '+fnl);
            SImg_ColorizeTest(cc, fi, map, lx,ly*4);
            SImg_ColorizeLegend(cc, fi, map, lx,ly,nl, lm);
            SImg_WriteImageFile(cc, fnl, fi);
            SImg_Free(cc,fi);
        end;


        done;
        end;

    cc.destroy;
end.
