{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name convert2}
{@description Misc simple image processing routines}

{$I directives.pas}
uses simage,getparam,sysutils;
var par                 : TParams;
    cc                  : CSImg_Config;
    img                 : TSImg_FPImage;
    mi                  : CSImg_MetaInfo;
    zl                  : double;
    mz,mo,ix,iy,ic,ac,r1: boolean;
    fi,fo,v,c           : ansistring;
    i,rot               : longint;
    mean,cf,co,r        : double;
    a,b                 : simgfp;
    d                   : ^simgfp;
    zld                 : boolean;

begin
    cc:=CSImg_Config.create(SImg_Config_Default);
    with par do
    begin
        init;
        paramusage:='[<infile> [<outfile>]]';
        fi:='';      defparam('<infile>','Source image (default: stdin)',fi);
        fo:='';      defparam('<outfile>','Destination image (default: stdout)',fo);
                     defparam('-h','Help',helpf);
        ac:=false;   defparam('-ac','Add all colors (monochrome conversion)',ac);
        rot:=0;      defparam('-r','Rotate image clockwise in 90° steps',rot);
        ix:=false;   defparam('-ix','Flip X',ix);
        iy:=false;   defparam('-iy','Flip Y',iy);
        ic:=false;   defparam('-ic','Flip color',ic);

        zl:=0.1;     defparam('-zl','Zero level quantile',zl);
        mz:=false;   defparam('-mz','Mean value is zero',mz);
        mo:=false;   defparam('-mo','Mean value is one',mo);
        r1:=false;   defparam('-1r','Calculate reciprocal flatfield image (e.g. for offset correction using the -bi parameter of sdbadd. Implies -mo)',r1);
                     defparam('-o','Destination image',fo); deffmt('-o','<outfile>');
        run;

        cc.debug(cc.name+': values: source='+fi+' dest='+fo+' rot='+int2str(rot)+' ix='+boolean2str(ix)+' iy='+boolean2str(iy)+' ic='+boolean2str(ic)+condstring(paramdef('-zl'),'zl='+real2str(zl),'')+' mz='+boolean2str(mz)+' mo='+boolean2str(mo)+' r1='+boolean2str(mo));

        zld:=paramdef('-zl');

        if zld and mz then
        begin
            cc.warning(cc.name+': -zl and -mz can''t be defined together, ignoring both');
            zld:=false;
            mz:=false;
        end;
        if mo and mz then
        begin
            cc.warning(cc.name+': -mo and -mz can''t be defined together, ignoring both');
            mo:=false;
            mz:=false;
        end;
        if r1 and (zld or mz) then
        begin
            cc.warning(cc.name+': -r1 cannot be defined with -zl or -mz, ignoring them');
            zld:=false;
            mz:=false;
        end;

        cc.info(cc.name+': Reading input image '+fi);
        mi:=CSImg_MetaInfo.create();
        SImg_ReadImageFile(cc,fi,img,mi);

        if ix or iy or (round(rot/90) and 3<>0) then
        begin
            cc.info(cc.name+': Rotating and/or flipping image');
            SImg_RotateFlip(cc, img,img, rot,ix,iy);
        end;

        mean:=1e-100;
        if mo or mz or r1 then
        begin
            d:=img.pt;
            for i:=0 to img.size-1 do
                mean+=d[i];
            mean/=img.size;
        end;

        if r1 then
        begin
            a:=mean*1e-6;
            if a<simg_mina then cc.error(cc.name+': Mean value of input image must be positive');
            mi.delete('EGAIN');
            mi.delete('RDNOISE');
            mi.delete('DATAMIN');
            mi.delete('DATAMAX');
            mi.delete('BUNIT');
            for i:=0 to img.size-1 do
            begin
                b:=img.pt[i];
                if b<a then b:=a;
                img.pt[i]:=mean/b;
            end;
        end
        else begin
            co:=0;
            if zld or mz then
            begin
                mi.delete('EGAIN');
                if mz then co:=mean;
                if zld then
                begin
                    cc.info(cc.name+': Calculation of '+ real2str(zl*100,4) + '% quantile');
                    co:=SImg_GetQuantile(cc,img,zl);
                end;
            end;

            cf:=1;
            if mo then
            begin
                cf:=1/(mean-co);
                mi.delete('EGAIN');
                mi.delete('RDNOISE');
            end;

            if mo or mz or zld then
            begin
                d:=img.pt;
                a:=co; b:=cf;
                for i:=0 to img.size-1 do
                    d[i]:=(d[i]-a)*b;
            end;
        end;

        if ac and (img.nc>1) then
        begin
            if mi.get('RDNOISE',r,@c) then mi.add('RDNOISE' ,real2str(sqrt(img.nc)*r,15), c);
            mi.delete('DATAMIN');
            mi.delete('DATAMAX');
            cc.info(cc.name+': Adding colors');
            SImg_AddColors(cc,img,img);
        end;

        try
            if mi.get('BAYERPAT',v,@c) then mi.add('BAYERPAT',''''+SImg_BayerFlipPattern(cc,ix and (img.nx and 1=0), iy and (img.ny and 1=0), trimquotes(v))+'''', c);
        except
            SImg_OutputCatchedException(cc);
        end;
        if mi.get('DATAMIN',r,@c) then mi.add('DATAMIN', real2str((r-co)*cf,15), c);
        if mi.get('DATAMAX',r,@c) then mi.add('DATAMAX', real2str((r-co)*cf,15) ,c);
        if (abs(cf-1)>1e-9) and mi.get('BUNIT',v,@c) then mi.add('BUNIT', '''' + real2str(cf,8) + '*' + trimquotes(v) + '''' ,c);

        cc.info(cc.name+': Writing output image '+fo);
        SImg_WriteImageFile(cc,fo,img,mi);
        SImg_Free(cc,img);

        done;
    end;
    cc.destroy;
end.
