{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name fitsinfo}
{@description Obtain information from FITS headers}
{@longdescription For drift computation two peak lists with telescope position and observation time are required First source list must contain a WCS header.}

{$I directives.pas}
uses simage,getparam,sysutils;
var par                            : TParams;
    cc                             : CSImg_Config;
    mi,mj                          : CSImg_MetaInfo;
    ri,rj                          : TSImg_RDImage;
    wcs                            : TSImg_WCS;
    ka                             : TAAStr;
    sm,rs,ds,tp,wp,pp,dp,dn,sc,fr  : boolean;
    xp,yp,xd,yd,dx,dy,dr,dd        : double;
    ifn,dfn,afn,ofn,va,co          : ansistring;
    tra,tdec,wra,wdec              : double;
    tra2,tdec2,wra2,wdec2          : double;
    wcs_nx,wcs_ny,nx,ny,a,b        : double; // wcs image width, real image width
    t1,t2                          : int64;
    res                            : TSImg_AlignmentParams;
    con                            : TSImg_AlignmentConstraints;
    i                              : longint;

function nsm(s:ansistring):ansistring;
begin
    if sm
    then result:=''
    else result:=s;
end;

function format(s:boolean; cf:double; d:double):shortstring;
var i,j   : longint;
begin
    if s then
    begin
        d/=cf;
        result:='';
        if d<0 then
        begin
            d:=-d;
            result:='-';
        end;
        for i:=0 to 1 do
        begin
            j:=trunc(d); d:=(d-j)*60;
            result+=int2str(j)+':';
        end;
        result+=real2str(d,-3);
    end
    else result:=real2str(d,-5);
end;

begin
    cc:=CSImg_Config.create(SImg_Config_Default);
    with par do
    begin
        init;
        paramusage:='[<infile> [<outfile>]';
        ifn:='';     defparam('<infile>','Input file (default: stdin)',ifn);
        ofn:='';     defparam('<outfile>','Output for modified alignment data if -a is given, default: stdout',ofn);
                     defparam('-h','Help',helpf);
        sm:=false;   defparam('-s','Script mode: Output results only and suppress warnings',sm);
        ka.size:=0;  defparam('-k','Return value for given keyword or '''' inf not defined. Can be specified multiple times.',ka);
        rs:=false;   defparam('-rs','Output RA coordinates in hours and sexagesimal format (H:M:S, default: degree, decimal)',rs);
        ds:=false;   defparam('-ds','Output DEC coordinates in degrees and sexagesimal format (D:M:S, default: degree, decimal)',ds);
        tp:=false;   defparam('-tp','Get telescope RA and DEC coordinates',tp);
        wp:=false;   defparam('-wp','Get WCS RA and DEC coordinates of image center',wp);
        sc:=false;   defparam('-sc','Return pixel scale',sc);
        fr:=false;   defparam('-fr','Return field rotation of x-axis and y-axis',fr);
        xp:=0;       defparam('-x','Return coordinates of a pixel specified by -x and -y (default: 0)',xp);
        yp:=0;       defparam('-y','Return coordinates of a pixel specified by -x and -y (default: 0)',yp);
        xd:=0;       defparam('-xd','Return differential coordinates of a differential pixel position specified -xd and -yd (default: 0)',xd);
        yd:=0;       defparam('-yd','Return differential coordinates of a differential pixel position specified -xd and -yd (default: 0)',yd);
        dfn:='';     defparam('-d','Return telescope drift in RA and DEC in arcsec/hour using a 2nd source list', dfn);   deffmt('-d','<infile>');
        dn:=false;   defparam('-dn','Output negative drift values, i.e. values for drift compensation', dn);
        dx:=0;       defparam('-dx','Desired X drift (residual) in pixel/hour. Default: 0', dx);
        dy:=0;       defparam('-dy','Desired Y drift (residual) in pixel/hour. Default: 0', dy);
        afn:='';     defparam('-a','Add drift given by -dr and -dd to that alignment data. <infile> if used as reference and must contain WCS data.', afn);   deffmt('-a','<infile>');
        dr:=0;       defparam('-dr','RA object/viewpoint drift drift in arcsec/hour used if -a is given. Default: 0', dr);
        dd:=0;       defparam('-dd','DEC object/viewpoint drift drift in arcsec/hour used if -a is given. Default: 0', dd);
                     defparam('-o','Output for modified alignment data if -a is given',ofn); deffmt('-o','<outfile>');

        run;
        pp:=paramdef('-x') or paramdef('-y');
        dp:=paramdef('-xd') or paramdef('-yd');

        done;
    end;

    // read meta info
    ri.pt:=nil;
    mi:=CSImg_MetaInfo.create;
    if dfn='' then SImg_RHeader(cc, ifn, ri, mi)
    else SImg_ReadImageFile(cc, ifn, ri, mi);

    // return keywords
    for i:=0 to ka.size-1 do
    begin
        co:='';
        if not mi.get(ka.aa[i],va,@co) then va:='''''';
        writeln(nsm(ka.aa[i]+' = '), va, select(sm or (co=''), '', ' / '+co));
    end;

    // telescope position
    if tp or (dfn<>'') then
    begin
        if not mi.getRaDec(tra,tdec) then cc.error('fitsinfo: Cannot determine telescope position because keyword ''RA''/''OBJCTRA'' and ''DEC''/''OBJCTDEC'' are not defined.')
        else if tp then writeln(nsm('Telescope coordinates: RA='), format(rs,15,tra), ' ', nsm('   DEC='), format(ds,1,tdec));
    end;

    // read wcs data
    if wp or pp or dp or (dfn<>'') or sc or fr or (afn<>'') then
    begin
        if not SImg_GetWCSData(cc, mi, wcs) then cc.error('fitsinfo: Invalid or incomplete WCS data.');

        // image width and height
        nx:=ri.nx; ny:=ri.ny;
        if mi.get('IBNDXMAX',a) and mi.get('IBNDXMIN',b) then nx:=abs(a-b)+1;
        if mi.get('IBNDYMAX',a) and mi.get('IBNDYMIN',b) then ny:=abs(a-b)+1;

        // wcs width and height (may be scaled)
        wcs_nx:=nx; wcs_ny:=ny;
        mi.get('IMAGEW', wcs_nx);
        mi.get('IMAGEH', wcs_ny);

        // estimate  width and height using reference positions
        if wcs_nx<2 then cc.error('fitsinfo: Invalid or undefined image width');
        if wcs_ny<2 then cc.error('fitsinfo: Invalid or undefined image height');
        if nx<2 then nx:=wcs_nx;
        if ny<2 then ny:=wcs_ny;
//      writeln('nx=',nx:0:1,'  ny=',ny:0:1, '  wcs_nx=',wcs_nx:0:1,'  wcs_ny=',wcs_ny:0:1);

        // WCS position of given pixel
        if pp then
        begin
            SImg_GetRaDec(cc,  wcs,  xp*wcs_nx/nx, yp*wcs_ny/ny,  wcs_ny,  wra, wdec);
            writeln(nsm('WCS coordinates of image at FITS coordinates ('+real2str(xp,-1)+', '+real2str(yp,-1)+'): RA='), format(rs,15,wra), ' ', nsm('   DEC='), format(ds,1,wdec));
        end;

        // WCS position of center
        if wp or (dfn<>'') or dp or (afn<>'') then
        begin
            SImg_GetRaDec(cc, wcs, 0.5*(wcs_nx-1), 0.5*(wcs_ny-1), wcs_ny, wra, wdec);
            if wp then writeln(nsm('WCS coordinates of image center: RA='), format(rs,15,wra), ' ', nsm('   DEC='), format(ds,1,wdec));
        end;

        // WCS pixel scale
        if sc then
        begin
            SImg_GetPixelScale(wcs, xp,yp);
            writeln(nsm('Pixel scale in arcsecs: x: '), real2str(xp,4), nsm('   y:'), real2str(yp,4));
        end;

        // Field rodation
        if fr then writeln(nsm('Rotation of x-axis: '), real2str(arctan2(wcs.cd2_1,-wcs.cd1_1)*180/pi,-2), nsm('°'), ' ', nsm('   rotation of y-axis: '), real2str(arctan2(wcs.cd1_2,wcs.cd2_2)*180/pi,-2),nsm('°'));

        // differential WCS position
        if dp then
        begin
            SImg_GetRaDec(cc, wcs,  0.5*(wcs_nx-1)+xd*wcs_nx/nx, 0.5*(wcs_ny-1)+yd*wcs_ny/ny,  wcs_ny,  wra2, wdec2);
            writeln(nsm('Differential WCS coordinates: RA='), format(rs,15,wra2-wra), ' ', nsm('   DEC='), format(ds,1,wdec2-wdec));
        end;

        if dfn<>'' then
        begin
            if not mi.getDateTime(t1) then cc.error('fitsinfo: Invalid or missing observation time information.');

            rj.pt:=nil;
            mj:=CSImg_MetaInfo.create;
            SImg_ReadImageFile(cc, dfn, rj, mj);
            if not mj.getDateTime(t2) then cc.error('fitsinfo: Invalid or missing observation time information for 2nd source list.');
            if not mj.getRaDec(tra2,tdec2) then cc.error('fitsinfo: Cannot determine telescope position for 2nd source list because keyword ''RA''/''OBJCTRA'' and ''DEC''/''OBJCTDEC'' are not defined.');

            { calculate shift }
            SImg_ReadAlignmentParams(cc,res,mj);
            con.nq:=false;
            con.flip:=false;
            con.smin:=0.99;
            con.smax:=1.01;
            con.dmax:=sqrt(sqr(res.xmin-res.xmax)+sqr(res.ymin-res.ymax))*0.1;
            con.rx:=0;
            con.ry:=0;
            con.sh:=2;
            con.frc:=0;
            con.frd:=10;
            SImg_GetAlignment(cc, ri,rj, 0, res, con);
            a:=(t2-t1)/3600.0;
            res.ax-=dx*a;
            res.ay-=dy*a;

            // calculate drift
            a:=3600*3600/((t1-t2)+0.25);
{           writeln(stderr,'t1-t2=',t1-t2);
            writeln(stderr,'nx=',nx:0:1,'   ny=',ny:0:1); }
            SImg_GetRaDec(cc, wcs, 0.5*(wcs_nx-1)-res.ax*wcs_nx/nx, 0.5*(wcs_ny-1)-res.ay*wcs_ny/ny, wcs_ny,  wra2,wdec2);
{           writeln(stderr,'wra=',wra:0:6,'  wdec=',wdec:0:6);
            writeln(stderr,'wra2=',wra2:0:6,'  wdec2=',wdec2:0:6);
            writeln(stderr,'wra-wra2=',(wra-wra2)*3600:0:6,'  wdec-wdec2=',(wdec-wdec2)*3600:0:6);
            writeln(stderr,'(wra-wra2)*a=',(wra-wra2)*a:0:1,'  (wdec-wdec2)*a=',(wdec-wdec2)*a:0:1);
            writeln(stderr,'(tra-tra2)*a=',(tra-tra2)*a:0:1,'  (tdec-tdec2)*a=',(tdec-tdec2)*a:0:1); }

            wra2:=(wra-wra2-tra+tra2)*a;
            wdec2:=(wdec-wdec2-tdec+tdec2)*a;
            if dn then writeln(nsm('Telescope drift compensation: RA='), -wra2:0:1, ' ', nsm('   DEC='), -wdec2:0:1)
            else writeln(nsm('Telescope drift: RA='), wra2:0:1, ' ', nsm('   DEC='), wdec2:0:1);

            SImg_free(cc,ri);
            SImg_free(cc,rj);
            mj.destroy;
        end;

        if afn<>'' then
        begin
            if not mi.getDateTime(t1) then cc.error('fitsinfo: Invalid or missing observation time information.');

            rj.pt:=nil;
            mj:=CSImg_MetaInfo.create;
            SImg_RHeader(cc, afn, rj, mj);
            SImg_ReadAlignmentParams(cc,res,mj);
            if not mj.getDateTime(t2) then cc.error('fitsinfo: Invalid or missing observation time information for 2nd source list.');

            { calculate drift position of center in reference image }
            wra2:=wra+(t2-t1)/(3600*3600)*dr;
            wdec2:=wdec+(t2-t1)/(3600*3600)*dd;
            if not SImg_GetPos(wcs, wra,wdec, wcs_ny, xd,yd) then cc.error('fitsinfo: Singular WCS data');
            if not SImg_GetPos(wcs, wra2,wdec2, wcs_ny, dx,dy) then cc.error('fitsinfo: Singular WCS data');
            xd:=nx/wcs_nx*(dx-xd);
            yd:=ny/wcs_ny*(dy-yd);

            { transform to position in destination image }
            xp:=sqr(xd);
            yp:=sqr(yd);
            a:=res.d*(xp+yp) + res.c1*xd + res.c2*yd;
            xd+=xd*a + res.c3*xp + res.c5*yp;
            yd+=yd*a + res.c4*yp + res.c6*xp;
            res.ax+=res.b11*xd + res.b12*yd;
            res.ay+=res.b21*xd + res.b22*yd;

            cc.Info('fitsinfo: Writing alignment data '+ofn);
            SImg_WriteAlignmentParams(res,mj);
            SImg_WFITSHeader(cc,CSImg_FileOutput.init(ofn),mj).close(cc);
        end;

    end;

    mi.destroy;
    cc.destroy;
end.
