{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name getbg}
{@description Estimate background from an image set}

{$I directives.pas}
uses simage,getparam,sysutils;

const unmask_bncorr = 1;        { correction of minimum square background noise }
      unmask_lamb   = 2;        { smoothing factor relative to background noise in order to prevent overweighting small values }
      mosaic_approx = 2;        { boundary overweight in mosaic mode }
//      mosaic_approx = 1;      { boundary overweight in mosaic mode }

var par         : tparams;
    cc          : CSImg_Config;
    ofn,ifp,ps  : ansistring;
    ifn         : TAAStr;
    ss          : longint;
    fr,cs       : double;

    iset        : CSImg_ImageSet_Files;
    fi,fsi      : TSImg_FPImage;
    ei          : TSImg_RDImage;
    i,j         : longint;
    mi          : CSImg_MetaInfo;
    bi          : TSImg_BayerInfo;


begin
{ define getbg_debug}
    cc:=CSImg_Config.create(SImg_Config_Default);
    with par do
    begin
        init;
        paramusage:='<stack database> [<FWHM limits>]';
        ifn.size:=0;    defparam('-i','Input image, can be specified multiple times',ifn);      deffmt('-i','<image>');
        ifp:='';        defparam('-p','Define input files by file pattern',ifp);                deffmt('-p','<file pattern>');
        ps:='';         defparam('-b','Bayer pattern. If defined or detected input image is converted to RGB by downscaling. E.g. 12/23 or RG/GB (both are the same)', ps );
        ss:=8;          defparam('-s','Define sub-sampling size. Image is median filtered with diameter 2.4 times that size. Default: 8',ss);
        fr:=-1;         defparam('-r','Background filter radius in input image. Default: <smaller image dimension>/8',fr);
        cs:=0.25;       defparam('-c','Corner size for background filter relative to filter radius. Default: 0.25',cs);
        ofn:='';        defparam('-o','Output file (default: stdout)',ofn);                                     deffmt('-o','<outfile>');
                        defparam('-h','Help',helpf);
        run;
        done;
    end;
    cc.debug(cc.name+': ofn='+ofn+' ifn.size='+int2str(ifn.size)+' ifp='+ifp+'  ps='+ps+'  ss='+int2str(ss)+'  fr='+realstr(fr));

    if ss<2 then ss:=2;
    iset:=CSImg_ImageSet_Files.init(cc,ifp);
    for i:=0 to ifn.size-1 do
        iset.addfile(ifn.aa[i]);

    if iset.getMax<0 then
    begin
        cc.warning('No input defined -- nothing to do');
        SImg_Halt;
    end;


    { run 1 }
    fi.pt:=nil;
    fsi.pt:=nil;
    ei.pt:=nil;
    mi:=CSImg_MetaInfo.create();
    for i:=0 to iset.getmax do
    begin
        iset.getimage(cc, fi, mi, i);

        { process bayer pattern }
        if i=0 then
        begin
            if fr<0 then fr:=min(fi.nx,fi.ny)/8;
            fr/=ss;
            if (ps='') and mi.get('BAYERPAT',ifp) then ps:=trimquotes(ifp);
            if ps<>'' then
            begin
                SImg_BayerParsePattern(cc,ps,bi);
                cc.info('Treating input images as Bayer pattern images with pattern "'+ps+'"');
                fr*=0.5;
            end;
        end;
        if ps<>'' then SImg_BayerDownScale(cc, bi, fi,fi);

        { downscaling + median filter }
        SImg_Filter(cc, fi,fi, ss*1.2, -4, ss, 0.5);
{$ifdef getbg_debug}
        if i=0 then SImg_WriteImageFile(cc, 'getBG-fj0.fits', fi);
{$endif}

        { background filter }
        SImg_FBGFilter(cc, fi,fi,ei, fr,fr,cs, true,false,false);
{$ifdef getbg_debug}
        if i=0 then SImg_WriteImageFile(cc, 'getBG-fk0.fits', fi);
{$endif}

        { add }
        if i=0 then begin
            fsi:=fi;
            fi.pt:=nil;
        end
        else begin
            for j:=0 to fi.size-1 do
                fsi.pt[j]+=fi.pt[j];

        end;
    end;
    SImg_Free(cc,fi);
{$ifdef getbg_debug}
    SImg_WriteImageFile(cc, 'getBG-sum.fits', fsi);
{$endif}

    cc.info('Resizing'+select(ps<>'',', Bayer transforming','')+' and writing output image'+ofn);
    fr:=1/(iset.getmax+1);
    for j:=0 to fsi.size-1 do
        fsi.pt[j]*=fr;
    SImg_Resize(cc, fsi,fsi, fsi.nx/iset.nx, fsi.ny/iset.ny, 2);
    if ps<>'' then SImg_ToBayer(cc, fsi,fsi,bi);
    SImg_WriteImageFile(cc, ofn, fsi,mi);
    SImg_Free(cc, fsi);
    mi.destroy;
    iset.destroy;
end.
