{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name zocorrect}
{@description Bias-/flatfield corrections and calculation of pixel weight image}

{
    photon count image: 2nd pass (unit conversion)
}

{$I directives.pas}
uses simage,getparam;
var par                         : tparams;
    cc                          : CSImg_Config;
    fni,fno,fnw,fnw16,fno16     : ansistring;
    fnb,fn0,fn1,fnd,fng,fnm     : ansistring;
    ma,md                       : TAAStr;
    om,bl,fmin,fmax             : double;
    bayer,nc,ue,ues,us,o1,nn,nr : boolean;
    ob                          : boolean;

    mi                          : CSImg_MetaInfo;
    fi,f0,f1,fsn,fm,fg,fmb      : TSImg_FPImage;
    ri                          : TSImg_RDImage;
    str                         : ansistring;
    q,r,s,u,v,dmin,dmax         : double;
    rn,bn,dn,gain,adr           : double;
    i                           : longint;
    j                           : int64;
    bi,bj                       : ^double;

begin
{ define zocorrect_noiseinfo}
    with par do
    begin
        init;
        paramusage:='<infile1> [<infile2> [<outfile>]]';
        fni:='';      defparam('<infile>','Input image (default: stdin)',fni);
        fno:='';      defparam('<outfile>','Output image (default: stdout)',fno);
        fnw:='';      defparam('-w','Output pixel weight image',fnw);                   deffmt('-w','<outfile>');
        fnw16:='';    defparam('-w16','Output pixel weight image as 16 bit image',fnw16);       deffmt('-w16','<outfile>');

        fnb:='';      defparam('-b','Bias image',fnb);                                  deffmt('-b','<infile>');
        fn0:='';      defparam('-0','Fixed dark current image',fn0);                    deffmt('-0','<infile>');
        fnd:='';      defparam('-d','Variable dark current image',fnd);                 deffmt('-d','<infile>');
        fng:='';      defparam('-g','Variable amp glow image (only use it if necessary)',fng);                          deffmt('-g','<infile>');
        fn1:='';      defparam('-1','Flatfield image',fn1);                             deffmt('-1','<infile>');
        fnm:='';      defparam('-pm','Dead pixel mask as calculated by avgset -pm',fnm);        deffmt('-pm','<infile>');

        bl:=0.0;      defparam('-bl','Fixed bias level subtracted from image (default: 0)',bl);

        fmin:=-1;     defparam('-s','Power supply noise reduction (parameter is the frequency limit in periods per height. Recommended: 10..100',fmin);
        fmax:=0;      defparam('-br','Maximum frequency (in period per height) black level calibration using black reference columns.\\'+
                                     '<0 to disable that. Should be smaller than frequency specified by -s. Default: 0.',fmax);
        nc:=false;    defparam('-nc','Do not crop values less than -2*<readout noise>',nc);
        bayer:=false; defparam('-bp','Bayer pattern image (default: auto-detect)',bayer);

        o1:=false;    defparam('-1o','Offset correction using flatfield (ignored if no flatfield defined)',o1);
        ob:=false;    defparam('-bo','Background correction (ignored if no flatfield defined)',ob);

        om:=0;        defparam('-om','Mask multiplier for overflown pixels (default: 0)',om);
        nn:=false;    defparam('-nn','Disable photon noise estimator.',nn);

        ma.size:=0;   defparam('-m','Add meta info (can be defined multiple times)',ma);        par.deffmt('-m','<key>=<value>[/<comment>]');
        md.size:=0;  defparam('-md','Delete meta info (can be defined multiple times)',md); deffmt('-md','<key>');

        ue:=false;    defparam('-ue','Output image unit is electrons',ue);
        ues:=false;   defparam('-ues','Output image unit is electrons per second',ues);
        us:=false;    defparam('-us','Output image unit is ADU per second',us);

        nr:=false;    defparam('-nr','Do not trim to ROI',nr);

                      defparam('-h','Help',helpf);
                      defparam('-o','Output image',fno); deffmt('-o','<outfile>');
        fno16:='';    defparam('-o16','Write 16 bit output image',fno16); deffmt('-o16','<outfile>');
        run;
        done;
    end;

    cc:=CSImg_Config.create(SImg_Config_Default);
    cc.debug('zocorrect: in='+fni+' out='+fno+' out16='+fno16+' weight='+fnw+' weight16='+fnw16+' bias='+fnb+' fixed_dark='+fn0+' variable_dark='+fnd+' variable_ampglow='+fng+' flat='+fn1+' mask='+fnm+
        ' bl='+real2str(bl)+' om='+real2str(om)+' nc='+boolean2str(nc)+' nn='+boolean2str(nn)+' bayer='+boolean2str(bayer)+' fmin='+real2str(fmin)+' fmax='+real2str(fmax)+
        ' ue='+boolean2str(ue)+' ues='+boolean2str(ues)+' us='+boolean2str(us)+' 1o='+boolean2str(o1)+' nr='+boolean2str(nr));

    i:=0;
    if ue then i+=1;
    if ues then i+=1;
    if us then i+=1;
    if i>1 then cc.error('More than one output units defined');
    o1:=o1 and (fn1<>'');

    { input image }
    fi.pt:=nil;
    mi:=CSImg_MetaInfo.create();
    cc.Info('zocorrect: Reading input image '+fni);
    SImg_ReadImageFile(cc,fni,fi,mi);

    { meta infos }
    for i:=0 to ma.size-1 do
    begin
        str:=mi.add(trim(ma.aa[i]));
        if str<>'' then  cc.warning('zocorrect: Error parsing -m '+ma.aa[i]+': '+str+ '  (expected syntax: -m <key>=<value>[/<comment>])');
    end;
    if not bayer then bayer:=mi.get('BAYERPAT',str,nil);
    if not mi.get('RDNOISE',rn) then rn:=-1;
    if not mi.get('EGAIN',gain) then gain:=-1
    else if gain<1e-10 then gain:=-1;
    adr:=1/gain;  { ADU/e}

    { delete meta info keys }
    for i:=0 to md.size-1 do
        mi.delete(trim(md.aa[i]));

    { initialize square noise with sqr(rn) }
    fsn.pt:=nil;
    if (fnw<>'') or (fnw16<>'') then
    begin
        if rn<-1e-30 then cc.error('zocorrect: Read out noise not defined');
        if gain<1e-10 then cc.error('zocorrect: Gain not defined');
        SImg_New(cc, fsn, fi.nx,fi.ny,fi.nc, false);
        r:=sqr(rn);
        for i:=0 to fsn.size-1 do
            fsn.pt[i]:=r;
    end;

    { dead pixel mask }
    fm.pt:=nil;
    if fnm<>'' then
    begin
        cc.Info('zocorrect: Reading pixel mask '+fnm);
        SImg_ReadImageFile(cc,fnm,fm);
        if fm.size<>fi.size then cc.error('Invalid size of pixel mask');
    end;
    if (fsn.pt<>nil) and (fm.pt=nil) then
    begin
        SImg_New(cc,fm, fi.nx,fi.ny,fi.nc, false);
        for i:=0 to fm.size-1 do
            fm.pt[i]:=1.0;
    end;

    { overflow/underflow }
    if not mi.get('DATAMIN',dmin) then dmin:=-1e300;
    if not mi.get('DATAMAX',dmax) then dmax:=1e300;
    if (fm.pt<>nil) and ((dmin>-1e299) or (dmax<1e299)) then
    begin
        cc.Info('zocorrect: Calculating overflow/underflow mask');
        for i:=0 to fi.size-1 do
        begin
            r:=fi.pt[i];
            if r<dmin then fm.pt[i]:=0
            else if r>dmax then fm.pt[i]*=om;
        end;
     end;

    { bias }
    f0.pt:=nil;
    if fnb<>'' then
    begin
        cc.Info('zocorrect: Reading and subtracting bias image '+fn0);
        SImg_ReadImageFile(cc,fnb,f0);
        if f0.size<>fi.size then cc.error('Invalid size of bias image');
        for i:=0 to fi.size-1 do
            fi.pt[i]-=f0.pt[i]+bl;
        SImg_Free(cc,f0);
    end
    else if abs(bl)>1e-30 then
        for i:=0 to fi.size-1 do
            fi.pt[i]-=bl;

    { fixed dark current }
    f0.pt:=nil;
    if fn0<>'' then
    begin
        cc.Info('zocorrect: Reading and subtracting fixed dark current image '+fn0);
        SImg_ReadImageFile(cc,fn0,f0);
        if f0.size<>fi.size then cc.error('Invalid size of fixed dark current image');
        for i:=0 to fi.size-1 do
        begin
            r:=f0.pt[i];
            fi.pt[i]-=r;
            if (fsn.pt<>nil) and (r>0) then fsn.pt[i]+=r*adr;
        end;
        SImg_Free(cc,f0);
    end;

    { collect data for black reference calibration }
    fmb.pt:=nil;
    bi:=nil;
    if fmax>-simg_mina then
    begin
        if fm.pt<>nil then SImg_CropBRC(cc, fm, fmb, mi, nr);
//      if fmb.pt<>nil then SImg_WriteImageFile(cc, 'zocorrect-fmb.fits,', fmb);
        if SImg_GetBR(cc, fi,fmb, mi, nr, bi)<1 then bi:=nil;
    end;
    if bi=nil then fmax:=-1;

    { trim to ROI }
    if not nr then
    begin
        cc.Info(cc.name+': Cropping to ROI');
        SImg_CropROI(cc, fi, mi);
        if fm.pt<>nil then SImg_CropROI(cc, fm, mi);
        if fsn.pt<>nil then SImg_CropROI(cc, fsn, mi);
    end;

    { variable dark current image }
    f0.pt:=nil; f1.pt:=nil;
    if fnd<>'' then
    begin
        SImg_ReadImageFile(cc,fnd,f0);
        if bi<>nil then SImg_GetBR(cc, f0,fmb, mi, nr, bj);
        if (not nr) and (f0.size>fi.size) then SImg_CropROI(cc, f0, mi);
        if bayer then i:=2
        else i:=1;
        r:=SImg_DarkCurrentSubtraction(cc, fi,f0, fm, fi, i);
        cc.Info('zocorrect: Variable dark current subtraction with factor '+real2str(r,5));
        if fsn.pt<>nil then
        begin
            r*=adr;
            for i:=0 to fsn.size-1 do
                fsn.pt[i]+=abs(f0.pt[i]*r);
        end;
        if bi<>nil then
            for i:=0 to fi.ny*fi.nc-1 do
                bi[i]-=r*bj[i];
        SImg_Free(cc,f0);
        if bi<>nil then freemem2(bj);
    end;

    { variable amp glow image }
    f0.pt:=nil; f1.pt:=nil;
    if fng<>'' then
    begin
        SImg_ReadImageFile(cc,fng,f0);
        if bi<>nil then SImg_GetBR(cc, f0,fmb, mi, nr, bj);
        if (not nr) and (f0.size>fi.size) then SImg_CropROI(cc, f0, mi);
        r:=SImg_BGApprox3(cc, fi,f0,fm, fg, false);
        cc.Info('zocorrect: Amplow subtraction with factor '+real2str(r,5));
        for i:=0 to fi.size-1 do
        begin
            s:=f0.pt[i]*r;
            fi.pt[i]-=s;
            if fsn.pt<>nil then fsn.pt[i]+=abs(s*adr);
        end;
        if bi<>nil then
            for i:=0 to fi.ny*fi.nc-1 do
                bi[i]-=r*bj[i];
        SImg_Free(cc,f0);
        if bi<>nil then freemem2(bj);
    end;

    { black reference calibration }
    if bi<>nil then
    begin
        if fmin<=0.999999 then cc.Info(cc.name+': Black level calibration')
        else cc.Info(cc.name+': Black level calibration and power supply noise reduction up to frequency '+real2str(fmax,-1));
        SImg_BRCorrect(cc, fi, bi, fmax);
        freemem2(bi);
        SImg_Free(cc,fmb);
    end;

    { cropping to small values }
    if (not nc) and (rn>1e-50) then
    begin
        r:=-rn*5;
        cc.Info('zocorrect: Cropping values less than '+real2str(r,5));
        for i:=0 to fi.size-1 do
            if fi.pt[i]<r then fi.pt[i]:=r;
    end;

    { dark current estimation }
    if fsn.pt=nil then dn:=-1
    else begin
        r:=SImg_GetQuantile(cc,fsn,0.5) - sqr(rn);
        if r<0 then r:=0;
        dn:=sqrt(r);
{$ifndef zocorrect_noiseinfo}
        if fn1='' then
{$endif}
        cc.info(cc.name+': Estimated dark current noise: '+real2str(dn)+' ADU = '+real2str(dn*gain)+' e-');
    end;

    { read flatfield (also used for power supply noise reduction) }
    if fn1<>'' then
    begin
        cc.Info('zocorrect: Reading flatfield image '+fn1);
        SImg_ReadImageFile(cc,fn1,f1);
        if (not nr) and (f1.size>fi.size) then SImg_CropROI(cc, f1, mi);
        if f1.size<>fi.size  then cc.error('zocorrect: invalid dimensions: '+fn1);
    end;

    { power supply noise reduction }
    f0.pt:=nil;
    if fmin>-1e-10 then
    begin
        cc.Info('zocorrect: Reduction of power supply noise above frequency '+real2str(fmin,-1));
        SImg_PSDenoise(cc, fi,f0,f1,fm, fi, fmin, false, fmax>0.75*fmin);
    end;

    { image unit: fi must be divided by u }
    u:=1;
    if (ues or us) and (not mi.get('EXPTIME',u)) then cc.warning('zocorrect: Exposure time not defined. Assuming 1s');
    if u<1e-9 then
    begin
        cc.warning('zocorrect: Invalid exposure time. Assuming 1s');
        u:=1;
    end;
    if ue or ues then
    begin
        if gain<1e-10 then cc.warning('zocorrect: Gain not defined, assuming 1 (image unit is electrons)')
        else u*=adr;
    end;
    str:='';
    if ues then str:='electron/s'
    else if ue then str:='electron'''
    else if us then str:='adu/s';
    if str<>'' then mi.add('BUNIT',''''+str+'''','Image unit');

    bn:=-1;
    if fn1<>'' then
    begin
        { flatfield and unit correction }
        cc.Info('zocorrect: Dividing flatfield image '+fn1);
        f0.pt:=nil;
        if o1 then SImg_New(cc, f0, f1.nx,f1.ny,f1.nc, false);

        q:=SImg_GetQuantile(cc, f1, 0.98);
        if q<1e-20 then cc.error('zocorrect: non-positive flatfield');
        q:=1/q;
        v:=simg_mina; s:=simg_mina;
        for i:=0 to fi.size-1 do
        begin
            r:=f1.pt[i]*q;
            if r>1.05 then { this is most likely a hot pixel }
            begin
                if fm.pt<>nil then fm.pt[i]*=sqr(sqr(1.05/r));
                r:=1.05;
            end;
            if r<1e-2 then { this is most likely a dead pixel }
            begin
                if fm.pt<>nil then fm.pt[i]:=0;
                r:=1e-2;
            end;
            f1.pt[i]:=r;

            v+=r; s+=sqr(r);

            r*=u;
            if fm.pt<>nil then fm.pt[i]*=sqr(r);
            r:=1/r;
            fi.pt[i]*=r;
            if f0.pt<>nil then f0.pt[i]:=r;
        end;
        v/=s;   // weighted average amplification (of noise levels) due to flatfield correction

        if o1 or ob or (fsn.pt<>nil) then
        begin
            { background estimation }
            fg.pt:=nil;
            s:=SImg_BGApprox3(cc, fi,f0,fm, fg, true);
            if o1 then cc.Info(cc.name+': Offset corrected by '+real2str(-s)+' ADU = '+real2str(-s*gain)+' e-');

            { offset and background correction }
            if ob and o1 then
                for i:=0 to fi.size-1 do
                    fi.pt[i]-=s*f0.pt[i]+fg.pt[i]
            else if o1 then
                for i:=0 to fi.size-1 do
                    fi.pt[i]-=s*f0.pt[i]
            else if ob then
                for i:=0 to fi.size-1 do
                    fi.pt[i]-=fg.pt[i];

            { photon noise }
{$ifdef zocorrect_noiseinfo}
            r:=sqrt(fg.pt[(fg.ny div 2)*fg.sl + (fg.nx div 2)*fg.nc]*u*adr*v);
            cc.info(cc.name+': Estimated background photon noise in center: '+real2str(r)+' ADU = '+real2str(r*gain)+' e-');
{$endif}
            if (fsn.pt<>nil) and o1 then
            begin
                dmax:=s;
                { adjust noise, make sure that s is plausible }
                if sqr(dn)-s*adr<0 then s:=sqr(dn)/adr;                 { make sure that s is plausible, dn ist defined if fsn.pt<>nil }
                r:=sqr(rn);
                for i:=0 to fi.size-1 do
                begin
                    q:=fsn.pt[i] + (fg.pt[i]*u*f1.pt[i] - s)*adr;       { old image units }
                    fsn.pt[i]:=q;
                    if q<r then r:=q;
                end;
                r-=sqr(rn);
                if r<-simg_mina then            { either s or the gradient where not plausible }
                for i:=0 to fi.size-1 do
                    fsn.pt[i]-=r;
                { adjust dark current noise }
                dn:=sqr(dn)-s*adr;
                if dn<0 then dn:=0;
                dn:=sqrt(dn);
                if abs(dmax-s)>simg_mina then cc.warning(cc.name+': Offset correction used for noise estimation changed to plausible value of '+real2str(-s)+' ADU = '+real2str(-s*gain)+' e-. This may lead to wrong average dark current noise estimation: '+real2str(dn)+' ADU = '+real2str(dn*gain)+' e-');
            end
            else if ob and (gain>simg_mina) and (rn>simg_mina) then     { estimate background noise becuse it cant be estimated anymore after background subtraction }
            begin
                r:=SImg_GetQuantile(cc, fi, 0.48);
                if r<0 then r:=0;
                r+=fg.pt[(fg.ny div 2)*fg.sl + (fg.nx div 2)*fg.nc];    { gradient in center }
                if r<0 then r:=0;
                bn:=sqrt(sqr(rn)+r*adr*u*v)/u;  { in new image units }
            end;
            SImg_Free(cc,fg);
        end;
        SImg_Free(cc,f0);
        SImg_Free(cc,f1);
        dn*=v; rn*=v;
        if dn>0 then cc.Info(cc.name+': Estimated average dark current noise: '+real2str(dn)+' ADU = '+real2str(dn*gain)+' e-');
        if rn>0 then cc.Info(cc.name+': Average read out noise corrected by factor '+real2str(v)+' to '+real2str(rn)+' ADU = '+real2str(rn*gain)+' e-');
    end else
    begin
        if fsn.pt<>nil then
        begin
            r:=sqr(1/u);
            s:=SImg_GetQuantile(cc, fi, 0.48)*adr;      { background photon noise in ADU }
            for i:=0 to fsn.size-1 do
                fsn.pt[i]:=(fsn.pt[i]+s)*r;
        end;
        r:=1/u;
        for i:=0 to fi.size-1 do
            fi.pt[i]*=r;
        v:=u;
    end;
    if rn>-simg_mina then rn/=u;
    if dn>-simg_mina then dn/=u;
    if gain>-simg_mina then gain*=u;

    { combine mask image (fm) and square noise image (fsn), estimation of background noise }
    if (fsn.pt<>nil) and (fm.pt<>nil) then
    begin
        bn:=SImg_GetQuantile(cc,fsn,0.5);       { old image units }
        s:=0.25*bn;
        if bn<0 then bn:=0;
        bn:=sqrt(bn)*v/u;
        cc.Info('zocorrect: Estimated average background noise: '+real2str(bn)+' '+str+' = '+real2str(bn*u)+' ADU = '+real2str(bn*gain)+' e-');
        if s<simg_mina then s:=simg_mina;
        v:=1/s;
        for i:=0 to fm.size-1 do
        begin
            r:=fsn.pt[i];
            if r>s then fm.pt[i]/=r
            else fm.pt[i]*=v;
        end;
        SImg_Free(cc,fsn);
    end;

    { output image }
    if gain>1e-10 then mi.add('EGAIN',real2str(gain,12),'ADC resolution in electrons per image unit')
    else mi.delete('EGAIN');
    if bn>=0 then mi.add('BGNOISE',real2str(bn,15),'Background noise in image units')
    else mi.delete('BGNOISE');
    if rn>=0 then mi.add('RDNOISE',real2str(rn,15),'Readout noise in image units')
    else mi.delete('RDNOISE');
    if dn>=0 then mi.add('DCNOISE',real2str(dn,15),'Dark current noise in image units')
    else mi.delete('DCNOISE');
    mi.delete('DATAMIN');
    mi.delete('DATAMAX');
    if not nr then
    begin
        mi.delete('ROI_O1');
        mi.delete('ROI_O2');
        mi.delete('ROI_S1');
        mi.delete('ROI_S2');
    end;
    cc.Info('zocorrect: Writing output image '+fno);
    if (fno<>'') or (fno16='') then SImg_WriteImageFile(cc,fno,fi,mi);
    if fno16<>'' then
    begin
        ri.pt:=nil;
        if fm.pt=nil then cc.warning(cc.name+': 16 bit integer output only supported in combination with mask output, writing floating point image')
        else if not mi.get('RDNOISE',rn) then cc.warning(cc.name+': read out noise not defined, writing floating point image')
        else begin
            dmin:=-16*rn;
            dmax:=32760*rn;
            s:=dmax; u:=dmin; j:=0;
            for i:=0 to fi.size-1 do
            begin
                r:=fi.pt[i];
                if fm.pt[i]<simg_mina then j+=1
                else if (r>=dmin) and (r<=dmax) then
                begin
                    if r<s then s:=r;
                    if r>u then u:=r;
                    j+=1;
                end;
            end;
            if j<0.999*fi.size then cc.warning(cc.name+': to much large or small values, writing floating point image')
            else begin
                SImg_New(cc, ri, SImg_RawWord, fi.nx,fi.ny,fi.nc, false);
                u:=(u-s)/65535;
                if u<simg_mina then u:=simg_mina;
                ri.co:=s; ri.cf:=u;
                u:=1/u;
                for i:=0 to ri.size-1 do
                begin
                    j:=round((fi.pt[i]-s)*u);
                    if j<0 then
                    begin
                        j:=0;
                        fm.pt[i]:=0;
                    end;
                    if j>65535 then
                    begin
                        j:=65535;
                        fm.pt[i]*=om;
                    end;
                    ri.ptw[i]:=j;
                end;
            end;
        end;
        if ri.pt=nil then SImg_WriteImageFile(cc,fno16,fi,mi)
        else begin
            SImg_WriteImageFile(cc,fno16,ri,mi);
            SImg_Free(cc,ri);
        end;
    end;
    SImg_Free(cc,fi);

    { output weight image }
    if (fnw<>'') or (fnw16<>'') then
    begin
        cc.Info('zocorrect: Writing pixel weight image '+fnw);
        mi.delete('BGNOISE');
        mi.delete('RDNOISE');
        mi.delete('EGAIN');
        mi.delete('BUNIT');
        if fnw<>'' then SImg_WriteImageFile(cc,fnw,fm,mi);
        if fnw16<>'' then
        begin
            ri.pt:=nil;
            SImg_Convert(cc, fm,ri, true,false);
            SImg_Convert(cc, ri,ri, SImg_RawWord, nil, false,false);
            SImg_WriteImageFile(cc,fnw16,ri,mi);
            SImg_Free(cc, ri);
        end
        else SImg_Free(cc,fm);
    end
    else SImg_Free(cc,fm);

    cc.destroy;
end.
