{  simg -- image processing library
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{*************************************************************************}
{******* SImg_MinMax *****************************************************}
{*************************************************************************}
procedure SImg_MinMax(cc:CSImg_Config; mi:boolean; src1,src2:TSImg_FPImage; var dest:TSImg_FPImage; f1,f2:double);
var fi      : TSImg_FPImage;
    i       : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_MinMax: min='+boolean2str(mi) + ' dest= '+SImg_InfoStr2(dest)+' src1='+SImg_InfoStr2(src1)+' src2='+SImg_InfoStr2(src2)+
       ' f1='+real2str(f1)+' f2='+real2str(f2));
    if not SImg_Recover(cc,src1) then cc.Error('SImg_MinMax: Invalid image: src1');
    if not SImg_Recover(cc,src2) then cc.Error('SImg_MinMax: Invalid image: src2');
    if (src1.nx<>src2.nx) or (src1.ny<>src2.ny) or (src1.nc<>src2.nc) then cc.Error('SImg_MinMax: Images must have the same size');
    if (not SImg_OverwriteAlert(dest,src1)) and (not SImg_OverwriteAlert(dest,src2)) then SImg_Free(cc,dest);
    fi.pt:=nil;
    SImg_Duplicate(cc,src1,fi,false);

    if mi
        then for i:=0 to fi.size-1 do
            fi.pt[i]:=min(f1*src1.pt[i], f2*src2.pt[i])
    else for i:=0 to fi.size-1 do
        fi.pt[i]:=max(f1*src1.pt[i], f2*src2.pt[i]);

    SImg_Move(cc,fi,dest);
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_Mul ********************************************************}
{*************************************************************************}
procedure SImg_Mul(cc:CSImg_Config; src1,src2:TSImg_FPImage; var dest:TSImg_FPImage; e1,e2:double);
var fi      : TSImg_FPImage;
    i       : longint;
    d,e,f   : ^simgfp;
    r,s,t   : double;
    b1,b2   : boolean;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_Mul: dest= '+SImg_InfoStr2(dest)+' src1='+SImg_InfoStr2(src1)+' src2='+SImg_InfoStr2(src2)+
        ' e1='+real2str(e1)+' e2='+real2str(e2));
    if not SImg_Recover(cc,src1) then cc.Error('SImg_Mul: Invalid image: src1');
    if not SImg_Recover(cc,src2) then cc.Error('SImg_Mul: Invalid image: src2');
    if (src1.nx<>src2.nx) or (src1.ny<>src2.ny) or (src1.nc<>src2.nc) then cc.Error('SImg_Mul: Images must have the same size');
    if (not SImg_OverwriteAlert(dest,src1)) and (not SImg_OverwriteAlert(dest,src2)) then SImg_Free(cc,dest);
    fi.pt:=nil;
    SImg_Duplicate(cc,src1,fi,false);

    e1-=1; e2-=1;
    b1:=abs(e1)>1e-5;
    b2:=abs(e2)>1e-5;
    d:=src1.pt; e:=src2.pt; f:=fi.pt;
    for i:=0 to fi.size-1 do
    begin
        s:=d^; t:=e^;
        r:=s*t;
        if b1 then
        begin
            s:=abs(s);
            if s<simg_mina then s:=simg_mina;
            r*=exp(ln(s)*e1);
        end;
        if b2 then
        begin
            t:=abs(t);
            if t<simg_mina then t:=simg_mina;
            r*=exp(ln(t)*e2);
        end;
    f^:=r;
    d+=1; e+=1; f+=1;
    end;
    SImg_Move(cc,fi,dest);
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_Power ******************************************************}
{*************************************************************************}
procedure SImg_Power(cc:CSImg_Config; src:TSImg_FPImage; var dest:TSImg_FPImage; ex:double);
var fi    : TSImg_FPImage;
    i     : longint;
    d     : ^simgfp;
    r,s   : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_Power: dest= '+SImg_InfoStr2(dest)+' src='+SImg_InfoStr2(src)+' ex='+real2str(ex));
    if not SImg_Recover(cc,src) then cc.Error('SImg_Power: invalid image: src');
    if not SImg_OverwriteAlert(dest,src) then SImg_Free(cc,dest);
    fi.pt:=nil;
    SImg_Duplicate(cc,src,fi,true);

    ex-=1;
    if abs(ex)>1e-5 then
    begin
        d:=fi.pt;
        for i:=0 to fi.size-1 do
        begin
            r:=d^;
            s:=abs(r);
            if s<simg_mina then s:=simg_mina;
            d^:=r*exp(ln(s)*ex);
            d+=1;
        end;
    end;
SImg_Move(cc,fi,dest);
SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_FlipC ******************************************************}
{*************************************************************************}
procedure SImg_FlipC(cc:CSImg_Config; fi:TSImg_FPImage);
var i,j,k,l,m : longint;
    d,e       : ^simgfp;
    r         : simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_FlipC: fi= '+SImg_InfoStr2(fi));
    if not SImg_Recover(cc,fi) then cc.Error('SImg_FlipC: invalid image: fi');

    k:=fi.nc-1;
    l:=(fi.nc shr 1)-1; m:=fi.nc-l-1;
    d:=fi.pt;
    for i:=0 to fi.nx*fi.ny-1 do
    begin
        e:=@d[k];
        for j:=0 to l do
        begin
            r:=d^;  d^:=e^;  e^:=r;
            d+=1; e-=1;
        end;
        d+=m;
    end;
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_RotateFlip *************************************************}
{*************************************************************************}
{ fi    : input
  fo    : output
  rot   : CW rotation angle in degree. Rounded to 90° steps.
  flipx : flip in X direction (before rotation)
  flipy : flip in Y direction (before rotation)
}
procedure SImg_RotateFlip(cc:CSImg_Config; fi:TSImg_FPImage; var fo:TSImg_FPImage; rot:longint; flipx,flipy:boolean);
var fj                  : TSImg_FPImage;
    a11,a12,a21,a22     : longint;
    ax,ay,x,y,c,de      : longint;
    d,e                 : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_RotateFlip: fi= '+SImg_InfoStr2(fi)+' fo= '+SImg_InfoStr2(fo)+' rot='+int2str(rot)+' flipx='+boolean2str(flipx)+' flipy='+boolean2str(flipy));
    if not SImg_Recover(cc,fi) then cc.Error('SImg_RotateFlip: invalid image: fi');
    if not SImg_OverwriteAlert(fi,fo) then SImg_Free(cc,fo);

    { transformation matrix }
    rot:=round(rot/90) and 3;
    a11:= round(cos(rot*pi/2));
    a12:=-round(sin(rot*pi/2));
    a21:=-a12;
    a22:= a11;
    if flipx then
    begin
        a11:=-a11;
        a21:=-a21;
    end;
    if flipy then
    begin
        a12:=-a12;
        a22:=-a22;
    end;

    { transformation matrix }
    fj.pt:=nil;
    if rot and 1=0 then SImg_New(cc, fj, fi.nx,fi.ny,fi.nc, false)
    else SImg_New(cc, fj, fi.ny,fi.nx,fi.nc, false);
    ax:=select(a11+a12<0, fj.nx-1, 0);
    ay:=select(a21+a22<0, fj.ny-1, 0);
    de:=a21*fj.sl + a11*fj.nc;
{    writeln(stderr, a11,'   ',a12,'      ',ax);
    writeln(stderr, a21,'   ',a22,'      ',ay);}
    for y:=0 to fi.ny-1 do
    begin
        d:=fi.pt + y*fi.sl - 1;
        e:=fj.pt + (ay+a22*y)*fj.sl + (ax+a12*y)*fj.nc - 1;
        for x:=0 to fi.nx-1 do
        begin
            for c:=1 to fi.nc do
                e[c]:=d[c];
            d+=fi.nc;
            e+=de;
        end;
    end;
    SImg_Move(cc, fj,fo);
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_Cut ********************************************************}
{*************************************************************************}
{ cut an image
   x1,x2,y1,y2,c1,c2  :
         * area to cut
         * if values are <0 or >=fi.n[xyc], a border with background value bg added
         * if x1>x2, y1>y2, c1>c2, image is flipped
   bg  : background value
}
procedure SImg_Cut(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; x1,y1,x2,y2,c1,c2,rot:longint; bg:simgfp);
var fk            : TSImg_FPImage;
    y,ya,yb,yo    : longint;
    x,xa,xb,xo,xr : longint;
    c,ca,cb,co,cr : longint;
    d,e           : ^simgfp;
    ix,iy,ic      : boolean;
begin
    if cc=nil then cc:=SImg_Config_default;
    SImg_IncDebugLevel;
    cc.debug('SImg_Cut: called with: source='+SImg_InfoStr2(fi)+' dest='+SImg_InfoStr2(fj)+' x1='+int2str(x1)+' y1='+int2str(y1)+' x2='+int2str(x2)+' y2='+int2str(y2)+' c1='+int2str(c1)+' c2='+int2str(c2)+' rot='+int2str(rot)+' bg='+real2str(bg));
    if not SImg_Recover(cc,fi) then cc.error('SImg_Cut: Invalid image');
    if (x1=0) and (y1=0) and (x2=fi.nx-1) and (y2=fi.ny-1) and (c1=0) and (c2=fi.nc-1) then
    begin
        SImg_Duplicate(cc, fi,fj, true);
        SImg_DecDebugLevel;
        exit;
    end;

    x:=abs(x2-x1)+1; if (x>maxxsize) or (x<minxsize) then cc.error('SImg_Cut: Invalid X size');
    y:=abs(y2-y1)+1; if (y>maxysize) or (y<minysize) then cc.error('SImg_Cut: Invalid Y size');
    c:=abs(c2-c1)+1; if (c>maxcnsize) or (c<mincnsize) then cc.error('SImg_Cut: Invalid number of colors');
    if not SImg_OverwriteAlert(fi,fj) then SImg_Free(cc,fj);

    ix:=x1>x2;
    if ix then
    begin
        x:=x1;
        x1:=x2;
        x2:=x;
    end;
    iy:=y1>y2;
    if iy then
    begin
        y:=y1;
        y1:=y2;
        y2:=y;
    end;
    ic:=c1>c2;
    if ic then
    begin
        c:=c1;
        c1:=c2;
        c2:=c;
    end;

    fk.pt:=nil;
    SImg_New(cc,fk, x2-x1+1, y2-y1+1, c2-c1+1);

    ya:=max(0,-y1); yb:=max(ya-1, min(y2-y1,fi.ny-1-y1));
    xa:=max(0,-x1); xb:=max(xa-1, min(x2-x1,fi.nx-1-x1));
    ca:=max(0,-c1); cb:=max(ca-1, min(c2-c1,fi.nc-1-c1));
    yo:=max(y1,0)*fi.sl; {yr:=max(fi.ny-y2-1,0)*fi.sl;}
    xo:=max(x1,0)*fi.nc; xr:=max(fi.nx-x2-1,0)*fi.nc;
    co:=max(c1,0);       cr:=max(fi.nc-c2-1,0);
    d:=fk.pt; e:=fi.pt;
    for y:=0 to ya*fk.sl-1 do
    begin
        d^:=bg; d+=1;
    end;
//    writeln(stderr, ya,'  ',yb,'    ',xa,'  ',xb,'    ',ca,'  ',cb);
//    writeln(stderr, yo,'    ',xo,'  ',xr,'    ',co,'  ',cr);
    e+=yo;
    for y:=ya to yb do
    begin
        for x:=0 to xa*fk.nc-1 do
        begin
            d^:=bg; d+=1;
        end;
        e+=xo;
        for x:=xa to xb do
        begin
            for c:=0 to ca-1 do
            begin
                d^:=bg; d+=1;
            end;
            e+=co;
            for c:=ca to cb do
            begin
                d^:=e^;
                d+=1; e+=1;
            end;
            for c:=cb+1 to fk.nc-1 do
            begin
                d^:=bg; d+=1;
            end;
            e+=cr;
        end;
        for x:=(xb+1)*fk.nc to fk.sl-1 do
        begin
            d^:=bg; d+=1;
        end;
        e+=xr;
    end;
    for y:=(yb+1)*fk.sl to fk.size-1 do
    begin
        d^:=bg; d+=1;
    end;

    SImg_Move(cc,fk,fj);

    if ix or iy or (round(rot/90) and 3<>0) then SImg_RotateFlip(cc, fj,fj, rot,ix,iy);
    if ic then SImg_FlipC(cc,fj);

    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_CutSimple **************************************************}
{*************************************************************************}
{ simple cut without extension and color cutting }
procedure SImg_CutSimple(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; x0,y0,width,height:longint);
var fk  : TSImg_FPImage;
    i,j : longint;
    d   : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_default;
    SImg_IncDebugLevel;
    cc.debug('SImg_Cut_simple: called with: source='+SImg_InfoStr2(fi)+' dest='+SImg_InfoStr2(fj)+' x0='+int2str(x0)+' y0='+int2str(y0)+' width='+int2str(width)+' hight='+int2str(height));
    if not SImg_Recover(cc,fi) then cc.error('SImg_Cut_simple: Invalid image');
    if not SImg_OverwriteAlert(fi,fj) then SImg_Free(cc,fj);
    if x0<0 then x0:=0;
    if y0<0 then y0:=0;
    if width>fi.nx-x0 then width:=fi.nx-x0;
    if height>fi.ny-y0 then height:=fi.ny-y0;
    SImg_New(cc, fk, width,height,fi.nc, true);

    j:=fk.sl*sizeof(simgfp);
    d:=fi.pt + y0*fi.sl + x0*fi.nc;
    for i:=0 to height-1 do
        move( d[i*fi.sl], fk.pt[i*fk.sl], j );

    SImg_Move(cc, fk,fj);
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_FindCenter *************************************************}
{*************************************************************************}
procedure SImg_FindCenter(cc:CSImg_Config; fi:TSImg_FPImage; var xc,yc:double);
var x,y,c     : longint;
    d         : ^simgfp;
    xw,yw,w   : double;
    q,r,s     : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_FindCenter: img= '+SImg_InfoStr2(fi));
    if not SImg_Recover(cc,fi) then cc.error('SImg_FindCenter: Invalid image');

    xw:=fi.pt[0];
    d:=fi.pt;
    for x:=0 to fi.size-1 do
    begin
        r:=d^; d+=1;
        if r>xw then xw:=r;
    end;
    yw:=SImg_GetQuantile(cc,fi,0.9);
    q:=0.7*yw + 0.3*xw;

    xw:=0; yw:=0; w:=0;
    d:=fi.pt;
    for y:=0 to fi.ny-1 do
        for x:=0 to fi.nx-1 do
        begin
            r:=0;
            for c:=0 to fi.nc-1 do
            begin
                s:=(d^-q); d+=1;
                if s>0 then r+=s;
            end;
            xw+=x*r;
            yw+=y*r;
            w+=r;
        end;
    if abs(w)>1e-100 then
    begin
        xc:=xw/w;
        yc:=yw/w;
    end
    else begin
        xc:=fi.nx/2;
        yc:=fi.ny/2;
    end;
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_FindMaximum ************************************************}
{*************************************************************************}
procedure SImg_FindMaximum(cc:CSImg_Config; fi:TSImg_FPImage; var xc,yc:double);
var x,y,c  : longint;
    d      : ^simgfp;
    r      : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_FindMaximum: img= '+SImg_InfoStr2(fi));
    if not SImg_Recover(cc,fi) then cc.error('SImg_FindMaximum: Invalid image');

    d:=fi.pt;
    xc:=0; yc:=0; r:=d^;
    for y:=0 to fi.ny-1 do
        for x:=0 to fi.nx-1 do
            for c:=0 to fi.nc-1 do
            begin
                if d^>r then
                begin
                    r:=d^;
                    xc:=x;
                    yc:=y;
                end;
                d+=1;
            end;
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_AddColors **************************************************}
{*************************************************************************}
procedure SImg_AddColors(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage);
var fk    : TSImg_FPImage;
    d,e   : ^simgfp;
    i,c   : longint;
    r     : double;
begin
if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_AddColors: fi='+SImg_InfoStr2(fi)+' fj='+SImg_InfoStr2(fj));
    if not SImg_Recover(cc,fi) then cc.error('SImg_AddColors: Invalid image: fi');
    if not SImg_OverwriteAlert(fi,fj) then SImg_Free(cc,fj);
    fk.pt:=nil; SImg_New(cc,fk,fi.nx,fi.ny,1);

    d:=fk.pt; e:=fi.pt;
    for i:=0 to fk.size-1 do
    begin
        r:=0;
        for c:=0 to fi.nc-1 do
        begin
            r+=e^;
            e+=1;
        end;
        d^:=r; d+=1;
    end;
    SImg_Move(cc,fk,fj);
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_MaxColor ***************************************************}
{*************************************************************************}
procedure SImg_MaxColor(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage); { get brightes color }
var fk    : TSImg_FPImage;
    e     : ^simgfp;
    i,c   : longint;
    a,b   : simgfp;
begin
if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_MaxColor: fi='+SImg_InfoStr2(fi)+' fj='+SImg_InfoStr2(fj));
    if not SImg_Recover(cc,fi) then cc.error('SImg_MaxColor: Invalid image: fi');
    if not SImg_OverwriteAlert(fi,fj) then SImg_Free(cc,fj);
    fk.pt:=nil; SImg_New(cc,fk,fi.nx,fi.ny,1);

    e:=fi.pt;
    for i:=0 to fk.size-1 do
    begin
        a:=e^;
        for c:=1 to fi.nc-1 do
        begin
            b:=e[c];
            if b>a then a:=b;
        end;
        e+=fi.nc;
        fk.pt[i]:=a;
    end;
    SImg_Move(cc,fk,fj);
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_CartesianToPolar *******************************************}
{*************************************************************************}
{ fi     : input
  fj     : output
  xc,yc  : center
  dr,da  : radius/angle step
  a1,a2  : start/end angle (in degree) }
procedure SImg_CartesianToPolar(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; xc,yc, dr,da,a1,a2:double);
var i,j,c,x,y  : longint;
    d,e        : ^simgfp;
    fk         : TSImg_FPImage;
    ca,sa,r    : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_CartesianToPolar: fi='+SImg_InfoStr2(fi)+' fj='+SImg_InfoStr2(fj)+
        ' xc='+real2str(xc)+' yc='+real2str(yc)+
        ' dr='+real2str(dr)+' da='+real2str(da)+' a1='+real2str(a1)+' a2='+real2str(a2));
    if not SImg_Recover(cc,fi) then cc.error('SImg_CartesianToPolar: Invalid image: fi');
    if (a2-a1)/da+0.49<2 then cc.error('SImg_CartesianToPolar: Invalid angles, parameter da to large or wrong signed');
    if dr<0.01 then cc.error('SImg_CartesianToPolar: Parameter dr to small');
    if xc>fi.nx-1-xc then r:=sqr(xc+1)
    else r:=sqr(fi.nx-xc);
    if yc>fi.ny-1-yc then r+=sqr(yc+1)
    else r+=sqr(fi.ny-yc);
    fk.pt:=nil; SImg_New(cc,fk, round(sqrt(r)/dr), round((a2-a1)/da+1), fi.nc);
    r:=pi/180;
    a1*=r; a2*=r; da*=r;

    e:=fk.pt;
    for i:=0 to fk.ny-1 do
    begin
        ca:=cos(i*da+a1)*dr;
        sa:=sin(i*da+a1)*dr;
        for j:=0 to fk.nx-1 do
        begin
            x:=round(ca*j+xc);
            y:=round(sa*j+yc);
            if (x>=0) and (x<fi.nx) and (y>=0) and (y<fi.ny) then
            begin
                d:=fi.pt + y*fi.sl + x*fi.nc;
                for c:=0 to fi.nc-1 do
                    e[c]:=d[c];
            end
            else for c:=0 to fi.nc-1 do
                    e[c]:=0;
            e+=fi.nc;
        end;
    end;
    SImg_Move(cc,fk,fj);
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_PolarToCartesian *******************************************}
{*************************************************************************}
{ fi     : input
  fj     : output
  nx,ny  : center
  xc,yc  : center
  dr,da  : radius/angle step
  a1     : start angle (in degree) }
procedure SImg_PolarToCartesian(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; nx,ny:longint; xc,yc, dr,da,a1:double);
var i,j,c,x,y   : longint;
    d,e         : ^simgfp;
    fk          : TSImg_FPImage;
    a,xr,yr     : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_PolarToCartesian: fi='+SImg_InfoStr2(fi)+' fj='+SImg_InfoStr2(fj)+
        ' nx='+int2str(nx)+' ny='+int2str(ny)+' xc='+real2str(xc)+' yc='+real2str(yc)+
        ' dr='+real2str(dr)+' da='+real2str(da)+' a1='+real2str(a1));
    if dr<0.01 then cc.error('SImg_PolarToCartesian: Parameter dr to small');
    fk.pt:=nil; SImg_New(cc,fk, nx,ny, fi.nc);
    a:=pi/180;
    a1*=a; da:=1/(da*a);
    dr:=1/dr;

    d:=fk.pt;
    for y:=0 to ny-1 do
    begin
        yr:=y-yc;
        for x:=0 to nx-1 do
        begin
            xr:=x-xc;
            if yr>simg_mina then a:=pi*0.5+arctan(-xr/yr)
            else if yr<-simg_mina then a:=pi*1.5+arctan(-xr/yr)
            else if xr<0 then a:=pi
            else a:=0;
            i:=round((a-a1)*da);
            j:=round(sqrt(sqr(xr)+sqr(yr))*dr);
            if (j>=0) and (j<fi.nx) and (i>=0) and (i<fi.ny) then
            begin
                e:=fi.pt + i*fi.sl + j*fi.nc;
                for c:=0 to fi.nc-1 do
                    d[c]:=e[c];
            end
            else for c:=0 to fi.nc-1 do
                d[c]:=0;
            d+=fi.nc;
        end;
    end;
    SImg_Move(cc,fk,fj);
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_DeInterlace ************************************************}
{*************************************************************************}
{ fi        : input
  fj1,fj2   : output
  threshold : interpolate, if deviation is larger then threshold
  swap      : swap half frames }
procedure SImg_DeInterlace(cc:CSImg_Config; fi:TSImg_FPImage; var fj1,fj2:TSImg_FPImage; threshold:double; swap:boolean);
var fk      : array[0..1] of TSImg_FPImage;
    e       : array[0..1] of ^simgfp;
    x,y,c   : longint;
    h1,h2   : longint;
    r       : simgfp;
    s,t     : double;
    d       : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_DeInterlace: fi='+SImg_InfoStr2(fi)+' fj1='+SImg_InfoStr2(fj1)+' fj2='+SImg_InfoStr2(fj2)+' threshold='+real2str(threshold)+' swap='+boolean2str(swap));
    if not SImg_Recover(cc,fi) then cc.error('SImg_Deinterlace: Invalid image: fi');
    if not SImg_OverwriteAlert(fi,fj1) then SImg_Free(cc,fj1);
    if not SImg_OverwriteAlert(fi,fj2) then SImg_Free(cc,fj2);

    s:=0;
    t:=0;
    d:=fi.pt;
    for c:=0 to fi.size-1 do
    begin
        s+=d[c];
        t+=sqr(d[c]);
    end;
    s:=t/fi.size-sqr(s/fi.size);
    //writeln(stderr, s);
    if s<1e-20 then s:=1e-10
    else s:=sqrt(s);
    threshold:=abs(threshold)*s;
    cc.info('Abs. threshold='+real2str(threshold));

    fk[0].pt:=nil;  SImg_New(cc,fk[0],fi.nx,fi.ny,fi.nc);
    fk[1].pt:=nil;  SImg_New(cc,fk[1],fi.nx,fi.ny,fi.nc);
    d:=fi.pt;
    e[0]:=fk[0].pt; e[1]:=fk[1].pt;
    for y:=0 to fi.ny-1 do
    begin
        if (y=0) or (y=fi.ny-1) then
        begin
            move(d^, e[0]^, fi.sl*sizeof(simgfp));
            move(d^, e[1]^, fi.sl*sizeof(simgfp));
            d+=fi.sl;
            e[0]+=fi.sl;
            e[1]+=fi.sl;
        end else
        begin
            h1:=y and 1;
            h2:=(y+1) and 1;
            for x:=0 to fi.nx-1 do
                for c:=0 to fi.nc-1 do
                begin
                    e[h1]^:=d^;
                    r:=0.5*(d[-fi.sl]+d[fi.sl]);
                    if abs(d^-r)>threshold then e[h2]^:=r
                    else e[h2]^:=d^;
                    d+=1;
                    e[0]+=1;
                    e[1]+=1;
                end;
        end;
    end;

    if swap then
    begin
        SImg_Move(cc,fk[0],fj2);
        SImg_Move(cc,fk[1],fj1);
    end else
    begin
        SImg_Move(cc,fk[0],fj1);
        SImg_Move(cc,fk[1],fj2);
    end;
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_WeightsToNoise *********************************************}
{*************************************************************************}
procedure SImg_WeightToNoise(cc:CSImg_Config; src:TSImg_FPImage; var dest:TSImg_FPImage);
var fi      : TSImg_FPImage;
    i       : longint;
    r       : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_WeightToNoise: dest= '+SImg_InfoStr2(dest)+' src='+SImg_InfoStr2(src));
    if not SImg_Recover(cc,src) then cc.Error('SImg_WeightToNoise: Invalid image: src');
    if not SImg_OverwriteAlert(dest,src) then SImg_Duplicate(cc,src,dest,false);
    fi:=dest;
    for i:=0 to fi.size-1 do
    begin
        r:=src.pt[i];
        if r<simg_mina then r:=simg_mina;
        fi.pt[i]:=sqrt(1/r);
    end;
    SImg_DecDebugLevel;
end;

{*************************************************************************}
{******* SImg_WeightsToSNoise ********************************************}
{*************************************************************************}
procedure SImg_WeightToSNoise(cc:CSImg_Config; src:TSImg_FPImage; var dest:TSImg_FPImage);
var fi      : TSImg_FPImage;
    i       : longint;
    r       : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_WeightToSNoise: dest= '+SImg_InfoStr2(dest)+' src='+SImg_InfoStr2(src));
    if not SImg_Recover(cc,src) then cc.Error('SImg_WeightToSNoise: Invalid image: src');
    if not SImg_OverwriteAlert(dest,src) then SImg_Duplicate(cc,src,dest,false);
    fi:=dest;
    for i:=0 to fi.size-1 do
    begin
        r:=src.pt[i];
        if r<simg_mina then r:=simg_mina;
        fi.pt[i]:=1/r;
    end;
    SImg_DecDebugLevel;
end;
