{  simg -- image processing library
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{*************************************************************************}
{******* SImg_GetStatInfo ************************************************}
{*************************************************************************}
procedure SImg_GetStatInfo(cc:CSImg_Config; fi:TSImg_FPImage; var si:TSImg_StatInfo);
var i        : longint;
    r,s,t,u  : double;
    d        : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_default;
    if not SImg_Recover(cc,fi) then cc.error('SImg_GetStatInfo: illegal image');
    d:=fi.pt;
    s:=d^; t:=s;
    u:=0;
    for i:=0 to fi.size-1 do
    begin
        r:=d^; d+=1;
        if r<s then s:=r;
        if r>t then t:=r;
        u+=r;
    end;
    si.minv:=s; si.maxv:=t;
    u/=fi.size; si.av:=u;

    d:=fi.pt;
    s:=0; t:=0;
    for i:=0 to fi.size-1 do
    begin
        r:=d^-u; d+=1;
        s+=abs(r);
        t+=sqr(r);
    end;
    s/=fi.size; si.md:=s;
    t/=fi.size; si.rms:=sqrt(t);
end;


{*************************************************************************}
{******* SImg_GetMinMax **************************************************}
{*************************************************************************}
{
  Get minval and maxval.
  For 8 bit images these are 0 and 255.
  For 16 bit images these are 0 and 65536.
  cf anf co are ignored.
 }
procedure SImg_GetMinMax(cc:CSImg_Config; ri:TSImg_RDImage; var minv,maxv:double);
@rawptr d;
var i       : longint;
    r,s,t   : double;
begin
    if cc=nil then cc:=SImg_Config_default;
    if not SImg_Recover(cc,ri) then cc.error('SImg_GetMinMax: illegal image');
    case ri.fmt of
        SImg_RawByte :
            begin
                minv:=0;
                maxv:=255;
                exit;
            end;
        SImg_RawWord :
            begin
                minv:=0;
                maxv:=65535;
                exit;
            end;
    end;

    if not ri.bole then SImg_SwapBytes(ri);
    d:=ri.pt;
    r:=0; s:=0; // makes the compiler happy
    case ri.fmt of
        SImg_RawSingle :
            begin
                r:=ds^; s:=r+abs(r)*1e-15;
                for i:=0 to ri.size-1 do
                begin
                    t:=ds^;
                    if t<r then r:=t;
                    if t>s then s:=t;
                    ds+=1;
                end;
            end;
        SImg_RawDouble :
            begin
                r:=dd^; s:=r+abs(r)*1e-15;
                for i:=0 to ri.size-1 do
                begin
                    t:=dd^;
                    if t<r then r:=t;
                    if t>s then s:=t;
                    dd+=1;
                end;
            end;
        else cc.error('SImg_GetMinMax: illegal raw format');
    end;
    minv:=r;
    maxv:=s;
end;

procedure SImg_GetMinMax(cc:CSImg_Config; fi:TSImg_FPImage; var min,max:double);
var i        : longint;
    r,s,t,u  : double;
    d        : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_default;
    if not SImg_Recover(cc,fi) then cc.error('SImg_GetMinMax: illegal image');
    d:=fi.pt;
    s:=d^; t:=s;
    u:=0;
    for i:=0 to fi.size-1 do
    begin
        r:=d^; d+=1;
        if r<s then s:=r;
        if r>t then t:=r;
        u+=r;
    end;
    min:=s; max:=t;
end;


{*************************************************************************}
{******* SImg_GetSharpness1 **********************************************}
{*************************************************************************}
{ edge detect algorithm, best choice if physical resolution is higher than optical resolution
  thr: threshold; feasible: 0..1; 0.999..1 for stars
  returned value is the geometric mean of single color values }
function SImg_GetSharpness1(cc:CSImg_Config; fi:TSImg_FPImage; thr:double; var sharpness:array of double):double;
var r,v,u11,u12,u21,u22   : double;
    v11,v12,v21,v22       : double;
    buf                   : ^double;
    cnt                   : ^longint;
    d0,d1,d2,d3           : ^simgfp;
    x,y,c,x1,x2,x3        : longint;

begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_GetSharpness1: fi='+SImg_InfoStr2(fi)+'  thr='+real2str(thr));
    if not SImg_Recover(cc,fi) then cc.error('SImg_GetSharpness1: Invalid image: fi');
    getmem(buf,fi.nc*2*sizeof(double));
    getmem(cnt,fi.nc*sizeof(longint));

    { get max. number of saturated pixels for threshold limitation }
    y:=round(sqrt(fi.nx*fi.ny*1e-4));
    if fi.nc>1 then
    begin
        for c:=0 to fi.nc-1 do
        begin
            buf[c*2]:=1e-100;
            buf[c*2+1]:=0;
            cnt[c]:=10; // regularization
        end;

        d0:=fi.pt;
        for x:=0 to fi.nx*fi.ny-1 do
            for c:=0 to fi.nc-1 do
            begin
                r:=d0^; d0+=1;
                if r>buf[c*2] then buf[c*2]:=r;
                buf[c*2+1]+=r;
            end;

        for c:=0 to fi.nc-1 do
            buf[c]:=0.1*buf[c*2+1]/(fi.nx*fi.ny)+0.9*buf[c*2];

        d0:=fi.pt;
        for x:=0 to fi.nx*fi.ny-1 do
        begin
            for c:=0 to fi.nc-1 do
                if d0[c]>buf[c] then cnt[c]+=1;
            d0+=fi.nc;
        end;

        for c:=0 to fi.nc-1 do
            if cnt[c]>y then y:=cnt[c];
    end;
    r:=1-y/(fi.nx*fi.ny);
    if thr>r then thr:=r;

    for c:=0 to fi.nc-1 do
    begin
        v:=SImg_GetQuantile(cc,fi,c,thr);
        if thr<0.05 then r:=v
                    else r:=SImg_GetQuantile(cc,fi,c,0.05);
        buf[c*2]:=v+3*r;
        buf[c*2+1]:=0;
        cnt[c]:=1;
    end;

    x1:=fi.nc;
    x2:=fi.nc*2;
    x3:=fi.nc*3;
    for y:=0 to fi.ny-4 do
    begin
        d0:=fi.pt + y*fi.sl;
        d1:=d0 + fi.sl;
        d2:=d0 + 2*fi.sl;
        d3:=d0 + 3*fi.sl;
        for x:=0 to fi.nx-4 do
        begin
            for c:=0 to fi.nc-1 do
            begin
                v:=buf[c*2];
                u11:=d1[x1];
                u12:=d1[x2];
                u21:=d2[x1];
                u22:=d2[x2];
                v11:=d0[ 0] + d0[x1] + d1[ 0] + u11   ;
                v12:=d0[x2] + d0[x3] + u12    + d1[x3];
                v21:=d2[ 0] + u21    + d3[ 0] + d3[x1];
                v22:=u22    + d2[x3] + d3[x2] + d3[x3];
                if ((v11>v) or (v22>v)) and ((v12>v) or (v21>v)) then
                begin
                    buf[c*2+1]+= sqr(d1[ 0] + 2*u11 - 2*u12 - d1[x3])
                               + sqr(d2[ 0] + 2*u21 - 2*u22 - d2[x3])
                               + sqr(d0[x1] + 2*u11 - 2*u21 - d3[x1])
                               + sqr(d0[x2] + 2*u12 - 2*u22 - d3[x2])
                               + sqr(d0[ 0] + 2*u11 - 2*u22 - d3[x3])
                               + sqr(d3[ 0] + 2*u21 - 2*u12 - d0[x3]);
                    cnt[c]+=1;
                end;
                d0+=1; d1+=1; d2+=1; d3+=1;
            end;
        end;
    end;

    result:=1;
    for c:=0 to fi.nc-1 do
    begin
        v:=sqrt(buf[c*2+1]/(cnt[c]*54));
        if c<=high(sharpness) then sharpness[c]:=v;
        result*=v;
    end;
    if result>1e-50 then result:=exp(ln(result)/fi.nc);

    freemem(buf,fi.nc*2*sizeof(double));
    freemem(cnt,fi.nc*sizeof(longint));

    cc.debug('SImg_GetSharpness1: finished');
    SImg_DecDebugLevel;
end;

function SImg_GetSharpness1(cc:CSImg_Config; fi:TSImg_FPImage; thr:double):double;
var a : array[0..0] of double;
begin
    result:=SImg_GetSharpness1(cc,fi,thr,a);
end;

{*************************************************************************}
{******* SImg_GetSharpness2 **********************************************}
{*************************************************************************}
{ edge detect algorithm, optical resolution must be close or higher than sensor resolution
  thr: threshold; feasible: 0..1; 0.999..1 for stars
  returned value is the geometric mean of single color values }
function SImg_GetSharpness2(cc:CSImg_Config; fi:TSImg_FPImage; thr:double; var sharpness:array of double):double;
const qfact = 0.25;
var u,v,w,r,s       : double;
    buf             : ^double;
    cnt             : ^longint;
    d0,d1,d2        : ^simgfp;
    e               : ^double;
    x,y,c           : longint;
    x1,x2           : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_GetSharpness2: fi='+SImg_InfoStr2(fi)+'  thr='+real2str(thr));
    if not SImg_Recover(cc,fi) then cc.error('SImg_GetSharpness2: Invalid image: fi');

    getmem(buf,fi.nc*3*sizeof(double));

    { get max. number of saturated pixels for threshold limitation }
    y:=round(sqrt(fi.nx*fi.ny*1e-4));
    if fi.nc>1 then
    begin
        getmem(cnt,fi.nc*sizeof(longint));
        for c:=0 to fi.nc-1 do
        begin
            buf[c*2]:=1e-100;
            buf[c*2+1]:=0;
            cnt[c]:=0;
        end;

        d0:=fi.pt;
        for x:=0 to fi.nx*fi.ny-1 do
            for c:=0 to fi.nc-1 do
            begin
                r:=d0^; d0+=1;
                if r>buf[c*2] then buf[c*2]:=r;
                buf[c*2+1]+=r;
            end;

        for c:=0 to fi.nc-1 do
            buf[c]:=0.1*buf[c*2+1]/(fi.nx*fi.ny)+0.9*buf[c*2];

        d0:=fi.pt;
        for x:=0 to fi.nx*fi.ny-1 do
        begin
            for c:=0 to fi.nc-1 do
                if d0[c]>buf[c] then cnt[c]+=1;
            d0+=fi.nc;
        end;

        for c:=0 to fi.nc-1 do
            if cnt[c]>y then y:=cnt[c];
        freemem(cnt,fi.nc*sizeof(longint));
    end;
    r:=1-y/(fi.nx*fi.ny);
    if thr>r then thr:=r;

    for c:=0 to fi.nc-1 do
    begin
        buf[c*3]:=SImg_GetQuantile(cc,fi,c,thr)*qfact;
        buf[c*3+1]:=0;
        buf[c*3+2]:=10;  // regularization
    end;

    x1:=fi.nc;
    x2:=fi.nc*2;
    for y:=0 to fi.ny-3 do
    begin
        d0:=fi.pt + y*fi.sl;
        d1:=d0 + fi.sl;
        d2:=d0 + fi.sl*2;
        for x:=0 to fi.nx-3 do
        begin
            e:=buf;
            for c:=0 to fi.nc-1 do
            begin
                u:=d1[x1];
                w:=e[0]*0.5;
                if u>w then
                begin
                    v:=1.5*w-u;
                    if v<w then v:=w;
                    if (d1[0]>v) or (d1[x2]>v) then
                    begin
                        r:=sqr(0.5*(d1[x2]-d1[0]));
                        s:=sqr(u-d1[x2]);       if s>r then r:=s;
                        s:=sqr(u-d1[0]);        if s>r then r:=s;
                        e[1]+=r;
                        e[2]+=1;
                    end;
                    if (d0[x1]>v) or (d2[x1]>v) then
                    begin
                        r:=sqr(0.5*(d2[x1]-d0[x1]));
                        s:=sqr(u-d0[x1]);       if s>r then r:=s;
                        s:=sqr(u-d2[x1]);       if s>r then r:=s;
                        e[1]+=r;
                        e[2]+=1;
                    end;
                    if (d0[0]>v) or (d2[x2]>v) then
                    begin
                        r:=sqr(0.5*(d2[x2]-d0[0]));
                        s:=sqr(u-d0[0]);        if s>r then r:=s;
                        s:=sqr(u-d2[x2]);       if s>r then r:=s;
                        e[1]+=r*0.35;
                        e[2]+=0.7;
                    end;
                    if (d2[0]>v) or (d0[x2]>v) then
                    begin
                        r:=sqr(0.5*(d0[x2]-d2[0]));
                        s:=sqr(u-d2[0]);        if s>r then r:=s;
                        s:=sqr(u-d0[x2]);       if s>r then r:=s;
                        e[1]+=r*0.35;
                        e[2]+=0.7;
                    end;
                end;
                d0+=1; d1+=1; d2+=1;;
                e+=3;
            end;
        end;
    end;

    result:=1;
    for c:=0 to fi.nc-1 do
    begin
        v:=sqrt(buf[c*3+1]/buf[c*3+2]);
        if c<=high(sharpness) then sharpness[c]:=v;
        result*=v;
    end;
    if result>1e-50 then result:=exp(ln(result)/fi.nc);

    freemem(buf,fi.nc*3*sizeof(double));

    cc.debug('SImg_GetSharpness2: finished');
    SImg_DecDebugLevel;
end;

function SImg_GetSharpness2(cc:CSImg_Config; fi:TSImg_FPImage; thr:double):double;
var a : array[0..0] of double;
begin
    result:=SImg_GetSharpness2(cc,fi,thr,a);
end;

{*************************************************************************}
{******* SImg_GetSharpness3 **********************************************}
{*************************************************************************}
{ peak detect algorithm optimized for stars, optical resolution must be close to or higher than sensor resolution
  thr: threshold; valid: 0..1, feasible: 0.999..1
  returned value is the geometric mean of single color values }
function SImg_GetSharpness3(cc:CSImg_Config; fi:TSImg_FPImage; thr:double; var sharpness:array of double):double;
const qfact = 0.25;
var u,v,r           : double;
    buf             : ^double;
    cnt             : ^longint;
    d0,d1,d2        : ^simgfp;
    e               : ^double;
    x,y,c           : longint;
    x1,x2           : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_GetSharpness3: fi='+SImg_InfoStr2(fi)+'  thr='+real2str(thr));
    if not SImg_Recover(cc,fi) then cc.error('SImg_GetSharpness2: Invalid image: fi');

    getmem(buf,fi.nc*3*sizeof(double));

    y:=round(sqrt(fi.nx*fi.ny*1e-4));
    { get max. number of saturated pixels for threshold limitation }
    if fi.nc>1 then
    begin
        getmem(cnt,fi.nc*sizeof(longint));
        for c:=0 to fi.nc-1 do
        begin
            buf[c*2]:=1e-100;
            buf[c*2+1]:=0;
            cnt[c]:=0;
        end;

        d0:=fi.pt;
        for x:=0 to fi.nx*fi.ny-1 do
            for c:=0 to fi.nc-1 do
            begin
                r:=d0^; d0+=1;
                if r>buf[c*2] then buf[c*2]:=r;
                buf[c*2+1]+=r;
            end;

        for c:=0 to fi.nc-1 do
            buf[c]:=0.1*buf[c*2+1]/(fi.nx*fi.ny)+0.9*buf[c*2];

        d0:=fi.pt;
        for x:=0 to fi.nx*fi.ny-1 do
        begin
            for c:=0 to fi.nc-1 do
                if d0[c]>buf[c] then cnt[c]+=1;
            d0+=fi.nc;
        end;

        for c:=0 to fi.nc-1 do
            if cnt[c]>y then y:=cnt[c];
        freemem(cnt,fi.nc*sizeof(longint));
    end;
    r:=1-y/(fi.nx*fi.ny);
    if thr>r then thr:=r;

    for c:=0 to fi.nc-1 do
    begin
        buf[c*3]:=SImg_GetQuantile(cc,fi,c,thr)*qfact;
        buf[c*3+1]:=0;
        buf[c*3+2]:=10;  // regularization
    end;

    x1:=fi.nc;
    x2:=fi.nc*2;
    for y:=0 to fi.ny-3 do
    begin
        d0:=fi.pt + y*fi.sl;
        d1:=d0 + fi.sl;
        d2:=d0 + fi.sl*2;
        for x:=0 to fi.nx-3 do
        begin
            e:=buf;
            for c:=0 to fi.nc-1 do
            begin
                u:=d1[x1];
                v:=e[0];
                if u>v then
                begin
                    v*=0.5;
                    if (d1[0]>v) or (d1[x2]>v) then
                    begin
                        r:=2*u-d1[x2]-d1[0];
                        if r>=0 then
                        begin
                            e[1]+=r;
                            e[2]+=1;
                        end;
                    end;
                    if (d0[x1]>v) or (d2[x1]>v) then
                    begin
                        r:=2*u-d2[x1]-d0[x1];
                        if r>=0 then
                        begin
                            e[1]+=r;
                            e[2]+=1;
                        end;
                    end;
                    if (d0[0]>v) or (d2[x2]>v) then
                    begin
                        r:=2*u-d2[x2]-d0[0];
                        if r>=0 then
                        begin
                            e[1]+=r*0.35;
                            e[2]+=0.7;
                        end;
                    end;
                    if (d2[0]>v) or (d0[x2]>v) then
                    begin
                        r:=2*u-d0[x2]-d2[0];
                        if r>=0 then
                        begin
                            e[1]+=r*0.35;
                            e[2]+=0.7;
                        end;
                    end;
                end;
                d0+=1; d1+=1; d2+=1;;
                e+=3;
            end;
        end;
    end;

    result:=1;
    for c:=0 to fi.nc-1 do
    begin
        v:=buf[c*3+1]/buf[c*3+2];
        if c<=high(sharpness) then sharpness[c]:=v;
        result*=v;
    end;
    if result>1e-50 then result:=exp(ln(result)/fi.nc);

    freemem(buf,fi.nc*3*sizeof(double));

    cc.debug('SImg_GetSharpness3: finished');
    SImg_DecDebugLevel;
end;

function SImg_GetSharpness3(cc:CSImg_Config; fi:TSImg_FPImage; thr:double):double;
var a : array[0..0] of double;
begin
    result:=SImg_GetSharpness3(cc,fi,thr,a);
end;


{*************************************************************************}
{******* SImg_GetPSFInfo *************************************************}
{*************************************************************************}
{ Get PSF info }
procedure Simg_GetPSFInfo(cc:CSimg_Config; fs:TSImg_FPImage; var inf:TSImg_PSFInfo);
const zero_thr = 1e-6; { zero threshold relative to maximum }
      sig_thr  = 5e-3; { threshold for determining sig relative to maximum }
var x,y,xm,ym   : longint;
    xc,yc,sum   : double;
    ts,tz       : double;
    r,s,t,u     : double;
    max         : simgfp;
    d           : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fs) then cc.error('SImg_GetPSFInfo: Invalid image: fs');
    if fs.nc<>1 then cc.error('SImg_GetPSFInfo: Invalid dimensions: fs');

    { xc,yc,sum,max,xm,ym }
    sum:=0;
    xc:=0; yc:=0;
    d:=fs.pt;
    max:=d^; xm:=0; ym:=0;
    for y:=0 to fs.ny-1 do
        for x:=0 to fs.nx-1 do
        begin
            r:=d^; d+=1;
            sum+=r;
            xc+=x*r;
            yc+=y*r;
            if r>max then
            begin
                max:=r;
                xm:=x;
                ym:=y;
            end;
        end;
    if sum<simg_mina then cc.error('SImg_GetPSFInfo: Invalid PSF: sum<0');
    inf.sum:=sum;
    sum:=1/sum;
    xc*=sum; inf.xc:=xc;
    yc*=sum; inf.yc:=yc;
    inf.max:=max;
    inf.xm:=xm; inf.ym:=ym;

    { zrad, first sig estimation using sig_thr }
    tz:=max*zero_thr;
    ts:=max*sig_thr;
    if xc<yc then u:=xc+1
    else u:=yc+1;
    if fs.nx-xc<u then u:=fs.nx-xc;
    if fs.ny-yc<u then u:=fs.ny-yc;
    u:=sqr(u);
    t:=simg_mina; sum:=t;
    d:=fs.pt;
    for y:=0 to fs.ny-1 do
        for x:=0 to fs.nx-1 do
        begin
            r:=d^; d+=1;
            s:=sqr(x-xc)+sqr(y-yc);
            if (r<tz) and (s<u) then u:=s;
            if r>ts then
            begin
                r/=0.25+sqrt(s);        { perimeter compensation because 2D-sum }
                t+=s*r;
                sum+=r;
            end;
        end;
    inf.zrad:=sqrt(u);

    { final sig estimation using sig_thr and radius filter }
    u:=2+t/sum*sqr(ln(sig_thr));        { assuming exponential distribution }
    t:=simg_mina; sum:=t;
    d:=fs.pt;
    for y:=0 to fs.ny-1 do
        for x:=0 to fs.nx-1 do
        begin
            r:=d^; d+=1;
            s:=sqr(x-xc)+sqr(y-yc);
            if (r>ts) and (s<u) then
            begin
                r/=0.25+sqrt(s);        { perimeter compensation (critical because 2D-sum) }
                t+=s*r;
                sum+=r;
            end;
        end;
    inf.sig:=sqrt(t/sum);
end;

function SImg_PSFInfoString(const inf:TSImg_PSFInfo):ansistring;
begin
    with inf do
        result:='center=('+real2str(xc,-1)+','+real2str(yc,-1)+')  sig='+real2str(sig,-2)+'  sum='+real2str(sum,5)+'  max='+real2str(max,5)+' at ('+real2str(xm,-1)+','+real2str(ym,-1)+')  zrad='+real2str(zrad,-1);
end;

function SImg_PSFInfo_test(cc:CSImg_Config;sig:double):ansistring;
var fs          : TSImg_FPImage;
    inf         : TSImg_PSFInfo;
    x,y,rad     : longint;
    r,s         : double;
    d           : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;

    rad:=round(2+sig*10);
    sig:=-1/(2*sqr(sig));
    fs.pt:=nil;
    SImg_New(cc, fs, 2*rad+1,2*rad+1,1);
    d:=fs.pt;
    s:=0;
    for y:=-rad to rad do
        for x:=-rad to rad do
        begin
            r:=exp((sqr(x)+sqr(y))*sig);
            s+=r;
            d^:=r;
            d+=1;
        end;

    SImg_GetPSFInfo(cc,fs,inf);
    result:=SImg_PSFInfoString(inf);
    SImg_Free(cc,fs);
end;


{*************************************************************************}
{******* SImg_GetPSFWidth ************************************************}
{*************************************************************************}
{ Return PSF width (diameter) at a given brightness level relative to max }
function SImg_GetPSFWidth(cc:CSimg_Config; fs:TSImg_FPImage; l:double):double;
var r   : simgfp;
    i,j : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fs) then cc.error('SImg_GetPSFWidth: Invalid image: fs');
    if fs.nc<>1 then cc.error('SImg_GetPSFWidth: Invalid dimensions: fs');
    r:=0;
    for i:=0 to fs.size-1 do
        if fs.pt[i]>r then r:=fs.pt[i];
    r*=l; j:=0;
    for i:=0 to fs.size-1 do
        if fs.pt[i]>r then j+=1;
    result:=2*sqrt(j/pi);
end;


{*************************************************************************}
{******* SImg_GetPSFFWHM *************************************************}
{*************************************************************************}
{ Measure the FWHM at 25% level }
function SImg_GetPSFFWHM(cc:CSimg_Config; fs:TSImg_FPImage):double;
begin
    result:=SImg_GetPSFWidth(cc, fs, 0.25)*sqrt(0.5);
end;
