{  simg -- image processing library
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

// define[debug_rawcorrect]

{*************************************************************************}
{******* SImg_GetPixelMask ***********************************************}
{*************************************************************************}
{ Calculate (dead/bad) pixel mask using average and square noise image as calculated by SImg_AvgSet
    avg   : avg image
    sn    : square noise image
    dest  : mask image output
    rn    : read out noise in image units
    gain  : gain in electrons per image unit
}
procedure SImg_GetPixelMask(cc:CSImg_Config; avg,sn:TSImg_FPImage; var dest:TSImg_FPImage; rn,gain:double);
const thr = 1.5; { underweight threshold }
var fi          : TSImg_FPImage;
    i           : longint;
    adr,r,s,t,u : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_GetPixelMask: avg= '+SImg_InfoStr2(avg)+' sn= '+SImg_InfoStr2(sn)+' dest= '+SImg_InfoStr2(dest)+
        ' rn='+real2str(rn)+' gain='+real2str(gain));
    if not SImg_Recover(cc,avg) then cc.Error('SImg_GetPixelMask: Invalid input image');
    if not SImg_Recover(cc,sn) then cc.Error('SImg_GetPixelMask: Invalid input image');
    if avg.size<>dest.size then cc.Error('SImg_GetPixelMask: Size mismatch between input images');
    if (not SImg_OverWriteAlert(avg,dest)) and (not SImg_OverWriteAlert(sn,dest)) then SImg_Free(cc,dest);

    SImg_New(cc, fi, avg.nx,avg.ny,avg.nc, false);

    rn:=sqr(rn)*thr;
    if gain>=1e-50 then adr:=thr/gain
    else adr:=0;

    r:=SImg_GetQuantile(cc,sn,0.004);
    t:=1/r;
    for i:=0 to fi.size-1 do
    begin
        // dead pixel ?
        s:=sn.pt[i];
        if s<r then s*=t
        else begin
            u:=avg.pt[i]*adr+rn;
            if u<1e-30 then u:=1e-30;
            if s>u then s:=u/s
            else s:=1;
        end;
        fi.pt[i]:=s;
    end;

    SImg_Move(cc,fi,dest);
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_BGBrightness ***********************************************}
{*************************************************************************}
{ Calculates background brightness
  fi    : input
  gain  : gain in electrons per image unit
}
function SImg_BGBrightness(cc:CSImg_Config; fi:TSImg_FPImage; gain:double):double;
{ Assumed background brightness is the max(q1,r)*q2-quantile where r is the ratio of pixels darker than the threshold thr in electrons. }
const thr = 6;
      q1  = 0.20;
      q2  = 0.45;
var i,j    : longint;
    r      : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_BGBrightness: fi='+SImg_InfoStr2(fi)+ ' gain='+real2str(gain));
    if not SImg_Recover(cc,fi) then cc.error('SImg_BGBrightness: Invalid image: fi');
    if gain>1e-8 then
    begin
        r:=thr/gain;
        j:=0;
        for i:=0 to fi.size-1 do
            if fi.pt[i]<r then j+=1;
        r:=j/fi.size;
        if r<q1 then r:=q1;
    end
    else r:=q1;
    result:=SImg_GetQuantile(cc,fi,r*q2);
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_BGNoise ****************************************************}
{*************************************************************************}
{ Calculates background noise
  fi    : input
  rn    : read out noise in image units
  gain  : gain in electrons per image unit
}
function SImg_BGNoise(cc:CSImg_Config; fi:TSImg_FPImage; rn,gain:double):double;
var r      : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_BGNoise: fi='+SImg_InfoStr2(fi)+ ' rn='+real2str(rn)+ ' gain='+real2str(gain));
    if not SImg_Recover(cc,fi) then cc.error('SImg_BGNoise: Invalid image: fi');
    if rn<-1e-8 then cc.error('SImg_BGNoise: Invalid read out noise');
    if gain<1e-8 then cc.error('SImg_BGNoise: Invalid gain');

    r:=SImg_BGBrightness(cc,fi,gain);
    if r<0 then r:=0;
//    cc.info('bg='+real2str(r,5));
    result:=sqrt(r/gain+sqr(rn));
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_NoiseEstimator *********************************************}
{*************************************************************************}
{ single image square noise estimator, see ../test/noise-estimator.pas, y5
    fi          : input image
    dest        : output image
    bn          : Background noise, can be calculated using SImg_BGNoise(cc,fi,rn,gain);
    gain        : gain in electrons per image unit
    add         : square noise estimation is added to dest
    result      : maximum underestimation error
}
(*function SImg_NoiseEstimator(cc:CSImg_Config; fi:TSImg_FPImage; var dest:TSImg_FPImage; bn,gain:double; add:boolean):double;
var fw          : TSImg_FPImage;
    i           : longint;
    b,r2b,r,adr : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_NoiseEstimator: dest='+SImg_InfoStr2(dest)+' src='+SImg_InfoStr2(fi)+' bn='+real2str(bn)+' gain='+real2str(gain)+' add='+boolean2str(add));
    if not SImg_Recover(cc,fi) then cc.Error('SImg_NoiseEstimator: Invalid input image');
    if add then
    begin
        if not SImg_Recover(cc,dest) then cc.error('SImg_NoiseEstimator: Invalid output image');
        if SImg_OverwriteAlert(dest,fi) then cc.error('SImg_NoiseEstimator: Input and output images must not be the same in add mode');
        if fi.size<>dest.size then cc.error('SImg_NoiseEstimator: Invalid image dimensions of dest');
        fw:=dest;
    end else
    begin
        if not SImg_OverwriteAlert(dest,fi) then SImg_Free(cc,dest);
        SImg_Duplicate(cc, fi,fw, false)
    end;

    if gain>=1e-50 then adr:=1/gain
    else adr:=0;
    if bn>adr then bn:=sqr(bn)
    else bn*=adr;
    b:=25*sqr(sqr(adr))/bn;
    if b<1e-200 then b:=1e-200;
    r2b:=1/(2*b);
//    cc.info('Simg_NoseEstimator: sqr(bgNoise)='+real2str(bn)+'   b='+real2str(b,6));
    for i:=0 to fi.size-1 do
    begin
        r:=fi.pt[i]*adr-bn;
        if r<0 then r:=0
        else if r<b then r:=sqr(r)*r2b  { at µ*adr-bn=b square noise is underestimated by factor 2 which is the maximum }
        else r-=0.5*b;
        fw.pt[i]+=bn+r
    end;

    result:=0.5*b+bn;
    SImg_Move(cc,fw,dest);
    SImg_DecDebugLevel;
end; *)


{*************************************************************************}
{******* SImg_DarkCurrentSubtraction *************************************}
{*************************************************************************}
{ subtract variable dark current src2 from src1 (difference-fitting)
  pm     : pixel mask
  st     : step size (2 for Bayer images, 0 otherwise
  result : subtraction factor
}
function SImg_DarkCurrentSubtraction(cc:CSImg_Config; src1,src2,pm:TSImg_FPImage; var dest:TSImg_FPImage; st:longint):double;
var fi              : TSImg_FPImage;
    i,j,y,ix,iy     : longint;
    fg,gg,r,s,wx,wy : double;

begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_DarkCurrentSubtraction: dest= '+SImg_InfoStr2(dest)+' src1='+SImg_InfoStr2(src1)+' src2='+SImg_InfoStr2(src2)+' pm='+SImg_InfoStr2(pm)+' st='+int2str(st));
    if not SImg_Recover(cc,src1) then cc.Error('SImg_DarkCurrentSubtraction: Invalid image: src1');
    if not SImg_Recover(cc,src2) then cc.Error('SImg_DarkCurrentSubtraction: Invalid image: src2');
    if (src1.nx<>src2.nx) or (src1.ny<>src2.ny) or (src1.nc<>src2.nc) then cc.Error('SImg_DarkCurrentSubtraction: Images must have the same size');

    if pm.pt<>nil then
    begin
        if not SImg_Recover(cc,pm) then
        begin
            cc.warning('SImg_DarkCurrentSubtraction: Invalid image: pm');
            pm.pt:=nil;
        end;
        if (src1.nx<>pm.nx) or (src1.ny<>pm.ny) or (src1.nc<>pm.nc) then
        begin
            cc.warning('SImg_DarkCurrentSubtraction: Invalid size: pm');
            pm.pt:=nil;
        end;
    end;

    if (not SImg_OverwriteAlert(dest,src1)) and (not SImg_OverwriteAlert(dest,src2)) and (not SImg_OverwriteAlert(dest,pm)) then SImg_Free(cc,dest);

    if st<1 then st:=1;

    fg:=0; gg:=1e-200;
    for y:=0 to src1.ny-st-1 do
        for j:=0 to (src1.nx-st)*src1.nc-1 do
        begin
            i:=y*src1.sl + j;
            ix:=i+src1.nc*st;
            iy:=i+src1.sl*st;

            { weights }
            if pm.pt=nil then
            begin
                wx:=1;
                wy:=1;
            end
            else begin
                r:=pm.pt[i];
                if r<simg_mina then continue;
                r:=1/r;
                wx:=pm.pt[ix];
                if wx<simg_mina then wx:=0
                else wx:=1/(r+1/wx);
                wy:=pm.pt[iy];
                if wy<simg_mina then wy:=0
                else wy:=1/(r+1/wy);
            end;

            r:=src1.pt[i]-src1.pt[iy];
            s:=src2.pt[i]-src2.pt[iy];
            fg+=r*s*wy;
            gg+=s*s*wy;

            r:=src1.pt[i]-src1.pt[ix];
            s:=src2.pt[i]-src2.pt[ix];
            fg+=r*s*wx;
            gg+=s*s*wx;
        end;

    r:=fg/gg;
    result:=r;

    SImg_New(cc, fi, src1.nx,src1.ny,src1.nc, false);
    for i:=0 to fi.size-1 do
        fi.pt[i]:=src1.pt[i]-src2.pt[i]*r;

    SImg_Move(cc,fi,dest);
    SImg_DecDebugLevel;
end;



{*************************************************************************}
{******* SImg_PSDenoise **************************************************}
{*************************************************************************}
{ calculate 3 background values for each row
  bgbuf: output buffer
  thr  : brightness threshold
  mt   : mask threshold
}
threadprocedure SImg_PSDenoise_thread(ya:longint; yz:longint; fj:TSImg_FPImage; mi:TSImg_FPImage; bgbuf:pdouble; thr:double; mt:double);
const quant  = 0.15;
      sort_max = 100;
var buf                 : array[0..sort_max] of double;  { last element is a dummy in order to avoid conditionals }
    x,y,c,j,k,m,nc      : longint;
    n1,n2,n3,n4,xn,xo   : longint;
    d,e                 : ^simgfp;
    a                   : ^double;
    r,r1,r2,r3,s1,s2    : double;
    w1,w2,w3            : double;
    a0,a1,a2            : simgfp;
begin
    mt*=9;
    xn:=(fj.nx-6) div 12;
    xo:=((fj.nx-6) mod 12) div 2;
    nc:=fj.nc;
    n3:=round(quant*xn+0.50001);
    if n3<4 then n3:=4;
    if n3>sort_max then n3:=sort_max;
    n4:=round(n3*0.5);
    if n4<3 then n4:=3;
    buf[0]:=0;  { makes the compiler happy }
    w1:=3;
    for y:=ya to yz do
        for c:=0 to nc-1 do
        begin
            k:=+y*fj.sl + xo*nc + c;
            d:=fj.pt+k;
            e:=mi.pt+k;
            if mi.pt=nil then
            begin
                r2:=d[0]+d[nc]+d[nc*2];  d+=3*nc;
                r3:=d[0]+d[nc]+d[nc*2];  d+=3*nc;
                w2:=3;
                w3:=3;
            end
            else begin
                a0:=e[0];  a1:=e[nc];  a2:=e[2*nc];
                r2:=d[0]*a0 + d[nc]*a1 + d[nc*2]*a2;
                w2:=a0+a1+a2;
                d+=3*nc; e+=3*nc;
                a0:=e[0];  a1:=e[nc];  a2:=e[2*nc];
                r3:=d[0]*a0 + d[nc]*a1 + d[nc*2]*a2;
                w3:=a0+a1+a2;
                d+=3*nc; e+=3*nc;
            end;
            a:=bgbuf + 3*c*fj.ny + 3*y;
            for k:=0 to 3 do
            begin
                m:=0;
                for x:=0 to xn-1 do
                begin
                    // 9 pixel average calculated every 3 pixels
                    r1:=r2;
                    r2:=r3;
                    if mi.pt=nil then r3:=d[0]+d[nc]+d[nc*2]
                    else begin
                        w1:=w2;
                        w2:=w3;
                        a0:=e[0];  a1:=e[nc];  a2:=e[2*nc];
                        r3:=d[0]*a0 + d[nc]*a1 + d[nc*2]*a2;
                        w3:=a0+a1+a2;
                        e+=3*nc;
                    end;
                    d+=3*nc;
                    if (r1<=thr*w1) and (r2<=thr*w2) and (r3<=thr*w3) then m+=1;
                    r:=w1+w2+w3;
                    if r>mt then
                    begin
                        r:=(r1+r2+r3)/r;
                        // sort n darkest pixels
                        if x>n3 then j:=n3
                        else j:=x;
                        while (j>0) and (r<buf[j-1]) do
                        begin
                            buf[j]:=buf[j-1];
                            j-=1;
                        end;
                        buf[j]:=r;
                    end;
                end;
                s2:=n3*m/xn;
                if s2<n4 then s2:=n4;
                s1:=s2*0.667;
                n1:=round(s1+0.49999);
                n2:=round(s2-0.50001);
//              if (y>1500) and (y<1750) then write(stderr,n1:5, n2:5, n3:5, '        ');
                r:=(n1-s1)*buf[n1-1] + (s2-n2)*buf[n2];
                for j:=n1 to n2-1 do
                    r+=buf[j];
                r/=(s2-s1);
                case k of
                    0: a[0]:=r;
                    1: a[1]:=r;
                    2: if r<a[1] then a[1]:=r;
                    else a[2]:=r;
                end;
            end;
//      if (y>1500) and (y<1750) then writeln(stderr,y);
        end;
end;


{ Reduces line patterns caused by power supply noise
  ny   : number of rows
  fmin : non-repetitive frequency limit
  ibuf : input buffer
  obuf : output buffer
  rp   : reduced detection of periodic noise
}
procedure SImg_PSDenoise_filter(ny:longint; fmin:double; ibuf,obuf:pdouble; rp:boolean);
const maxit = 2;
var r,r1,r2,r3          : double;
    y,y2,yz,i,j,it      : longint;
    s1,s2,ft,ft1,ft2,nh : longint;
    tbuf                : ^double;
begin
    getmem2(tbuf,ny*sizeof(double));

    // filter non-periodic patterns
    for y:=0 to ny-1 do
        tbuf[y]:=ibuf[y*3];
    j:=round(ny/(3*fmin)+1);
    if j<3 then j:=3;
    for y:=0 to ny-1 do
    begin
        yz:=ny-1-y;
        if j<yz then yz:=j;
        if y<yz then yz:=y;
        r:=0;
        for y2:=y-yz to y+yz do
            r+=tbuf[y2];
        r/=2*yz+1;
        obuf[y]:=tbuf[y]-r;
    end;

    // filter periodic patterns: window width for frequency fmin/n is y/fmin*sqr(n) = <period>*n
    ft1:=round(fmin*1.5+0.5);   // maximum test frequency
    ft2:=round(sqrt(1.5*fmin)); // minimum test frequency
    if rp then
    begin
        i:=round(fmin*0.5-0.5);
        if i>ft2 then ft2:=i;
    end;
    if ft2<3 then ft2:=3;
    for it:=1 to maxit do
    begin
        for ft:=ft1 downto ft2 do // test frequency
        begin
            nh:=round(fmin/ft+0.5);     // number of test periods
            r1:=0.25*ny/ft;             // 1/4 period

            for y:=0 to ny-1 do
                tbuf[y]:=ibuf[y*3]-obuf[y];

            for y:=0 to ny-1 do
            begin
                { calculate constant component }
                r3:=0; y2:=0;
                for i:=1 to nh do
                begin
                    j:=round((i*2-1)*r1);
                    if y>=j then
                    begin
                        r3+=tbuf[y-j];
                        y2+=1;
                    end;
                    if y+j<ny then
                    begin
                        r3+=tbuf[y+j];
                        y2+=1;
                    end;
                end;
                if y2<2 then continue;
                r3/=y2;
                { calculate periodic component }
                r:=tbuf[y]-r3;
                s1:=1;
                if r<0 then
                begin
                    r:=-r;
                    s1:=-1;
                end;
                for i:=1 to nh do
                begin
                    j:=round(i*2*r1);
                    if (i and 1)=0 then s2:=s1
                    else s2:=-s1;
                    if y>=j then
                    begin
                        r2:=s2*(tbuf[y-j]-r3);
                        if r2<r then r:=r2;
                    end;
                    if y+j<ny then
                    begin
                        r2:=s2*(tbuf[y+j]-r3);
                        if r2<r then r:=r2;
                    end;
                    if r<0 then break;
                end;
                if r>0 then obuf[y]+=r*s1;
//              if (it=1) and (ft=12) then writeln(stderr,y:4, tbuf[y]-r3:16:6, r:16:6);
            end;
        end;
    end;
    freemem2(tbuf,ny*sizeof(double));
end;


{ Reduces line patterns caused by power supply noise
  fi    : input
  zi    : dark field, zi.pt=nil to ignore it
  oi    : flat field, oi.pt=nil to ignore it
  mi    : mask image, oi.pt=nil to ignore it
  fj    : output
  fmin  : limiting frequency
  o1    : enable order 1 corrections (prone to artifacts)
  rp    : reduced detection of periodic noise, recommended if black reference correction was applied
}
procedure SImg_PSDenoise(cc:CSImg_Config; fi,zi,oi,mi:TSImg_FPImage; var fj:TSImg_FPImage; fmin:double; o1,rp:boolean);
const minxsize    = 512;  { minimum ratio of pixels within threshold }
      invalid_thr = 8;    { invalid data threshold relative to mean deviation }
      quant_thr   = 0.8;  { ignore pixels (3 x down-sampled) brighter than this threshold }
      quant_oi    = 0.3;  { values below a threshold defined by this parameter are used for determination of the factor for flatfield _subtraction_ }
var ya,yz,y,c,i,j,tn    : longint;
    bgbuf,ob1,ob2,ob3,d : ^double;
    oifact              : ^double;
    e,f,g               : ^simgfp;
    vbuf,b              : ^byte;   { 1: valid }
    r,r1,r2,r3          : double;
    s,s1,s2,s3          : double;
    v                   : matrix;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_PSDenoise: fi='+SImg_InfoStr2(fi)+' zi='+SImg_InfoStr2(zi)+' oi='+SImg_InfoStr2(zi)+' mi='+SImg_InfoStr2(mi)+' fj='+SImg_InfoStr2(fj)+ ' fmin='+real2str(fmin)+' o1='+boolean2str(o1)+' rp='+boolean2str(rp));
    if not SImg_Recover(cc,fi) then cc.error('SImg_PSDenoise: Invalid image: fi');
    if fi.nx<minxsize then cc.error('SImg_PSDenoise: Image to small');
    if fmin<3 then fmin:=3;
    if not SImg_OverwriteAlert(fi,fj) then
    begin
        SImg_Free(cc,fj);
        SImg_Duplicate(cc,fi,fj,true);
    end;
    if (zi.pt<>nil) and (not SImg_Recover(cc,zi)) then
    begin
        cc.warning('SImg_PSDenoise: Invalid dark current image, ignoring it');
        zi.pt:=nil;
    end;
    if (zi.pt<>nil) and (zi.size<>fj.size) then
    begin
        cc.warning('SImg_PSDenoise: Invalid size of dark current image, ignoring it');
        zi.pt:=nil;
    end;
    if (oi.pt<>nil) and (not SImg_Recover(cc,oi)) then
    begin
        cc.warning('SImg_PSDenoise: Invalid dark current image, ignoring it');
        oi.pt:=nil;
    end;
    if (oi.pt<>nil) and (oi.size<>fj.size) then
    begin
        cc.warning('SImg_PSDenoise: Invalid size of dark current image, ignoring it');
        oi.pt:=nil;
    end;
    if (mi.pt<>nil) and (mi.size<>fj.size) then
    begin
        cc.warning('SImg_PSDenoise: Invalid size of mask image, ignoring it');
        mi.pt:=nil;
    end;

    // subtract zi
    if zi.pt<>nil then
    begin
        for i:=0 to fj.size-1 do
            fj.pt[i]-=zi.pt[i];
    end;

    // subtract oi
    getmem2(oifact,fj.nc*sizeof(double));
    for c:=0 to fj.nc-1 do
        oifact[c]:=-1;
    if oi.pt<>nil then
    begin
        s1:=SImg_GetQuantile(cc,fj,quant_oi*0.3);
        s2:=SImg_GetQuantile(cc,fj,quant_oi);  s2+=2*(s2-s1);
        for c:=0 to fj.nc-1 do
        begin
            r1:=0; r2:=0;
            e:=fj.pt+c;
            f:=oi.pt+c;
            g:=mi.pt+c;
            for i:=0 to fj.nx*fj.ny-1 do
            begin
                if mi.pt<>nil then r3:=g^
                else r3:=1;
                r:=e^;
                if (r>=s1) and (r<=s2) then
                begin
                    r1+=r*r3;
                    r2+=f^*r3;
                end;
                e+=fj.nc;
                f+=fj.nc;
            end;
//ifdef[debug_rawcorrect]
                cc.info('rawcorrect: flatfield factor for color '+int2str(c)+': '+real2str(r1/r2,6)+' (thr: '+real2str(s1,6)+' ... ' + real2str(s2,6) +')');
//endif
            if (r2>1-10) and (r1>0) then
            begin
                r1/=r2;
                oifact[c]:=r1;
                e:=fj.pt+c;
                f:=oi.pt+c;
                for i:=0 to fj.nx*fj.ny-1 do
                begin
                    e^-=r1*f^;
                    e+=fj.nc;
                    f+=fj.nc;
                end;
            end;
        end;
    end;


    // calculate background
    r:=SImg_GetQuantile(cc,fj,quant_thr);
    if mi.pt<>nil then s:=SImg_GetQuantile(cc,mi,0.9)*0.5       { mask threshold }
    else s:=0.5;
    getmem2(bgbuf,sizeof(double)*fj.ny*fj.nc*3);
    tn:=cc.threads;
    ya:=0;
    for i:=0 to tn-1 do
    begin
        yz:=fj.ny*(i+1) div tn;
        dothread SImg_PSDenoise_thread(ya,yz-1, fj,mi, bgbuf, r,s);
        ya:=yz;
    end;
    threadwait;

    am2(v,zmatrix(1,3));
    getmem2(vbuf,fj.ny*3);
    for c:=0 to fj.nc-1 do
    begin
        { mark invalid rows }
        b:=vbuf;
        d:=bgbuf + c*fj.ny*3;
        r1:=d[0];
        r2:=d[0];
        for i:=1 to fj.ny*3-1 do
        begin
            r:=d[i];
            if r<r1 then r1:=r;
            if r>r2 then r2:=r;
        end;
        r:=(r2-r1)*1e-6;
        s1:=0; s2:=0; s3:=0; j:=0;
        for y:=0 to fj.ny-1 do
        begin
            r1:=d[0];
            r2:=d[1];
            r3:=d[2];
            d+=3;
            if (abs(r1-r2)<r) and (abs(r1-r3)<r) and (abs(r2-r3)<r) then i:=0
            else begin
                s1+=r1;
                s2+=r2;
                s3+=r3;
                j+=1;
                i:=1;
            end;
            b[0]:=i;
            b[1]:=i;
            b[2]:=i;
            b+=3;
        end;

        { mark invalid (unlikely bright) sub-rows }
        b:=vbuf;
        d:=bgbuf + c*fj.ny*3;
        r:=1/j;
        s1*=r; s2*=r; s3*=r;
        r1:=0; r2:=0; r3:=0;
        for y:=0 to fj.ny-1 do
        begin
            if b^>0 then
            begin
                r1+=abs(d[0]-s1);
                r2+=abs(d[1]-s2);
                r3+=abs(d[2]-s3);
            end;
            b+=3; d+=3;
        end;
        r*=invalid_thr;
        s1+=r1*r;
        s2+=r2*r;
        s3+=r3*r;
        b:=vbuf;
        d:=bgbuf + c*fj.ny*3;
        for y:=0 to fj.ny-1 do
        begin
            if d[0]>s1 then b[0]:=0;
            if d[1]>s2 then b[1]:=0;
            if d[2]>s3 then b[2]:=0;
//ifdef[debug_rawcorrect]
            if b[0]+b[1]+b[2]<3 then cc.info( 'SImg_PSDenoise: Invalid row: '+int2str(y)+' ('+int2str(b[0])+int2str(b[1])+int2str(b[2])+')' );
//endif[debug_rawcorrect]
            b+=3; d+=3;
        end;

        { polynomial approximation up to order 1 }
        d:=bgbuf + c*fj.ny*3;
        b:=vbuf;
        for i:=0 to 2 do
            v.pt[i]:=0;
        j:=0;  s:=0;  s1:=0;  s2:=0;  s3:=0;
        r:=2/(fj.ny-1);
        for y:=0 to fj.ny-1 do
        begin
            r3:=0; i:=0;
            if b[0]>0 then
            begin
                r3:=d[0];
                i:=1;
            end;
            if b[2]>0 then
            begin
                r3+=d[2];
                i+=1;
            end;
            d+=3;
            b+=3;

            r1:=y*r-1;
            r2:=sqr(r1);

            j+=i;
            s+=r1*i;
            s1+=r2*i;
            s2+=r2*r1*i;
            s3+=sqr(r2)*i;

            v.pt[0]+=r3;  v.pt[1]+=r3*r1;  v.pt[2]+=r3*r2;
            end;
        am(v,solve(rmatrix(3,3, [  j,  s, s1,
                                   s, s1, s2,
                                  s1, s2, s3]), v));

        { subtraction }
        d:=bgbuf + c*fj.ny*3;
        for y:=0 to fj.ny-1 do
        begin
            r1:=y*r-1;
            r2:=sqr(r1);
            r3:=(v.pt[0] + r1*v.pt[1] + r2*v.pt[2]);
            d[0]-=r3;
            d[1]-=r3;
            d[2]-=r3;
            d+=3;
        end;

        { interpolation of invalid rows }
        for i:=0 to 2 do
        begin
            d:=bgbuf + c*fj.ny*3 + i;
            b:=vbuf + i;
            ya:=-1;
            yz:=-1;
            for y:=0 to fj.ny-1 do
                if b[y*3]>0 then ya:=y
                else begin
                    if yz<=y then
                    begin
                        yz:=y+1;
                        while (yz<fj.ny) and (b[yz*3]=0) do
                            yz+=1;
                    end;
                    if (ya>=0) and (yz<fj.ny) then d[y*3]:=(d[ya*3]*(yz-y)+d[yz*3]*(y-ya))/(yz-ya)
                    else if ya>=0 then d[y*3]:=d[ya*3]
                    else if yz<fj.ny then d[y*3]:=d[yz*3];
//ifdef[debug_rawcorrect]
                    cc.info('SImg_PSDenoise: Interpolation: '+int2str(y)+' '+int2str(ya)+' '+int2str(yz));
//endif[debug_rawcorrect]
                end;
        end;
    end;
    free(v);
    freemem2(vbuf,fj.ny*3);


    getmem2(ob1,sizeof(double)*fj.ny);
    getmem2(ob2,sizeof(double)*fj.ny);
    getmem2(ob3,sizeof(double)*fj.ny);
    for c:=0 to fj.nc-1 do
    begin
        { filter order 1 data }
        if o1 then
        begin
            SImg_PSDenoise_filter(fj.ny, fmin*3, bgbuf+c*fj.ny*3  , ob1,rp);
            SImg_PSDenoise_filter(fj.ny, fmin*3, bgbuf+c*fj.ny*3+2, ob3,rp);
        end else
            for y:=0 to fj.ny-1 do
            begin
                ob1[y]:=0;
                ob3[y]:=0;
            end;

        { filter order 0 data }
        d:=bgbuf + c*fj.ny*3;
        for y:=0 to fj.ny-1 do
        begin
            r1:=ob1[y];
            r3:=ob3[y];
            r2:=0.5*(r1+r3);
            ob1[y]:=r2;
            ob2[y]:=r3-r1;

            s1:=d[0]-r1;
            r2:=d[1]-r2;
            if r2<s1 then s1:=r2;
            r3:=d[2]-r3;
            if r3<s1 then s1:=r3;
            d[0]:=s1;           // order 0
            d+=3;
        end;
        SImg_PSDenoise_filter(fj.ny, fmin, bgbuf+c*fj.ny*3, ob3,rp);

        { noise subtraction }
        e:=fj.pt + c;
        f:=zi.pt + c;
        g:=oi.pt + c;
        r:=1/(fj.nx-1);
        s:=oifact[c];
        for y:=0 to fj.ny-1 do
        begin
            r2:=ob2[y];
            r1:=ob1[y]+ob3[y] - 0.6*r2;
            r2*=1.2;
            for i:=0 to fj.nx-1 do
            begin
                e^-=r1 + i*r*r2;
                e+=fj.nc;
            end;
            if zi.pt<>nil then
            begin
                e-=fj.sl;
                for i:=0 to fj.nx-1 do
                begin
                    e^+=f^;
                    e+=fj.nc; f+=fj.nc;
                end
            end;
            if s>0 then
            begin
                e-=fj.sl;
                s:=oifact[c];
                for i:=0 to fj.nx-1 do
                begin
                    e^+=s*g^;
                    e+=fj.nc; g+=fj.nc;
                end
            end;
        end;
    end;

    freemem2(bgbuf,sizeof(double)*fj.ny*fj.nc*3);
    freemem2(ob1,sizeof(double)*fj.ny);
    freemem2(ob2,sizeof(double)*fj.ny);
    freemem2(ob3,sizeof(double)*fj.ny);
    if oifact<>nil then  freemem2(oifact,sizeof(double)*fj.nc);

    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_PhotonCount ************************************************}
{*************************************************************************}
{ Count photons. Output image is an integer image (16 unsigned) with an offset of 16 (16: 0 electrons, 17: 1 electron, ...)
  Scaling parameters preserve original scaling.
  fi    : input
  ri    : output
  gain  : gain
  nq    : noise quantization for better compression, about 0.5% SNR loss.
}
procedure SImg_PhotonCount(cc:CSImg_Config; fi:TSImg_FPImage; var ri:TSImg_RDImage; gain:double; nq:boolean);
var i,j,k  : longint;
    m1,m2  : longint;
    r      : double;
    w      : ^word;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_PhotonCount: fi='+SImg_InfoStr2(fi)+' fj='+SImg_InfoStr2(ri)+ ' gain='+real2str(gain)+' nq='+boolean2str(nq));
    if not SImg_Recover(cc,fi) then cc.error('SImg_PhotonCount: Invalid image: fi');

    SImg_New(cc,ri,SImg_RAWWord, fi.nx,fi.ny,fi.nc,false);
    if gain>1e-10 then ri.cf:=1/gain;
    ri.co:=-16*ri.cf;
    w:=ri.ptw;
    for i:=0 to fi.size-1 do
    begin
        r:=fi.pt[i]*gain+16;
        if r<0 then j:=0
        else if r>65535 then j:=65535
        else j:=round(r);
        if nq and (j>=25+16) then
        begin
            j-=16;
            k:=j div (25*4);
            m1:=65532;
            m2:=1;
            while k>1 do
            begin
                k:=k shr 2;
                m1:=m1 shl 1;
                m2:=m2 shl 1;
            end;
            j:=((j and m1) or m2)+16;
        end;
        w[i]:=j;
    end;
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_CropBRC ****************************************************}
{*************************************************************************}
{ Crop region of black reference columns.
  If no such region is defined result is false and output image is untouched
  nr : ignore ROI
}
function SImg_CropBRC(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSimg_FPImage; mi:CSImg_MetaInfo; nr:boolean):boolean;
var x0,y0,width,height  : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi) then cc.error('SImg_CropBRC: Invalid image');

    SImg_GetBRC(mi, fi.nx,fi.ny, nr, x0,y0,width,height);
//    cc.info('SImg_CropBRC: x0='+int2str(x0)+' y0='+int2str(y0)+' width='+int2str(width)+' height='+int2str(height));
    if width<1 then
    begin
        result:=false;
        exit;
    end;

    SImg_CutSimple(cc, fi,fj, x0,y0,width,height);
    result:=true;
end;


{*************************************************************************}
{******* SImg_GetBR ******************************************************}
{*************************************************************************}
{ Get black reference per color and line.
    fi    : input image
    fm    : mask image (optional)
    mi    : meta infos
    nr    : ignore ROI
    br    : output buffer allocated using getmem2
    result: length of the buffer (height*fi.nc)
  If no black reference columns are defined, result is null and buf is untouched
}
function SImg_GetBR(cc:CSImg_Config; fi,fm:TSImg_FPImage; mi:CSImg_MetaInfo; nr:boolean; var br:pdouble):longint;
var x0,y0,width,height  : longint;
    x,y,c               : longint;
    r,s                 : double;
    a                   : simgfp;
    d,f                 : ^simgfp;
    e                   : ^double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi) then cc.error('SImg_GetBRC: Invalid image: fi');

    SImg_GetBRC(mi, fi.nx,fi.ny, nr, x0,y0,width,height);
//    cc.info('SImg_GetBR: x0='+int2str(x0)+' y0='+int2str(y0)+' width='+int2str(width)+' height='+int2str(height));
    if width<1 then
    begin
        result:=0;
        exit;
    end;

    if fm.pt<>nil then
    begin
        if not SImg_Recover(cc,fm) then cc.error('SImg_GetBRC: Invalid image: fm');
        if (fm.nx<>width) or (fm.ny<>height) or (fm.nc<>fi.nc) then cc.error('SImg_GetBRC: Invalid dimensions: fm');
    end;

    result:=height*fi.nc;
    getmem2(br, result*sizeof(double));
    e:=br;
    s:=1/width;
    for y:=0 to height-1 do
    begin
        for c:=0 to fi.nc-1 do
        begin
            d:=fi.pt + (y0+y)*fi.sl + x0*fi.nc + c;
            r:=0;
            if fm.pt<>nil then
            begin
                f:=fm.pt + y*fm.sl + c;
                s:=0;
                for x:=0 to width-1 do
                begin
                    a:=f[x*fm.nc];
                    if a<simg_mina then a:=simg_mina;
                    r+=d[x*fi.nc]*a;
                    s+=a;
                end;
                r/=s;
            end
            else begin
                for x:=0 to width-1 do
                    r+=d[x*fi.nc];
                r*=s;
            end;
            e^:=r;
            e+=1;
        end;
    end;
end;


{*************************************************************************}
{******* SImg_BRCorrect **************************************************}
{*************************************************************************}
{ Get black reference per color and line.
    fi    : input and outout image
    br    : black reference data generated ny SImg_GetBR
    fmax  : maximum frequency; <.5: constant subtraction
}
procedure SImg_BRCorrect(cc:CSImg_Config; fi:TSImg_FPImage; br:pdouble; fmax:double);
var x,y,y2,ya,yz,c,j    : longint;
    r                   : double;
    a                   : simgfp;
    d                   : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_BRCorrect: fi='+SImg_InfoStr2(fi)+' fmax='+real2str(fmax));

    if fmax<0.50001 then
    begin
        for c:=0 to fi.nc-1 do
        begin
            { lowpass filtering }
            r:=0;
            for y2:=0 to fi.ny-1 do
                r+=br[y2*fi.nc+c];
            a:=r/fi.ny;
//          cc.info('SImg_BRCorrect: '+real2str(a));
            { subtraction }
            d:=fi.pt + c;
            for x:=0 to fi.nx*fi.ny-1 do
                d[x*fi.nc]-=a;
        end;
    end
    else begin
        j:=round(fi.nx/(2*fmax));
        for y:=0 to fi.ny-1 do
        begin
            ya:=y-j;
            if ya<0 then ya:=0;
            yz:=y+j;
            if yz>fi.ny-1 then yz:=fi.ny-1;
            for c:=0 to fi.nc-1 do
            begin
                { low-pass filtering }
                r:=0;
                for y2:=ya to yz do
                    r+=br[y2*fi.nc+c];
                a:=r/(yz-ya+1);
                { subtraction }
                d:=fi.pt + y*fi.sl + c;
                for x:=0 to fi.nx-1 do
                    d[x*fi.nc]-=a;
            end;
        end;
    end;

    SImg_DecDebugLevel;
end;
