{  simg -- image processing library
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{ todo:
    * manual modification mask:
        - disable stars
        - disable extended interpolation
}

// define[debug_findStars]
// define[debug_starSubtract]
// define[debug_getpsf]


{*************************************************************************}
{******* SImg_Subtract_NoiseFloor ****************************************}
{*************************************************************************}
{ Calculate the noise floor from a background corrected image, also see SImg_NoiseFloor_int.
  Due to the larger background filter image contains more object information. Therefore quantiles are shifted left.
  fj : bg corrected: minimum filter with radius > 2*<superresolution factor> should be subtracted
  sig,offs : noise floor and offset
}
procedure SImg_StarSubtract_NoiseFloor(cc:CSImg_Config; fj: TSImg_FPImage; var sig,offs:simgfp);
var qs          : array[0..1] of double;
    ps          : array[0..1] of double = (0.25, 0.57);         { assuming with stars }
begin
    if not SImg_Recover(cc,fj) then cc.error('SImg_StarSubtract_NoiseFloor: Invalid image: fj');
    SImg_GetQuantile(cc, fj, cc.threads,-1,ps,qs);
    sig:=qs[1]-qs[0];
    offs:=0.5*(qs[1]+qs[0]);
end;

{*************************************************************************}
{******* SImg_SubtractBackground *****************************************}
{*************************************************************************}
{ Subtract background from image, process saturated pixels and calculate noise floor and  offset
    fi       : input
    sh       : output
    rad      : Background interpolation radius
    fwhm     : Star FWHM, used for denoising and noise level estimation.
    ithr     : Interpolation threshold relative to noise >=-1, recommended: >0. Used for weight calculation.
    iwdt     : Interpolation width relative to FWHM. Valid: > 1, Recommended: > 2
    sat_thr  : saturation threshold relative to brightest pixel
}
procedure SImg_SubtractBackground(cc:CSImg_Config; fi:TSImg_FPImage; out sh:TSImg_StarSubtract_shared; rad:longint; fwhm,ithr,iwdt,sat_thr:double);
const lq      = 0.001;  { lower quantile  }
var fj,fk,fl    : TSImg_FPImage;
    i,x,y       : longint;
    xa,xz,ya,yz : longint;
    q,r,s,t,u,v : simgfp;
    sig         : simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_SubtractBackground: fi='+SImg_InfoStr2(fi)+' rad='+real2str(rad)+' fwhm='+real2str(fwhm)+' ithr='+real2str(ithr)+' iwdt='+real2str(iwdt)+' sat_thr='+real2str(sat_thr));
    if not SImg_Recover(cc,fi) then cc.error('SImg_SubtractBackground: Invalid image: fi');
    if fwhm<1 then fwhm:=1;
    sig:=FWHM_TO_SIG*fwhm;
    i:=round(4*sig);
    if rad<i then rad:=i;
    sh.rad:=rad;
    if iwdt<1 then iwdt:=1;
    r:=sqrt(1+sqr(fwhm))*0.5*iwdt;
    sh.irad:=r/(1+r/rad);       { sh.irad < sh.rad }
    if sat_thr>0.75 then sat_thr:=0.75;
    sh.sat_thr:=sat_thr;

    { minimum / maximum, lower quantile }
    s:=SImg_GetQuantile(cc,fi,lq);
    fj.pt:=nil;
    SImg_New(cc, fj, fi.nx,fi.ny,fi.nc, false);
    t:=s;
    for i:=0 to fi.size-1 do
    begin
        r:=fi.pt[i];
        if r>t then t:=r;
        if r<s then r:=s;
        fj.pt[i]:=r;
    end;
    t:=s+sat_thr*(t-s);
    SImg_Filter(cc, fj,fj, sig);

    { bg filter }
    fk.pt:=nil;  fl.pt:=nil;
//  SImg_WriteImageFile(cc,'subtractBackground-fj1.fits',fj);
    SImg_FBGFilter(cc, fj,fk, 2*sig, true, false);
    SImg_FBGFilter(cc, fk,fl, sqrt(sqr(rad)-sqr(2*sig)), true, true);

    { noise floor }
    for i:=0 to fi.size-1 do
        fk.pt[i]:=fi.pt[i]-fk.pt[i];
    SImg_StarSubtract_NoiseFloor(cc, fk, sh.sig, sh.offs);
    SImg_free(cc, fk);

    { subtract BG and handle saturated pixels }
    u:=1/t;
    s:=t-s;
    for i:=0 to fi.size-1 do
    begin
        r:=fi.pt[i];
        v:=fl.pt[i];
        fj.pt[i]-=v;
        if r<t then fl.pt[i]:=r-v
        else begin
            y:=i div fl.sl;
            x:=(i-y*fl.sl) div fl.nc;
            xa:=0;
            while (x+xa>0) and (fi.pt[i+xa*fi.nc]>t) do
                xa-=1;
            xz:=0;
            while (x+xz<fi.nx-1) and (fi.pt[i+xz*fi.nc]>t) do
                xz+=1;
            ya:=0;
            while (y+ya>0) and (fi.pt[i+ya*fi.sl]>t) do
                ya-=1;
            yz:=0;
            while (y+yz<fi.ny-1) and (fi.pt[i+yz*fi.sl]>t) do
                yz+=1;
            v:=fi.nx+fi.ny;
            for y:=ya to yz do
                for x:=xa to xz do
                begin
                    r:=fi.pt[i+y*fi.sl+x*fi.nc];
                    if r<t then
                    begin
                        r:=sqrt(sqr(y)+sqr(x))-1+r*u; { interpolated distance }
                        if r<v then v:=r;
                    end;
                end;
            fl.pt[i]:=s*(1 + v*0.1);
{           y:=i div fl.sl;
            x:=(i-y*fl.sl) div fl.nc;
            x-=2270;
            y-=1908;
            if sqr(x)+sqr(y)<10 then writeln(stderr,x,', ',y,'   ',v);}
        end;
    end;
    sh.sat:=s;
    sh.fi2:=fl;

    { weight image }
    SImg_FBGFilter(cc, fj,fk, sh.irad*0.7, true, false);
    t:=1/(sh.sig*ithr);
    u:=sh.offs+sh.sig;
    for i:=0 to fk.size-1 do
    begin
        q:=fl.pt[i];
        if q>s then fk.pt[i]:=0
        else begin
            q:=(q-u)*t*0.25-1;
            r:=(fj.pt[i]-u-fk.pt[i])*t;
            if (q<0) and (r<0) then fk.pt[i]:=1
            else begin
                if q<0 then q:=0;
                if r<0 then r:=0;
                fk.pt[i]:=(1+q)/((1+q+sqr(q))*(1+sqr(r)));
            end;
        end;
    end;
    sh.fw:=fk;
    SImg_Free(cc,fj);
//    SImg_WriteImageFile(cc,'starSubtract-fw.fits',sh.fw);
//    SImg_WriteImageFile(cc,'starSubtract-fi.fits',fi);
//    SImg_WriteImageFile(cc,'starSubtract-fi2.fits',sh.fi2);
//    cc.info('sat='+real2str(sh.sat));

    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_FindStars **************************************************}
{*************************************************************************}
type TSImg_FindStars_peak = record
            x,y : word;         { position }
            b   : simgfp;       { brightness }
            w,a : single;       { FWHM width and aspect ratio }
            s,v : boolean;      { saturated / valid }
        end;
     PSImg_FindStars_peak = ^TSImg_FindStars_peak;

{ find peaks }
threadprocedure SImg_FindStars_thread1(ti:longint; tn:longint; fi:TSImg_FPImage; gx:integer; gy:integer; minb:double; offs:simgfp; stars:PSImg_FindStars_peak);
const mbd = 20; { minimum distance to border }
var i,ia,iz,xi,xa,xz    : longint;
    yi,ya,yz,x,y        : longint;
    r,s                 : simgfp;
    d                   : ^simgfp;
begin
    ia:=(gx*gy*ti) div tn;
    iz:=(gx*gy*(ti+1)) div tn - 1;
    for i:=ia to iz do
    begin
        yi:=i div gx;
        xi:=i - yi*gx;
        xa:=(xi*fi.nx) div gx;
        xz:=((xi+1)*fi.nx) div gx - 1;
        ya:=(yi*fi.ny) div gy;
        yz:=((yi+1)*fi.ny) div gy - 1;
        x:=-1;  y:=-1;
        r:=0;
        for yi:=ya to yz do
        begin
            d:=fi.pt + yi*fi.sl;
            for xi:=xa to xz do
            begin
                s:=d[xi];
                if s>r then
                begin
                    r:=s;
                    x:=xi;
                    y:=yi;
                end;
            end;
        end;
        r-=offs;
        stars[i].x:=x;
        stars[i].y:=y;
        stars[i].b:=r;
        stars[i].v:=(x>=mbd) and (y>=mbd) and (x<fi.nx-mbd) and (y<fi.ny-mbd) and (r>minb);
    end;
end;

{ calculate FWHM's }
threadprocedure SImg_FindStars_thread2(ti:longint; tn:longint; fi:TSImg_FPImage; gx:integer; gy:integer; stars:PSImg_FindStars_peak; offs:simgfp; sat:simgfp);
const k_fact = 2/pi;
      limit  = exp(ln(0.5)*4/(pi*k_fact));  { 0.25 @ k_fact=2/pi }
var i,x,y,ia,iz         : longint;
    j,k                 : int64;
    xi,xa,xz,yi,ya,yz   : longint;
    nx,ny               : longint;
    r                   : double;
    d                   : ^simgfp;
begin
    ia:=(gx*gy*ti) div tn;
    iz:=(gx*gy*(ti+1)) div tn - 1;
    nx:=fi.nx;
    ny:=fi.ny;
    d:=fi.pt;
    for i:=ia to iz do
        if stars[i].v  then
        begin
            x:=stars[i].x;
            y:=stars[i].y;
            d:=fi.pt + y*nx + x;
            { search boundaries }
            if stars[i].s then r:=sat*limit*0.5
            else r:=stars[i].b*limit*0.75+offs;
            xa:=0;
            while ((x+xa)>0) and (d[xa]>r) do
                xa-=1;
            xz:=0;
            while ((x+xz)<nx-1) and (d[xz]>r) do
                xz+=1;
            ya:=0;
            while ((y+ya)>0) and (d[ya*nx]>r) do
                ya-=1;
            yz:=0;
            while ((y+yz)<ny-1) and (d[yz*nx]>r) do
                yz+=1;

            { search boundaries }
            d+=ya*nx;
            if stars[i].s then r:=sat*limit*0.8+offs
            else r:=stars[i].b*limit+offs;
            j:=0; k:=0;
            for yi:=ya to yz do
            begin
                for xi:=xa to xz do
                    if d[xi]>r then
                    begin
                        j+=sqr(yi)+sqr(xi);
                        k+=1;
                    end;
                d+=nx;
            end;
//          if (k<0) or (j<0) then simg_error('k='+int2str(k)+' j='+int2str(j)+' xa='+int2str(ya)+' xz='+int2str(yz)+' ya='+int2str(ya)+' yz='+int2str(yz));
            stars[i].w:=sqrt(k*k_fact);
            stars[i].a:=sqrt((6*j+0.5)/(sqr(k)+0.5));
//          if (abs(x-4502)<10) and (abs(y-8693)<10) then with stars[i] do writeln(stderr,'x=',x,' y=',y,' s=',s,' w=',w:0:2,' a=',a:0:2);
//          if stars[i].s then writeln(stderr,stars[i].x,'  ',stars[i].y,'  ',stars[i].a:0:3);
        end
        else begin
            stars[i].w:=1;
            stars[i].a:=1;
        end;
end;

{ Find save stars
    fi     : input image (1 color image)
    stars  : output array, sorted by brightness
    tol    : FWHM tolerance
    rad    : Minimumn distance to boundaries.
    result : FWHM
}
function SImg_FindStars(cc:CSImg_Config; sh:TSImg_StarSubtract_Shared; var stars:array of TSImg_FindStars_peak; tol:double; rad:longint):double;
const fwhm_its  = 3;     { FWHM iterations }
      minb_fact = 15;    { minimum brightness relative to noise floor }
var i,xi,yi,gx,gy,tn,it : longint;
    n0,n1               : longint;
    a,minb              : simgfp;
    r,r1,t              : double;
    ss,s                : PSImg_FindStars_peak;
//ifdef[debug_findStars]
    fd                  : TSImg_FPImage;
//endif
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,sh.fi2) then cc.error('SImg_FindStars: Invalid image: sh.fi2');
    if sh.fi2.nc<>1 then cc.error('SImg_FindStars: Invalid dimensions: sh.fi2');
    if tol<1.1 then tol:=1.1;
    minb:=sh.sig*minb_fact;
    cc.info(cc.name+': Searching bright stars');

    { grid size }
    r:=sqrt(sh.fi2.size/(high(stars)+1));
    gx:=round(3*sh.fi2.nx/r)+10;
    gy:=round(3*sh.fi2.ny/r)+10;
    getmem2(ss,gx*gy*sizeof(TSImg_FindStars_peak));

    { find local maxima }
    tn:=cc.threads;
    for i:=0 to tn-1 do
        dothread SImg_FindStars_thread1(i,tn, sh.fi2, gx,gy, minb, sh.offs, ss);
    threadwait;
    for yi:=0 to gy-1 do
        for xi:=0 to gx-1 do
        begin
            s:=ss+yi*gx+xi;
            a:=s[0].b;
            if (yi>0   ) and (xi>0   ) and (a>s[-gx-1].b) then s[-gx-1].v:=false;
            if (yi>0   ) and               (a>s[-gx  ].b) then s[-gx  ].v:=false;
            if (yi>0   ) and (xi<gx-1) and (a>s[-gx+1].b) then s[-gx+1].v:=false;
            if               (xi>0   ) and (a>s[-1   ].b) then s[-1   ].v:=false;
            if               (xi<gx-1) and (a>s[ 1   ].b) then s[ 1   ].v:=false;
            if (yi<gy-1) and (xi>0   ) and (a>s[ gx-1].b) then s[ gx-1].v:=false;
            if (yi<gy-1) and               (a>s[ gx  ].b) then s[ gx  ].v:=false;
            if (yi<gy-1) and (xi<gx-1) and (a>s[ gx+1].b) then s[ gx+1].v:=false;
            s[0].s:=a>sh.sat;
        end;

    { disable stars which are to close to borders }
    for i:=0 to gx*gy-1 do
        with ss[i] do
            if (x<rad) or (y<rad) or (x+rad>=sh.fi2.nx) or (y+rad>=sh.fi2.ny) then v:=false;

    { calculate fwhm's }
    for i:=0 to tn-1 do
        dothread SImg_FindStars_thread2(i,tn, sh.fi2, gx,gy, ss, sh.offs, sh.sat);
    threadwait;
    xi:=0; r:=0; n0:=0;
    for i:=0 to gx*gy-1 do
        with ss[i] do
        begin
            if not s then
            begin
//              if v then writeln(stderr,x,', ',y,':   w=',w:0:3,'  a=',a:0:3);
//              v:=v and (sqr(a*w)+1<sqr(tol)*(sqr(w)+1));      { FWHM regularization with 1 pixel in order to avoid problems due to discretization }
                if v then
                begin
                    r+=w;
                    xi+=1;
                end;
            end
//          else if sqr(a-0.1)>tol then v:=false;
            else if a>tol then v:=false;
            if v then n0+=1;
        end;
    for it:=0 to fwhm_its do
    begin
        if xi<1 then cc.error('No valid and non-saturated sources found within tolerance');
        r/=xi;
        t:=exp(ln(tol+0.7/r)*(2-it/fwhm_its));
//ifdef[debug_findStars]
        cc.info(cc.name+': tol='+real2str(t,-3)+'   fwhm='+real2str(r,-2)+'   n='+int2str(xi));
//endif
//      r0:=sqr(r/t);
//      if r0<0 then r0:=0 else r0:=sqrt(r0);
        r1:=r*t;
        r:=0; xi:=0;
        for i:=0 to gx*gy-1 do
            with ss[i] do
                if not s then
                begin
//                  if v and (w>=r0) and (w<=r1) then
                    if v and (w*a<=r1) then
                    begin
                        r+=w;
                        xi+=1;
                    end
                    else if it=fwhm_its then v:=false;
//                  if v and (abs(x-4502)<10) and (abs(y-8693)<10) then writeln(stderr,'x=',x,' y=',y,' s=',s,' w=',w:0:2,' a=',a:0:2,' r1=',r1:0:1,' b=',b);
                end;
    end;
    result:=r/xi;

    { sort output }
    n1:=n0;
    if high(stars)>=0 then
    begin
        xi:=-1; { idx last valid element }
        for i:=0 to gx*gy-1 do
            with ss[i] do
                if v then
                begin
                    n1-=1;
                    a:=b;
                    if xi<high(stars) then xi+=1
                    else if a<=stars[xi].b then continue;
                    yi:=xi;
                    while (yi>0) and (a>stars[yi-1].b) do
                    begin
                        stars[yi]:=stars[yi-1];
                        yi-=1;
                    end;
                    stars[yi]:=ss[i];
                end;
        for i:=xi+1 to high(stars) do
            stars[i].v:=false;
    end;
    cc.info(cc.name+': Filtered '+real2str((n1/n0)*100,-1)+'% ('+int2str(n1)+' / '+int2str(n0)+') sources.');

//ifdef[debug_findStars]
{    if fi2.size<50e6 then
    begin
        fd.pt:=nil;
        SImg_New(cc,fd, fi2.nx,fi2.ny,2);
        for i:=0 to gx*gy-1 do
            if ss[i].v then fd.pt[ ss[i].y*fd.sl + ss[i].x*2]:=ss[i].b;
        for i:=0 to high(stars)-1 do
            if stars[i].v then fd.pt[ stars[i].y*fd.sl + stars[i].x*2 + 1]:=stars[i].b;
        SImg_Filter(cc,fd,fd,result*FWHM_to_rad);
        SImg_WriteImageFile(cc,'findStars-1.fits',fd);

        SImg_Free(cc,fd);
    end;}
//endif

    freemem2(ss,gx*gy*sizeof(TSImg_FindStars_peak));
end;


{*************************************************************************}
{******* SImg_GetPSF *****************************************************}
{*************************************************************************}
{ get star background using weighted average between 40% and 60% quntiles
  ti     : thread index
  tn     : number of threads
  fi2    : input image
  fw     : weight image
  stars  : stars
  n      : number of stars
  rad    : PSF radius
  rad2   : extended PSF radius used for background detection
}
threadprocedure SImg_GetPSF_getBgMed(ti:longint; tn:longint; fi2:TSImg_FPImage; fw:TSImg_FPImage; stars:PSImg_FindStars_peak; n:longint; rad:longint; rad2:longint);
const hist_n = 512;
var hist                : array[0..hist_n,0..1] of simgfp;  { [*,0]: weight; [*,1]:value*weight }
    s                   : TSImg_FindStars_peak;
    i,j,k,l,ia,iz,x,y   : longint;
    sr,ry               : longint;
    mi,ma,a,b           : simgfp;
    r,t,u,rrd           : double;
begin
    ia:=(ti*n) div tn;
    iz:=((ti+1)*n) div tn - 1;
    sr:=sqr(rad);
    rrd:=1/(rad2+1-rad);
    for i:=ia to iz do
    begin
        s:=stars[i];
        { find minimum / maximum }
        mi:=simg_large;
        ma:=-simg_large;
        for y:=-rad2 to rad2 do
        begin
            k:=s.y+y;
            if (k<0) or (k>=fi2.ny) then continue;
            ry:=round(sqrt(sqr(rad2+1)-sqr(y)-0.51));
            j:=(s.y+y)*fi2.sl + s.x;
            for x:=-ry to ry do
            begin
                k:=s.x+x;
                if (k<0) or (k>=fi2.nx) then continue;
                k:=sqr(x)+sqr(y);
                if k>sr then
                begin
                    a:=fi2.pt[j+x];
                    if a<mi then mi:=a;
                    if a>ma then ma:=a;
                end;
            end;
        end;
        ma-=mi;
        if ma<simg_mina then    { should not happen }
        begin
            writeln(stderr,'SImg_GetPSF_getBgMed: internal error');
            stars[i].b:=0;
            continue;
        end;
        ma:=hist_n/ma;

        { calculate histogramm }
        for l:=0 to hist_n do
        begin
            hist[l,0]:=0;
            hist[l,1]:=0;
        end;
        for y:=-rad2 to rad2 do
        begin
            k:=s.y+y;
            if (k<0) or (k>=fi2.ny) then continue;
            ry:=round(sqrt(sqr(rad2+1)-sqr(y)-0.51));
            j:=(s.y+y)*fi2.sl + s.x;
            for x:=-ry to ry do
            begin
                k:=s.x+x;
                if (k<0) or (k>=fi2.nx) then continue;
                k:=sqr(x)+sqr(y);
                if k>sr then
                begin
                    a:=(1-sqr((sqrt(k)-rad)*rrd))*fw.pt[j+x];
                    b:=fi2.pt[j+x];
                    l:=round((b-mi)*ma);        { range check not required }
                    hist[l,0]+=a;
                    hist[l,1]+=a*b;
                end;
            end;
        end;

        { average between 40% and 60% quantiles }
        r:=0;
        for l:=0 to hist_n do
            r+=hist[l,0];
        mi:=r*0.4;
        ma:=r*0.6;
        r:=hist[0,0]; l:=0;
        u:=0; t:=0;
        while (l<hist_n) and (r<ma) do
        begin
            l+=1;
            r+=hist[l,0];
            if r>mi then
            begin
                t+=hist[l-1,0];
                u+=hist[l-1,1];         { quantile of first quantization interval is less than 0.4 }
            end;
        end;
        t+=hist[l,0];
        u+=hist[l,1];                   { quantile of last quantization interval is larger than 0.6 }

        stars[i].b:=u/t;
    end;
end;

{******* SImg_GetPSF_getdat **********************************************}
{ extract normalized data from image
  ti    : thread index
  tn    : number of threads
  fi2   : input image
  fw    : weight image
  fs    : reference PSF
  stars : stars
  dat   : output buffer: n * 2 * sqr(2*rad+1) * 4; data samples:
          0. normalized brightness fi2/r where r is the approximation factor such that fi2(x-y)-bg(y) \approx r(y)*fs(x) where bg is the background by SImg_GetPSF_getBgMed
          1. fw(x-y) * r(y)^wex
          2. deviation weight (parameters sigr and sigs)
          3. r(y)
  n     : number of stars
  rad   : PSF radius
  sat   : saturation / undefined limit
  offs  : noise offset
  sig   : noise level of fi2
  rd    : relative deviation, >=0, recommended: 0..1
  sym   : symmetry enforcement: 0 (unsymmetrical) .. 1 (symmetrical)
  wex   : brightness exponent (optimal SNR at 1)
  isat  : ignore saturated stars
  ao    : offset adjustment factor (0..1)
}
threadprocedure SImg_GetPSF_getdat(ti:longint; tn:longint; fi2:TSImg_FPImage; fw:TSImg_FPImage; fs:TSImg_FPImage; stars:PSImg_FindStars_peak; dat:psimgfp; n:longint; rad:longint; sat:simgfp; sig:double; rd:double; sym:double; wex:double; isat:boolean; ao:double);
const min_re = 0.1;     { minimum relative error }
      fact_var = 0.9;   { if approximation factor changes by fact_var or 1/fact_var from step 1 to step 2, weight is halved, ignored if isat=true }
      var_lamb = ln(0.5)/sqr(ln(fact_var));
var s                   : TSImg_FindStars_peak;
    a,b                 : simgfp;
    q,r,t,u,w,rr,rsig   : double;
    a11,a12,a22         : double;
    sym2,df             : double;
    i,j,ia,iz,x,y,en,ry : longint;
    k                   : longint;
    d,e,f,g             : ^simgfp;
label ignore;
begin
    ia:=(ti*n) div tn;
    iz:=((ti+1)*n) div tn - 1;
    en:=sqr(rad*2+1)*4;
    sig*=2;
    rsig:=1/(2*sig);
    if rd<0 then rd:=0;
    if sym<0 then sym:=0;
    if sym>1 then sym:=1;
    sym2:=2+min_re-sym+rd;
    rd:=min_re+2*rd;
    for i:=ia to iz do
    begin
        s:=stars[i];
        if isat and s.s then goto ignore;
        j:=s.y*fi2.sl + s.x;
        d:=fi2.pt + j;
        e:=fw.pt  + j;
        j:=rad*fs.sl + rad;
        f:=fs.pt  + j;
        g:=dat + i*2*en + j*4;

        { get factor and offset: step 1: without deviation weight }
        q:=0; r:=0;
        a11:=0; a12:=0; a22:=0;
        for y:=-rad to rad do
        begin
            ry:=trunc(sqrt(sqr(rad)-sqr(y)));
            for x:=-ry to ry do
            begin
                { distance factor }
                df:=1/(0.25+sqrt(sqr(y)+sqr(x)));
                { value and weight }
                j:=y*fi2.sl+x;
                t:=d[j]-s.b;
                w:=e[j];
                { symmetry weight }
                u:=t - (d[-j]-s.b)*(sym2+df);
                if u>0 then w/=1+sqr(u*rsig);
                { scalar products }
                a:=f[y*fs.sl+x];
                a11+=sqr(a)*w;
                a12+=a*w;
                a22+=w;
                q+=t*a*w;
                r+=t*w;
            end;
        end;
        { only lower the background because it may be estimated to high }
        t:=a11*a22-sqr(a12);
        if (abs(t)<simg_mina) or (a11<simg_mina) then goto ignore;
        t:=ao*(a11*r-a12*q)/t;  { offset }
        if t>0 then t:=0;
        s.b+=t;
        b:=(q-t*a12)/a11;
        if b<simg_mina then goto ignore;
        rr:=b;

        { get factor: step 2: with deviation weight. No offset adjustment here because that may be unstable }
        for k:=1 to 2 do
        begin
            q:=0; a11:=0;
            for y:=-rad to rad do
            begin
                ry:=trunc(sqrt(sqr(rad)-sqr(y)));
                for x:=-ry to ry do
                begin
                    { distance factor }
                    df:=1/(0.25+sqrt(sqr(y)+sqr(x)));
                    { value }
                    j:=y*fi2.sl+x;
                    t:=d[j]-s.b;
                    w:=e[j];
                    { deviation and symmetry weight }
                    u:=t - (d[-j]-s.b)*(sym2+df);
                    if u<0 then u:=0
                    else u:=sqr(u*rsig);
                    a:=f[y*fs.sl+x];
                    if a<0 then a:=0;
                    w/=(1+u+sqr((a*b-t)/(sig+a*b*(rd+df))));
                    { scalar products }
                    a11+=sqr(a)*w;
                    q+=t*a*w;
                end;
            end;
            if (a11<simg_mina) or (q<simg_mina) then goto ignore;
            b:=q/a11;
        end;

        { extract data }
        r:=b;
        b:=ln(b/rr);
        if isat then b*=0.5;
        b:=exp(sqr(b)*var_lamb);
        q:=exp(ln(r)*wex);
        rr:=1/r;
        for y:=-rad to rad do
        begin
            for x:=-rad to rad do
            begin
                { distance factor }
                df:=1/(0.25+sqrt(sqr(y)+sqr(x)));
                { value }
                j:=y*fi2.sl+x;
                t:=d[j]-s.b;
                w:=q*e[j];
                { deviation and symmetry weight }
                u:=t - (d[-j]-s.b)*(sym2+df);
                if u<0 then u:=0
                else u:=sqr(u*rsig);
                j:=y*fs.sl+x;
                a:=f[j];
                if a<0 then a:=0;
                u:=b/(1+u+sqr((a*r-t)/(sig+a*r*(rd+df))));
                { save data }
                j*=4;
                t*=rr;

                g[j     ]:=t;
                g[j+1   ]:=w;
                g[j+2   ]:=u;
                g[j+3   ]:=r;

                g[en-j  ]:=t;
                g[en-j+1]:=w;
                g[en-j+2]:=u*sym;
                g[en-j+3]:=r;
            end;
        end;
        continue;

ignore:
        g:=dat + i*2*en;
        for x:=0 to 2*en-1 do
            g[x]:=0;

    end;
end;

procedure SImg_GetPSF_getdat(cc:CSImg_Config; fi2,fw,fs:TSImg_FPImage; stars:PSImg_FindStars_peak; dat:psimgfp; n,rad:longint; sat:simgfp; sig,rd,sym,wex:double; isat: boolean; ao:double);
var ti  : longint;
threadvars;
begin
    for ti:=0 to cc.threads-1 do
        dothread SImg_GetPSF_getdat(ti,cc.threads, fi2,fw,fs, stars, dat, n,rad, sat,sig, rd,sym, wex, isat,ao);
    threadwait;
end;


{******* SImg_GetPSF_quant ***********************************************}
{ calculate psf using averages between quantiles qmin and qmax
  ti    : thread index
  tn    : number of threads
  res   : m; output PSF
  resw  : m; output PSF weight
  dat   : n x m x 4; data samples ( brightness + 3 weight components, see SImg_GetPSF_getdat )
  qmin  : lower quantile
  qmax  : upper quantile
}
threadprocedure SImg_GetPSF_quant(ti:longint; tn:longint; res:psimgfp; resw:psimgfp; dat:psimgfp; n:longint; m:longint; qmin:double; qmax:double);
const hist_n = 512;
      wl_n   = 4;       { limit weight to the w_n'th largest value }
var hist        : array[0..hist_n,0..2] of simgfp;  { [*,0]: weight; [*,1]:value*weight }
    wl          : array[0..wl_n-1] of simgfp;
    q,r,s,t     : double;
    a,b,c,ra,rz : simgfp;
    i,j,k,ja,jz : longint;
    hi          : longint;
    d           : ^simgfp;

begin
    ja:=m*ti div tn;
    jz:=m*(ti+1) div tn-1;
    for j:=ja to jz do
    begin
        { get min/max and weight limit }
        for i:=0 to wl_n-1 do
            wl[i]:=simg_mina;
        ra:=simg_large; rz:=-simg_large;
        d:=dat+4*j;
        for i:=0 to n-1 do
        begin
            if d[1]>simg_mina then
            begin
                b:=d[1]*d[2];
                if b>wl[0] then
                begin
                    k:=0;
                    while (k<wl_n-1) and (b>wl[k+1]) do
                    begin
                        k+=1;
                        wl[k-1]:=wl[k];
                    end;
                    { (k==wl_n-1 || b<wl[k+1] ) && (wl[k-1]==wl[k]) };
                    wl[k]:=b;
                end;
                a:=d[0];
                if a<ra then ra:=a;
                if a>rz then rz:=a;
            end;
            d+=4*m;
        end;

        rz-=ra;
        if rz<simg_mina then
        begin
            res[j]:=0;
            resw[j]:=0;
        end
        else begin
            { get histogram }
            d:=dat+4*j;
            rz:=hist_n/rz;
            for i:=0 to hist_n do
            begin
                hist[i,0]:=0;
                hist[i,1]:=0;
                hist[i,2]:=0;
            end;
            for i:=0 to n-1 do
            begin
                a:=d[0];
                c:=d[2];
                b:=d[1]*c;
                if b>wl[0] then b:=wl[0];
                b*=d[3];
                s:=(a-ra)*rz;
                if s<0 then s:=0;
                if s>hist_n then s:=hist_n;
                hi:=round(s);
                hist[hi,0]+=b;
                hist[hi,1]+=a*b;
                hist[hi,2]+=c*b;
                d+=4*m;
            end;

            { average between quantiles }
            r:=0;
            for i:=0 to hist_n do
                r+=hist[i,0];
            ra:=r*qmin;
            rz:=r*qmax;
            q:=hist[0,0]; i:=0;
            r:=0; s:=0; t:=0;
            while (i<hist_n) and (q<rz) do
            begin
                i+=1;
                q+=hist[i,0];
                if q>ra then
                begin
                    s+=hist[i-1,0];             { quantile of first quantization interval is less than qmin }
                    r+=hist[i-1,1];
                    t+=hist[i-1,2];
                end;
            end;
            s+=hist[i,0];                       { quantile of last quantization interval is larger than qmax }
            s:=1/s;
            res [j]:=(r+hist[i,1])*s;
            resw[j]:=(t+hist[i,2])*s;
//          if s<=simg_mina then writeln(stderr,'qa=',qa,'   qz=',qz,'   q=',q,'   r=',r,'  s=',s);
        end;
    end;
end;

procedure SImg_GetPSF_quant(cc:CSImg_Config; res,resw,dat:psimgfp; n,m:longint; qmin,qmax:double);
var ti,tn       : longint;
threadvars;
begin
    tn:=cc.threads;
    for ti:=0 to tn-1 do
        dothread SImg_GetPSF_quant(ti,tn, res,resw, dat, n,m, qmin,qmax);
    threadwait;
end;


{******* SImg_GetPSF_normalize_smooth *********************************}
{ PSF smoothing algorithm. Undefined region is extended by linear approximation in order to avoid artifacts at boundary.
 offset subtraction, denoising, normalization
  bufi  : input buffer
  bufo  : output buffer
  rad   : radius
  clear : clear bufo
  smth  : smoothing radius relative to radius
  result: gradient at border
}
function SImg_GetPSF_normalize_smooth(bufi,bufo:pdouble; rad:longint; clear:boolean; smth:double):double;
var i,j,k               : longint;
    l                   : int64;
    a,b,r,s,rr,t        : double;
begin
    if clear then
        for i:=0 to rad do
            bufo[i]:=0;

    { linear approximation extension }
    j:=round(rad*smth-0.5001);
    if j<1 then
    begin
        for i:=0 to rad do
            bufo[i]+=bufi[i];
        result:=0;
        exit;
    end;
    if j>rad then j:=rad;
    a:=0; l:=0;
    b:=0;
    for i:=0 to j do
    begin
        k:=j-2*i;
        r:=bufi[rad-i];
        a+=r*k; l+=sqr(k);
        b+=r;
    end;
    a/=0.5*l;
    b:=b/(j+1) + 0.5*j*a;
    result:=a;
//    writeln(stderr,'a='+real2str(a)+' b='+real2str(b)+' bufi[rad]='+real2str(bufi[rad]));

    { smoothing }
    for i:=0 to rad do
    begin
        j:=round(i*smth-0.5001);
        if j<1 then
        begin
            bufo[i]+=bufi[i];
            continue;
        end;
        rr:=1/(i*smth);
        if j>i then j:=i;
        r:=0; s:=0;
        for k:=i-j to i+j do
        begin
            t:=1-abs(i-k)*rr;
            if t>0.5 then t:=0.5;
            if k>rad then r+=(b+(i-rad)*a)*t
            else r+=bufi[k]*t;
            s+=t;
        end;
        bufo[i]+=r/s;
    end;
end;


{******* SImg_GetPSF_normalize ****************************************}
{ offset subtraction, denoising, normalization
  fs    : PSF
  fsw   : PSF weight
  rad   : radius
}
procedure SImg_GetPSF_normalize(cc:CSImg_Config; fs,fsw:TSImg_FPImage; rad:longint; dbg:boolean);
const asmth = 5;       { angle smoothing }
      rsmth = 0.15;    { fixed radial smoothing }
var rbuf,rbuf1          : ^double;
    abuf,abuf1          : ^double;
    x,y,i,j,k,na        : longint;
    sr                  : longint;
    r,s,a1,a2,b1,rr,rrd : double;
    d,e                 : ^simgfp;
//ifdef[debug_getpsf]
    fa                  : TSImg_FPImage;
//endif

begin
    sr:=sqr(rad);

    { get average per radius }
    getmem2(rbuf,(rad+1)*2*sizeof(double));
    rbuf1:=rbuf +    rad+1;
    rr:=1/rad;
    for i:=0 to 2*rad+1 do
        rbuf[i]:=0;
    for y:=-rad-1 to rad+1 do
    begin
        e:=fs.pt + (y+rad)*fs.sl + rad;
        d:=fsw.pt + (y+rad)*fs.sl + rad;
        for x:=-rad-1 to rad+1 do
        begin
            i:=round(sqrt(sqr(x)+sqr(y)));
            if i>rad then continue;
            r:=d[x];
            if r<simg_mina then r:=simg_mina;
            rbuf[i*2  ]+=e[x]*r;
            rbuf[i*2+1]+=r;
        end;
    end;
    for i:=0 to rad do
        rbuf[i]:=rbuf[i*2]/rbuf[i*2+1];

    { enforce monotony }
    s:=0.02;
    repeat
        a2:=rbuf[rad];
        for i:=rad downto 0 do
        begin
            a1:=rbuf[i];
            if a1>a2 then a2:=a1;
            rbuf1[i]:=0.5*(a1-a2);
        end;
        a2:=rbuf[0];
        for i:=0 to rad do
        begin
            a1:=rbuf[i];
            if a1<a2 then a2:=a1;
            rbuf1[i]+=0.5*(a1-a2);
        end;
        a2:=0;
        for i:=0 to rad do
        begin
            a1:=rbuf1[i];
            rbuf[i]-=a1;
            a1:=abs(a1);
            if a1>a2 then a2:=a1;
        end;
        SImg_GetPSF_normalize_smooth(rbuf1,rbuf, rad, false, s);
        s*=1.2;
//      cc.info(real2str(a2)+'   '+real2str(s));
    until (s>4) or (a2<1e-9*rbuf[0]);

    { fixed smoothing + fixing offset }
    SImg_GetPSF_normalize_smooth(rbuf,rbuf1, rad, true, rsmth);
    SImg_GetPSF_normalize_smooth(rbuf1,rbuf, rad, true, rsmth*0.5);
    SImg_GetPSF_normalize_smooth(rbuf,rbuf1, rad, true, rsmth*0.25);
    r:=SImg_GetPSF_normalize_smooth(rbuf1,rbuf, rad, true, rsmth*0.125);
    r+=rbuf[rad];
    for i:=0 to rad do
        if rbuf[i]<r then r:=rbuf[i];
    for i:=0 to rad do
        rbuf[i]-=r;
    for i:=0 to fs.size-1 do
        fs.pt[i]-=r;
//    cc.info('offset: '+real2str(r));

    { get noise }
    a1:=0; a2:=0;
    b1:=rbuf[0]*1e-5;
    for y:=-rad to rad do
    begin
        d:=fs.pt + (y+rad)*fs.sl + rad;
        e:=fsw.pt + (y+rad)*fs.sl + rad;
        for x:=-rad to rad do
        begin
            i:=sqr(x)+sqr(y);
            if i>sr then continue;
            s:=d[x];
            if (s>b1) and (i*3<sr*2) then continue;
            i:=round(sqrt(i));
            r:=e[x];
            if r<simg_mina then r:=simg_mina;
            a1+=sqr(rbuf[i]-s)*r;
            a2+=r;
        end;
    end;
    b1:=sqrt(a1/a2);
//ifdef[debug_getpsf]
//  cc.info('psf noise: '+real2str(b1));
//endif

    { noise weight }
    for i:=0 to rad do
    begin
        r:=rbuf[i];
        rbuf1[i]:=r/(r+b1);
    end;

    { averages per angle }
    na:=rad*4;
    getmem2(abuf, na*2*sizeof(double));
    abuf1:=abuf+na;
    for i:=0 to na-1 do
        abuf[i]:=0;
    rr:=pi/na;
    d:=fs.pt + rad*fs.nx + rad;
    for i:=0 to na-1 do
    begin
        r:=i*rr;
        a1:=cos(r)*0.5;
        a2:=sin(r)*0.5;
        s:=0;
        for j:=(rad+8) div 16 to 2*rad do
        begin
            k:=round(j*a2)*fs.nx + round(j*a1);
            s+=(d[k]+d[-k]) * rbuf1[j div 2]*j;
        end;
        abuf[i]:=s;
    end;
    s:=0;
    for i:=0 to na-1 do
        s+=abuf[i];
    r:=s/na;
    s:=0;
    for i:=0 to na-1 do
    begin
        if abuf[i]<r then abuf[i]:=r
        else abuf[i]+=abuf[i]-r;
        s+=abuf[i];
    end;
    r:=na/((2*asmth+1)*s);
    for i:=0 to na-1 do
    begin
        s:=0;
        for j:=-asmth to asmth do
        begin
            k:=i+j;
            if k<0 then k+=na;
            if k>=na then k-=na;
            s+=abuf[k];
        end;
        abuf1[i]:=r*s;
    end;

    { apply constraints and output }
//ifdef[debug_getpsf]
    fa.pt:=nil;
    SImg_New(cc, fa, fs.nx,fs.ny,1);
    if dbg then SImg_WriteImageFile(cc,'getpsf_normalize-fs0.fits',fs);
//endif
    rr:=na/pi;
    rrd:=1/rbuf[0];
    for y:=-rad to rad do
    begin
        i:=(y+rad)*fs.sl + rad;
        d:=fs.pt  + i;
//ifdef[debug_getpsf]
        e:=fa.pt + i;
//endif
        for x:=-rad to rad do
        begin
            i:=sqr(x)+sqr(y);
            if i>sr then
            begin
                d[x]:=0;
//ifdef[debug_getpsf]
                e[x]:=0;
//endif
            end
            else begin
                i:=round(sqrt(i));
                r:=rbuf[i];
                j:=round(arctan2(y,x)*rr) mod na;
                if j<0 then j+=na;
                s:=sqr(sqr(1-r*rrd));
                r*=1-s+s*abuf1[j];
                a2:=rbuf1[i]*r;
                a1:=r-a2;
                a2:=r+a2;
                s:=d[x];
                if s<a1 then d[x]:=a1
                else if s>a2 then d[x]:=a2
                else d[x]:=s;
//ifdef[debug_getpsf]
                e[x]:=r;
//endif
            end;
        end;
    end;
    freemem2(rbuf,(rad+1)*2*sizeof(double));
    freemem2(abuf, na*2*sizeof(double));

    { normalize }
    r:=0;
    for i:=0 to fs.size-1 do
        r+=fs.pt[i];
//    cc.info('r='+real2str(r));
    r:=1/r;
    for i:=0 to fs.size-1 do
        fs.pt[i]*=r;

//ifdef[debug_getpsf]
    if dbg then
    begin
        SImg_WriteImageFile(cc,'getpsf_normalize-fs1.fits',fs);
        SImg_WriteImageFile(cc,'getpsf_normalize-fa.fits',fa);
    end;
    SImg_Free(cc,fa);
//endif
end;

{******* SImg_GetPSF *****************************************************}
{ Calculate PSF
    sh     : data shared with other routines
    fpsf   : reference psf (input, optional) and output PSF
    tol    : FWHM tolerance as factor
    rad    : PSF radius
    rm     : star mask. Or'ed with parameter mask at positions found by SImg_FindStars (optional)
    mask   : see rm
    sym    : symmetry enforcement (0..1)
    rrad   : use reference psf above this radius. Ignored if no reference psf is given. Or if <= 0
}
procedure SImg_GetPSF(cc:CSImg_Config; sh:TSImg_StarSubtract_shared; var fpsf:TSImg_FPImage; tol:double; rad:longint; rm:TSImg_RDImage; mask:byte; sym,rrad:double);
const min_stars = 100;  { minimum desired number of stars for PSF calculation, half as much if reference PSF is given and rrad is used }
      minb      = 25;   { minimum brightness relative to noise floor }
      itn       = 4;    { psf extraction iterations }
var fs,fsw      : TSImg_FPImage;
    stars       : array of TSImg_FindStars_peak;
    n,m         : longint;      { see below }
    dat         : ^simgfp;      { n x m x 4; data samples ( brightness + 3 weight components, see SImg_GetPSF_getdat ) }
    fwhm,r,t,u  : double;
    i,x,y       : longint;
    d,e         : ^simgfp;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_GetPSF: sh.fi2='+SImg_InfoStr2(sh.fi2)+' sh.fw='+SImg_InfoStr2(sh.fw)+' fpsf='+SImg_InfoStr2(fpsf)+' tol='+real2str(tol)+' rad='+int2str(rad)+' rm='+SImg_InfoStr2(rm)+' mask='+int2str(mask)+' sym='+real2str(sym)+' rrad='+real2str(rrad));
    if not SImg_Recover(cc,sh.fi2) then cc.error('SImg_GetPSF: Invalid image: sh.fi2');
    if sh.fi2.nc<>1 then cc.error('SImg_GetPSF: Invalid dimensions: sh.fi2');
    if not SImg_Recover(cc,sh.fw) then cc.error('SImg_GetPSF: Invalid image: sh.fw');
    if (sh.fi2.nx<>sh.fw.nx) or (sh.fi2.ny<>sh.fw.ny) or (sh.fw.nc<>1) then cc.error('SImg_GetPSF: Invalid dimensions: sh.fw');
    if fpsf.pt<>nil then
    begin
        if not SImg_Recover(cc,fpsf) then cc.error('SImg_GetPSF: Invalid reference PSF');
        if (fpsf.nx<>rad*2+1) or (fpsf.ny<>rad*2+1) or (fpsf.nc<>1) then cc.error('SImg_GetPSF: Invalid dimensions: fpsf');
    end;
    if mask=0 then rm.pt:=nil;
    if rm.pt<>nil then
    begin
        if not SImg_Recover(cc,rm) and (rm.fmt<>SImg_RawByte) then cc.error('SImg_GetPSF: Invalid image: rm');
        if rm.size<>sh.fi2.size then cc.error('SImg_GetPSF: Invalid dimensions: rm');
    end;

    if tol<1.1 then tol:=1.1;
    m:=sqr(2*rad+1);

    { number od stars }
    if fpsf.pt<>nil then
    begin
        if rrad<1e-6 then r:=1
        else begin
            r:=rrad*0.75;
            if r>1 then r:=1;
            if r<0.25 then r:=0.25;
        end;
        n:=round( sh.fi2.size / sqr(rad*2*r+1) );
        i:=min_stars div 2;
    end
    else begin
        i:=min_stars;
        n:=sh.fi2.size div sqr(rad*2+1);
    end;
    if n*4<i then n:=i
    else if n<i*4 then n:=round(sqrt(n*i*4));

    { find stars }
    setlength(stars,n);
    fwhm:=SImg_FindStars(cc, sh, stars, tol, rad);
    while (n>0) and not stars[n-1].v do
        n-=1;
    cc.info(cc.name+': Found '+int2str(n)+' sources with average FWHM '+real2str(fwhm,-2)+'.');

    { calculate star background }
    for i:=0 to cc.threads-1 do
       dothread SImg_GetPSF_getBgMed(i,cc.threads, sh.fi2, sh.fw, @stars[0], n, rad,round(rad*1.4+1));
    threadwait;

    { save stars }
    if (rm.pt<>nil) and (mask<>0) then
    begin
        for i:=0 to n-1 do
        begin
            x:=stars[i].y*rm.sl + stars[i].x;
            rm.ptb[x]:=rm.ptb[x] or mask;
        end;
    end;

    fs.pt:=nil;  fsw.pt:=nil;
    SImg_New(cc, fs,  2*rad+1, 2*rad+1,1 );
    SImg_New(cc, fsw, 2*rad+1, 2*rad+1,1 );
    if fpsf.pt<>nil then
    begin
        if rrad>1e-8 then rrad:=sqr(1/(rad*rrad));
        r:=0;
        d:=fpsf.pt;
        for x:=0 to fpsf.size-1 do
            r+=d[x];
        if r<simg_mina then cc.error('SImg_GetPSF: Non-positive reference PSF');
        r:=1/r;
        for y:=-rad to rad do
        begin
            d:=fs.pt  +(y+rad)*fs.sl   + rad;
            e:=fpsf.pt+(y+rad)*fpsf.sl + rad;
            for x:=-rad to rad do
            begin
                t:=e[x]*r;
                e[x]:=t;
                d[x]:=t;
            end;
        end;
        for x:=0 to fsw.size-1 do
            fsw.pt[i]:=1;
//ifdef[debug_getpsf]
        SImg_WriteImageFile(cc, 'getpsf_fs0.fits',fs);
//endif

        { psf extraction iterations }
        getmem2(dat, n*2*m*4*sizeof(simgfp));
        for i:=1 to itn do
        begin
            r:=0.5+0.5*(i-1)/(itn-1);
            SImg_GetPSF_getdat(cc,  sh.fi2, sh.fw, fs, @stars[0], dat, n, rad,  sh.sat,sh.sig,
                1-r,            { relative deviation parameter }
                sym,            { symmetry parameter }
                r,              { exponent }
                i=1,            { only non-saturated stars in first iteration }
                1{r});          { adjust offset }
            SImg_GetPSF_quant(cc, fs.pt,fsw.pt, dat, n*2,m, 0.35,0.65);
            SImg_GetPSF_normalize(cc, fs,fsw, rad, i=itn);

            if (rrad>1e-5) and (rrad<0.1) then
            begin
                t:=simg_mina; u:=0;
                for y:=-rad to rad do
                begin
                    d:=fs.pt  +(y+rad)*fs.sl   + rad;
                    e:=fpsf.pt+(y+rad)*fpsf.sl + rad;
                    for x:=-rad to rad do
                    begin
                        r:=(sqr(x)+sqr(y))*rrad;
                        if r<1 then
                        begin
                            r:=1-r*(2-r);
                            t+=d[x]*r;
                            u+=e[x]*r;
                        end;
                    end;
                end;
                u/=t;
                t:=0;
                for y:=-rad to rad do
                begin
                    d:=fs.pt  +(y+rad)*fs.sl   + rad;
                    e:=fpsf.pt+(y+rad)*fpsf.sl + rad;
                    for x:=-rad to rad do
                    begin
                        r:=(sqr(x)+sqr(y))*rrad;
                        if r>=1 then r:=e[x]
                        else begin
                            r:=r*(2-r);
                            r:=r*e[x]+(1-r)*d[x]*u;
                        end;
                        d[x]:=r;
                        t+=r;
                    end;
                end;
//              cc.info('t='+real2str(t,10));
            end;

//ifdef[debug_getpsf]
            SImg_WriteImageFile(cc, 'getpsf_fsw'+int2str(i)+'.fits',fsw);
            SImg_WriteImageFile(cc, 'getpsf_fs'+int2str(i)+'.fits',fs);
//endif
        end;
    end
    else begin
        { initial psf }
        t:=-1/sqr(fwhm*FWHM_TO_RAD);
        u:=0;
        d:=fs.pt;  e:=fsw.pt;
        for y:=-rad to rad do
            for x:=-rad to rad do
            begin
                r:=exp((sqr(x)+sqr(y))*t);
                d^:=r; d+=1;
                e^:=1; e+=1;
                u+=r;
            end;
        u:=1/u;
        for i:=0 to fs.size-1 do
            fs.pt[i]*=u;
//ifdef[debug_getpsf]
        SImg_WriteImageFile(cc, 'getpsf_fs0.fits',fs);
//endif

        { psf extraction iterations }
        getmem2(dat, n*2*m*4*sizeof(simgfp));
        for i:=1 to itn do
        begin
            r:=(i-1)/(itn-1);
            SImg_GetPSF_getdat(cc,  sh.fi2, sh.fw, fs, @stars[0], dat, n, rad,  sh.sat,sh.sig,
                1-r,            { relative deviation parameter }
                sym,            { symmetry parameter }
                r,              { exponent }
                i=1,            { only non-saturated stars in first iteration }
                1{r});          { adjust offset }
            SImg_GetPSF_quant(cc, fs.pt,fsw.pt, dat, n*2,m, 0.35,0.65);
            SImg_GetPSF_normalize(cc, fs,fsw, rad, i=itn);
//ifdef[debug_getpsf]
            SImg_WriteImageFile(cc, 'getpsf_fsw'+int2str(i)+'.fits',fsw);
            SImg_WriteImageFile(cc, 'getpsf_fs'+int2str(i)+'.fits',fs);
//endif
        end;
        SImg_New(cc, fpsf, rad*2+1,rad*2+1,1, false);
    end;

    { generate output }
    e:=fpsf.pt + rad*fpsf.sl + rad;
    for y:=-rad to rad do
    begin
        d:=fs.pt  +(y+rad)*fs.sl   + rad;
        e:=fpsf.pt+(y+rad)*fpsf.sl + rad;
        for x:=-rad to rad do
            e[x]:=d[x];
    end;
    SImg_Free(cc,fs);
    SImg_Free(cc,fsw);
{    r:=0;
    for i:=0 to fpsf.size-1 do
        r+=fpsf.pt[i];
    writeln(stderr,'sum=',r:0:3);  }
    freemem2(dat, n*2*m*4*sizeof(simgfp));
    setlength(stars,0);
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_StarSubtract ***********************************************}
{*************************************************************************}
{******* SImg_StarSubtractGetMask **************************************}
{ pattern 2: line }
function SImg_StarSubtract_GetMask_pat2(cc:CSImg_Config; fs:TSImg_FPImage; xm,ym:longint):double;
var x,y : longint;
begin
    result:=0;
    for y:=0 to fs.ny-1 do
        for x:=0 to fs.nx-1 do
            result+=fs.pt[y*fs.sl+x]/sqrt((sqr(x-xm)+sqr(y-ym))*sqr(pi)+1);
end;

{ Find stars using deconvolution
    fi     : input from which maxima are calculated
    rm     : mask input and output
             bit 0: star candidate filtered by getmask
             bit 1: star candidate not filtered by getmask
             bit 2: save star
    add    : find new sources
    fi1    : background subtracted image used for saturation detection
    sat    : saturation threshold
}
threadprocedure SImg_StarSubtract_GetMask_thread1(it:longint; tn:longint; fi:TSImg_FPImage; rm:TSImg_RDImage; add:boolean; fi1:TSImg_FPImage; sat:simgfp);
var r           : simgfp;
    x,y,ya,yz   : longint;
    a           : byte;
    d,e         : ^simgfp;
    b           : ^byte;
begin
    ya:=1 + ((rm.ny-2)*it) div tn;
    yz:=((rm.ny-2)*(it+1)) div tn;
    sat*=0.95;
    for y:=ya to yz do
    begin
        d:=fi.pt + y*fi.sl + 1;
        e:=fi1.pt + y*fi1.sl + 1;
        b:=rm.ptb + y*rm.sl + 1;
        for x:=1 to rm.nx-2 do
        begin
            a:=b^;
            if add and (a and 7=0) then { no input star }
            begin
                r:=d^;      { wiener deconvolution }
                if (r>0) and (r>d[-fi.sl-1]) and (r>d[-fi.sl]) and (r>d[-fi.sl+1]) and (r>d[-1]) and (r>d[1]) and (r>d[fi.sl-1]) and (r>d[fi.sl]) and (r>d[fi.sl+1])                                                    { maximum with 3x3 area }
                    and (b[-rm.sl-1] and 7=0) and (b[-rm.sl] and 7=0) and (b[-rm.sl+1] and 7=0) and (b[-1] and 7=0) and (b[1] and 7=0) and (b[rm.sl-1] and 7=0) and (b[rm.sl] and 7=0) and (b[rm.sl+1] and 7=0)         { no input star within 3x3 area }
                    then a:=a or 1;
            end;
            if (a and 1<>0) and
                ((e^>sat) or (e[-fi1.sl-1]>sat) or (e[-fi1.sl]>sat) or (e[-fi1.sl+1]>sat) or (e[-1]>sat) or (e[1]>sat) or (e[fi1.sl-1]>sat) or (e[fi1.sl]>sat) or (e[fi1.sl+1]>sat))
                then a:=(a and 254) or 2;               { not filtered by getmask }
{           if ((abs(x-7658*2)<3) and (abs(y-5739*2)<3)) or
                ((abs(x-9870)<3) and (abs(y-14977)<3))
                then writeln(stderr,'x=',x,' y=',y,': r='+real2str(r),'  s='+real2str(s),'  b='+int2str(b^),'  a='+int2str(a),'  sat='+real2str(sat)); }
            b^:=a;
            d+=1;
            e+=1;
            b+=1;
        end;
    end;
end;

{ Calculate i / (o*s)
    fi     : i
    ti     : o*s
}
threadprocedure SImg_StarSubtract_GetMask_thread2(it:longint; tn:longint; fi:TSImg_FPImage; ti:TSImg_FFTImage);
var x,y,ya,yz   : longint;
    r,s         : simgfp;
    d           : ^fftfp;
    e           : ^simgfp;
begin
    ya:=(ti.ny*it) div tn;
    yz:=(ti.ny*(it+1)) div tn - 1;
    for y:=ya to yz do
    begin
        d:=ti.pt + y*ti.nxc*2;
        x:=0;
        if y<fi.ny then
        begin
            e:=fi.pt + y*fi.sl;
            for x:=0 to fi.nx-1 do
            begin
                r:=e[x];
                if r<0 then r:=0;
                s:=d[x];
                if s<=1e-2*r then d[x]:=1e2
                else d[x]:=r/s;
            end;
            x:=fi.nx;
        end;
        for x:=x to ti.nxr-1 do
            d[x]:=0;
    end;
end;


{ Find initial source mask and initial solution
    fi1    : Background subtracted input image
    fs     : PSF
    rm     : mask input and output.
             bit 1: star candidate input and output
             bi2 2: save star input and output
    fj     : initial solution for a (undefined if rm and 6=0, otherwise non-negative)
    getmask: bit 0: find new stars
             bit 1: filter input candidates, otherwise they will be set to save stars
    inf    : PSF info
    sat    : saturation threshold
    sig2   : brightness limit
    offset : fi2 offset
    tol    : tolerance
    ff     : filter strength
}
procedure SImg_StarSubtract_GetMask(cc:CSImg_Config; fi1,fs:TSImg_FPImage; rm:TSImg_RDImage; var fj:TSImg_FPImage; getmask:longint; inf:TSImg_PSFinfo; sat,sig2,offs,ff:double);
const brad_thr  = 0.05;  { threshold for determination of background filter size }
      rl_iter   = 3;     { Richardson-Lucy iteration for initial solution }
      min_sfit  = 4;     { minimum shape filter iterations }
      bgfact    = 0.25;  { background factor for fi3 at ff=1 }
      dc_snr    = 20;    { SNR parameter for wiener deconvolution }
      dc_q      = 0.999; { Use (convoluted) input for star center calculation above this quantile }
      r9        = 1/9;

var fi2,fs2             : TSImg_FPImage; { background subtracted input image and PSF, filter radius brad  }
    fi3                 : TSImg_FPImage; { denoised fi2 and PSFs (with and w/o tolerance), filter radius nrad }
    fi4,fs4             : TSImg_FPImage; { smoothed fi3 and PSF for shape filtering }
    ts,ti               : TSImg_FFTImage;
    m2,m4               : simgfp;        { maximum of fs2 and fs4 }
    brad,nrad,srad      : double;
    srad_prev           : double;
    q,r,s,t,u,offs2     : double;
    i,j,x,y,it          : longint;
    b,c                 : byte;
    bo                  : boolean;
    d,e,f               : ^simgfp;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi1) then cc.error('SImg_StarSubtract_GetMask: Invalid image: fi1');
    if fi1.nc<>1 then cc.error('SImg_StarSubtract_GetMask: Invalid dimensions: fi1');
    if not SImg_Recover(cc,fs) then cc.error('SImg_StarSubtract_GetMask: Invalid image: fs');
    if fs.nc<>1 then cc.error('SImg_StarSubtract_GetMask: Invalid dimensions: fs');
    if not SImg_Recover(cc,rm) and (rm.fmt<>SImg_RawByte) then cc.error('SImg_StarSubtract_GetMask: Invalid image: rm');
    if rm.size<>fi1.size then cc.error('SImg_StarSubtract_GetMask: Invalid dimensions: rm');
    if not SImg_OverWriteAlert(fj,fi1) and not SImg_OverWriteAlert(fj,fs) then SImg_Free(cc,fj);

    { denoising filter }
    fs2.pt:=nil;
    nrad:=SImg_GetPSFWidth(cc,fs,0.25)/(4*sqrt(ln(4)));
    SImg_Filter(cc, fs,fs2, nrad);

    { background subtraction radius }
    brad:=SImg_GetPSFWidth(cc,fs2,brad_thr)*0.5;        { = FWHM at Gaussian PSF }
    cc.info(cc.name+': Background subtraction with interpolation radius '+real2str(brad,-1)+' and deconvolution');

    { subtract offset and crop negative values }
    fi3.pt:=nil;
    SImg_New(cc, fi3, fi1.nx,fi1.ny,1);
    for i:=0 to fi1.size-1 do
    begin
        r:=fi1.pt[i]-offs;
        if r<0 then r:=0;
        fi3.pt[i]:=r;
    end;
    SImg_Filter(cc, fi3,fi3, nrad);

    { subtract background }
    fi2.pt:=nil;
    SImg_FBGFilter(cc, fi3,fi2, brad, true);
    SImg_FBGFilter(cc, fs2,fs2, brad, true);
    t:=SImg_GetQuantile(cc, fi2, 0.75);
    u:=ff*bgfact;
    if u<0 then u:=0;
    for i:=0 to fi1.size-1 do
    begin
        s:=fi2.pt[i];
        r:=fi1.pt[i]-offs-s;
        if r<0 then r:=0;
        fi2.pt[i]:=r;
        if s<=t then fi3.pt[i]-=s
        else fi3.pt[i]-=s-(s-t)*u;
        rm.ptb[i]:=rm.ptb[i] and 254;
    end;
    for i:=0 to fs.size-1 do
    begin
        r:=fs.pt[i]-fs2.pt[i];
        if r<0 then r:=0;
        fs2.pt[i]:=r;
    end;
{    SImg_WriteImageFile(cc,'starSubtract-getmask-fi1.fits',fi1);
    SImg_WriteImageFile(cc,'starSubtract-getmask-fi2.fits',fi2);
    SImg_WriteImageFile(cc,'starSubtract-getmask-fs.fits',fs);
    SImg_WriteImageFile(cc,'starSubtract-getmask-fs2.fits',fs2); }

    { wiener deconvolution }
    offs2:=SImg_GetQuantile(cc, fi2, 0.5);
    for i:=0 to fi2.size-1 do
        fi2.pt[i]-=offs2;
    ts.pt:=nil; ti.pt:=nil;
    SImg_New_findbase(cc, ts, fi2.nx+fs2.nx-1, fi2.ny+fs2.ny-1);
    SImg_CopyR(cc, fs2,ts, 0, -inf.xm,-inf.ym);
    SImg_FFT_r2c(cc,ts);
    SImg_New(cc, ti, ts.nxr,ts.ny);
    SImg_Copy(cc, fi2,ti,0);
    SImg_FFT_r2c(cc,ti);
    s:=8*(2-sqrt(fs2.pt[inf.ym*fs2.sl+inf.xm]/ts.pt[0]));  { SNR for wiener regularization: 8..16 }
    SImg_FFTWiener(cc, dc_snr, ti,ti,ts);
    SImg_FFT_c2r(cc,ti);

    { Post-processing deconvolution result used to estimate star centers.
      fi1>sat   : fi1
      <=0       : deconvolution result is used
      >=q(fc_q) : background subtracted input fi2 (=deconvolution input)
    }
    fi4.pt:=nil;
    SImg_Duplicate(cc, fi2,fi4, true);
    q:=SImg_GetQuantile(cc,fi2,dc_q);
    if q<simg_mina then cc.error('SImg_StarSubtract_GetMask: Input image seems to contain no data');
    if q*8>sat then q:=sat/8;
    q:=1/q;
    for y:=1 to fi4.ny-2 do
    begin
        d:=fi2.pt + y*fi2.sl;
        e:=fi4.pt + y*fi4.sl;
        f:=fi1.pt + y*fi1.sl;
        j:=y*2*ti.nxc;
        for x:=1 to fi4.nx-2 do
        begin
            d+=1; e+=1; f+=1;
            s:=f^;
            if s>sat then
            begin
                e^:=s;
                continue;
            end;
            s:=d^;
            t:=s*q;
            if t<=0 then
            begin
                e^:=ti.pt[j+x];
                continue;
            end;
            if t<=1 then
            begin
                e^:=s*t + ti.pt[j+x]*(1-t);
                continue;
            end;
            e^:=s;
        end;
    end;
    SImg_Filter(cc, fi4,fi4, sqrt(0.5+sqr(SImg_GetPSFwidth(cc,fs,0.25)/8)) );   { denoising }
//    SImg_WriteImageFile(cc,'starSubtract-getmask-wiener.fits',fi4);

    { set input star either as save (getmask and 2=0) or as candidates }
    if getmask and 2=0 then c:=6        { input stars are save }
    else c:=1;
    for i:=0 to rm.size-1 do
    begin
        b:=rm.ptb[i] and 254;
        if b and 6=2 then b:=(b and (255-1-2-4)) or c;
        rm.ptb[i]:=b;
    end;

    { find new candidates center/maximum filter }
    cc.info(cc.name+': '
        + select(getmask and 1<>0, 'Searching star candidates, ', '')
        + select(getmask and 2<>0, 'Filtering input stars, ', '')
        + 'Calculating initial solution');
    if getmask<>0 then
    begin
        for i:=0 to cc.threads-1 do
            dothread SImg_StarSubtract_GetMask_thread1(i,cc.threads, fi4,rm, getmask and 1<>0, fi1,sat);
        threadwait;
//      SImg_WriteImageFile(cc,'starSubtract-getmask-mask0.fits',rm);
    end;

    { initial solution using masked RL deconvolution. Must be ran before filter otherwise result may be to large. }
    for y:=0 to ti.ny-1 do
    begin
        j:=y*ti.nxc*2;
        x:=0;
        if y<rm.ny then
        begin
            i:=y*rm.sl;
            for x:=0 to rm.nx-1 do
                if rm.ptb[i+x] and 7=0 then
                begin
                    ti.pt[j+x]:=0;
                    fi4.pt[i+x]:=0;
                end
                else begin
                    r:=fi4.pt[i+x];
                    if r<0 then
                    begin
                        r:=0;
                        fi4.pt[i+x]:=r;
                    end;
                    ti.pt[j+x]:=0;
                end;
            x:=rm.nx;
        end;
        for x:=x to ts.nxr-1 do
            ti.pt[j+x]:=0;
    end;
    for it:=1 to rl_iter do
    begin
        SImg_FFT_r2c(cc, ti);
        SImg_Mul_C(cc, ti,ts, ti);
        SImg_FFT_c2r(cc, ti);
        for i:=0 to cc.threads-1 do
            dothread SImg_StarSubtract_GetMask_thread2(i,cc.threads, fi2,ti);
        threadwait;
        SImg_FFT_r2c(cc, ti);
        SImg_MulK_C(cc, ti,ts, ti);
        SImg_FFT_c2r(cc, ti);
        bo:=(getmask<>0) and (it=rl_iter);
        u:=sig2/inf.max;
        for y:=0 to ti.ny-1 do
        begin
            j:=y*ti.nxc*2;
            x:=0;
            if y<rm.ny then
            begin
                i:=y*rm.sl;
                for x:=0 to rm.nx-1 do
                begin
                    b:=rm.ptb[i+x] and 7;
                    if b=0 then ti.pt[j+x]:=0
                    else begin
                        r:=fi4.pt[i+x]*ti.pt[j+x];
                        if r<0 then r:=0;
                        ti.pt[j+x]:=r;
                        fi4.pt[i+x]:=r;
                        { brightness filter using initial solution from deconvolution }
{                       if ((abs(x-7658*2)<10) and (abs(y-5739*2)<10)) or
                           ((abs(x-4934*2)<10) and (abs(y-7488*2)<10))
                            then writeln('x=',x,' y=',y,': r='+real2str(r)); }
                        if bo and (b=1) and (r<u) then rm.ptb[i+x]:=rm.ptb[i+x] and 254;
                    end;
                end;
                x:=rm.nx;
            end;
            for x:=x to ts.nxr-1 do
                ti.pt[j+x]:=0;
        end;
//        SImg_WriteImageFile(cc,'starSubtract-getmask-fj-'+int2str(it)+'.fits',fi4);
    end;
    SImg_Free(cc, ti);
    SImg_Free(cc, ts);
    SImg_Move(cc, fi4,fj);
//    SImg_WriteImageFile(cc,'starSubtract-getmask-mask1.fits',rm);

    { brightness filter using fi2 }
    if getmask<>0 then
    begin
        SImg_Filter(cc, fi2,fi4, nrad);
        for i:=0 to rm.size-1 do
            if (rm.ptb[i] and 1=1) and (fi2.pt[i]<sig2) and (fi4.pt[i]<sig2) then rm.ptb[i]:=rm.ptb[i] and 254;
    end;
    SImg_Free(cc, fi2);
    SImg_Free(cc, fi4);

    if getmask<>0 then
    begin
        { shape filter }
        fi4.pt:=nil; fs4.pt:=nil;
        srad:=2*nrad; srad_prev:=0;
        u:=SImg_StarSubtract_GetMask_pat2(cc, fs2 , inf.xm,inf.ym);
        m2:=fs2.pt[0];
        for i:=1 to fs2.size-1 do
            if fs2.pt[i]>m2 then m2:=fs2.pt[i];
        it:=0;
        repeat
            r:=sqrt(sqr(srad)-sqr(srad_prev));
            if fs4.pt=nil then SImg_Filter(cc, fs2,fs4, srad )
            else SImg_Filter(cc, fs4,fs4, r );
            if fi4.pt=nil then SImg_Filter(cc, fi3,fi4, srad)
            else SImg_Filter(cc, fi4,fi4, r);
            m4:=fs4.pt[0];
            for i:=1 to fs4.size-1 do
                if fs4.pt[i]>m4 then m4:=fs4.pt[i];
            q:=u/SImg_StarSubtract_GetMask_pat2(cc, fs4 , inf.xm,inf.ym);
            r:=1-m4/m2;
            cc.debug('SImg_StarSubtract_GetMask: srad='+real2str(srad,-2)+' fact='+real2str(q)+' sig2_fact='+real2str(r));
            { filter candidates  }
            r*=sig2;
            for i:=0 to rm.size-1 do
            begin
                b:=rm.ptb[i];
                if b and 3=1 then
                begin
                    s:=fi3.pt[i];
                    t:=fi4.pt[i];
                    if s<t+r then rm.ptb[i]:=b and 254
                    else begin
                        t-=offs2;
                        if t<0 then t:=0;
                        if s>t*q+offs2 then rm.ptb[i]:=(b and 254) or 2;
                    end;
{                    y:=i div rm.sl;
                    x:=i-y*rm.sl;
                    if (abs(x-8924)<10) and (abs(y-14040)<10) then writeln(x,', ',y,': ',rm.ptb[i]); }
                end;
            end;
            it+=1;
            { next step  }
            srad_prev:=srad;
            srad:=sqrt(sqr(nrad) + sqr(srad*1.2));
//          SImg_WriteImageFile(cc,'starSubtract-getmask-mask'+int2str(it+1)+'.fits',rm);
        until (srad*(0.2+ff)>brad*(1+ff)) and (it>min_sfit);
        SImg_Free(cc,fi4);
        SImg_Free(cc,fs4);
        for i:=0 to rm.size-1 do
        begin
            b:=rm.ptb[i];
            if b and 3=1 then rm.ptb[i]:=b and 254;
        end;
    end;
    SImg_Free(cc,fs2);
    SImg_Free(cc,fi3);
{    SImg_Copy(cc, ti,fi3, 0);
    for i:=0 to fi3.size-1 do
        if fi3.pt[i]<0 then fi3.pt[i]:=0;
    SImg_WriteImageFile(cc,'starSubtract-getmask-fj-0.fits',fi3);}
//    if getmask<>0 then SImg_WriteImageFile(cc,'starSubtract-getmask-mask1.pgm',rm);
end;

{******* SImg_StarSubtract_Getmask2 **************************************}
{ Find extended saturated areas
    fi2    : Background subtracted input image
    fs     : PSF
    rm     : Bit 5 (mask 32) is set at extended area and  Bits 1 and 2 are unset
    sat    : saturation level
    sat_thr: saturation threshold in order to make GetMask2 independent from this value
    ff     : filter factor
}
procedure SImg_StarSubtract_GetMask2(cc:CSImg_Config; fi2,fs:TSImg_FPImage; rm:TSImg_RDImage; sat,sat_thr,ff:simgfp);
const oe_ratio = 500;   { overexposure ratio }
var i,j,rad,x,y : longint;
    xa,ya,xz,yz : longint;
    s,t         : double;
begin
    if not SImg_Recover(cc,fi2) then cc.error('SImg_StarSubtract_GetMask2: Invalid image: fi2');
    if fi2.nc<>1 then cc.error('SImg_StarSubtract_GetMask2: Invalid dimensions: fi2');
    if not SImg_Recover(cc,fs) then cc.error('SImg_StarSubtract_GetMask2: Invalid image: fs');
    if fs.nc<>1 then cc.error('SImg_StarSubtract_GetMask: Invalid dimensions2: fs');
    if not SImg_Recover(cc,rm) and (rm.fmt<>SImg_RawByte) then cc.error('SImg_StarSubtract_GetMask2: Invalid image: rm');
    if rm.size<>fi2.size then cc.error('SImg_StarSubtract_GetMask2: Invalid dimensions: rm');
    if ff<0 then ff:=0;
    s:=SImg_GetPSFWidth(cc,fs,sat_thr/oe_ratio)*0.5*0.7/(ff+0.2)+0.7;
    rad:=round(2*s+3);
    s:=sat*(1+0.1*s); { see SImg_SubtractBackground }
    t:=sat*0.5;
    for i:=0 to fi2.size-1 do
        if fi2.pt[i]>s then
        begin
            y:=i div fi2.sl;
            x:=i - y*fi2.sl;
            xa:=x-rad; if xa<0 then xa:=0;
            ya:=y-rad; if ya<0 then ya:=0;
            xz:=x+rad; if xz>=fi2.nx then xz:=fi2.nx-1;
            yz:=y+rad; if yz>=fi2.ny then yz:=fi2.ny-1;
            for y:=ya to yz do
            begin
                j:=y*fi2.sl;
                for x:=xa to xz do
                  if fi2.pt[j+x]>t then rm.ptb[j+x]:=(rm.ptb[j+x] and 249) or 32;
            end;
        end;
end;

{******* SImg_StarSubtract_WidthFilter_approx ****************************}
{ Approximate fi2 at (x0,y0) by constant, normal PSF fpsf and a second PSF fr, a wider reference PSF.
    fi2    : background corrected input (1 color image)
    fpsf   : input PSF s
    fr0    : reference PSF 0
    fr1    : reference PSF 1
    tas    : a \ast s
    xc,yc  : PSF center position
    x0,y0  : position
    a      : PSF factor factor a (including optional regularization)
    sat    : saturation threshold. Return true if one of the pixels is equal or larger.
    rad    : test radius
    rr     : mixing factor for reference PSF: fr = fr0*(1-rr) + fr1*rr.
    ap,ar,b: result: factor for fpsf, factor for fr, constant b
    result : false if saturation or singularity occurred
}
function SImg_StarSubtract_WidthFilter_approx(const fi2,fpsf,fr0,fr1:TSImg_FPImage; const tas:TSImg_FFTImage; const xc,yc,x0,y0:longint; const a,sat,rad,rr:simgfp; out ap,ar,b:simgfp):boolean; inline;
const wreg  = 0.05;     { weight regularization }
      sqrt2 = sqrt(2);
var a11,a21,a31,a22,a32,a33     : double;
    b1,b2,b3,p,q,r,s,t,w,rsr    : double;
    x,y,xa,xz,ya,yz,xa2,xz2,sy  : longint;
    d,ep,e0,e1                  : ^simgfp;
    f                           : ^fftfp;
begin
    result:=false;

    rsr:=1/sqr(rad);
    x:=round(rad*sqrt2-0.501);
    if x<yc then ya:=y0-x else ya:=y0-yc;
    if ya<0 then ya:=0;
    if x<fpsf.ny-yc then yz:=y0+x else yz:=y0+fpsf.ny-1-yc;
    if yz>=fi2.ny then yz:=fi2.ny-1;
    if x<xc then xa:=x0-x else xa:=x0-xc;
    if xa<0 then xa:=0;
    if x<fpsf.nx-xc then xz:=x0+x else xz:=x0+fpsf.nx-1-xc;
    if xz>=fi2.nx then xz:=fi2.nx-1;

    { calculate matrix + vector }
    b1:=0;  b2:=0;  b3:=0;
    a11:=0; a21:=0; a31:=0;
    a22:=0; a32:=0; a33:=0;
    for y:=ya to yz do
    begin
        sy:=sqr(y-y0);
        d:=fi2.pt + y*fi2.sl;
        x:=(yc+y-y0)*fpsf.sl + xc-x0;
        ep:=fpsf.pt + x;
        e0:=fr0.pt + x;
        e1:=fr1.pt + x;
        f:=tas.pt + y*tas.nxc*2;
        x:=round(sqrt(2*sqr(rad)-sqr(y-y0))-0.5);
        xa2:=x0-x; if xa2<xa then xa2:=xa;
        xz2:=x0+x; if xz2>xz then xz2:=xz;
        for x:=xa2 to xz2 do
        begin
            q:=d[x];
            if q>sat then exit;
            w:=q*wreg;  if w<simg_mina then w:=simg_mina;
            p:=ep[x];                   { PSF }
            r:=e0[x]*(1-rr) + e1[x]*rr; { reference PSF }
            t:=a*p;     if t<0 then t:=0;
            s:=f[x];    if s<0 then s:=0;
            w:=(t+w)/(s+w);             { weight }
            q+=t-s;                     { corrected fi2 }
            t:=(sy+sqr(x-x0))*rsr;
            if t>1 then w*=2-t;
            a11+=w;
            a21+=p*w;
            a31+=r*w;
            a22+=sqr(p)*w;
            a32+=p*r*w;
            a33+=sqr(r)*w;
            q*=w;
            b1+=q;
            b2+=q*p;
            b3+=q*r;
        end;
    end;
//    printmatrix( solve( rmatrix(3,3, [a11,a21,a31, a21,a22,a32, a31,a32,a33]), rmatrix(1,3, [b1,b2,b3]) ) );
    if a33<simg_mina then exit;
    a33:=1/a33;
    s:=a32*a33; b2-=s*b3; a21-=s*a31;  a22-=s*a32;
    s:=a31*a33; b1-=s*b3; a11-=s*a31;
    r:=a22*a11 - sqr(a21);
    if r<simg_mina then exit;
    r:=1/r;
    b :=(a22*b1 - a21*b2)*r;   { constant = background }
    ap:=(a11*b2 - a21*b1)*r;   { factor for PSF fpsf }
    ar:=(b3-b*a31-ap*a32)*a33; { factor for reference PSF fr }
    result:=true;
end;


{******* SImg_AspectRatio ************************************************}
{ reg in pixels^2 }
function SImg_GetAspectRatio(const m,n,reg:simgfp):double; inline;
var r,s : double;
begin
    r:=2*pi*m;
    s:=sqr(n);
    if s>=r then
    begin
        result:=1;
        exit;
    end;
    r:=(r+reg)/(s+reg);
    result:=r+sqrt(sqr(r)-1);
end;


{******* SImg_StarSubtract_WidthFilter_width ****************************}
{ Width and shape test at (x0,y0)
    fi2    : background corrected input (1 color image)
    fpsf   : input PSF s
    tas    : a \ast s
    x0,y0  : position minimum distance to border must be 1
    a,b    : PSF factor factor a (including optional regularization), background b
    rad    : test radius
    thr    : brightness threshold
    m,n    : output
    result : false if if result is undefined
}
function SImg_StarSubtract_WidthFilter_width(const fi2,fpsf:TSImg_FPImage; const tas:TSImg_FFTImage; const xc,yc,x0,y0:longint; const a,b,rad,thr:simgfp; out m,n:double):boolean; inline;
const sqrt2 = sqrt(2);
var r,s,t,xr,yr,rsr     : double;
    x,y,xa,xz,ya,yz     : longint;
    xa2,xz2,x1,y1,sr    : longint;
    d,e                 : ^simgfp;
    f                   : ^fftfp;
begin
    result:=false;

    rsr:=1/sqr(rad);
    x:=round(rad*sqrt2-0.501);
    if x<yc then ya:=y0-x else ya:=y0-yc;
    if ya<0 then ya:=0;
    if x<fpsf.ny-yc then yz:=y0+x else yz:=y0+fpsf.ny-1-yc;
    if yz>=fi2.ny then yz:=fi2.ny-1;
    if x<xc then xa:=x0-x else xa:=x0-xc;
    if xa<0 then xa:=0;
    if x<fpsf.nx-xc then xz:=x0+x else xz:=x0+fpsf.nx-1-xc;
    if xz>=fi2.nx then xz:=fi2.nx-1;

    { search maximum within 3x3 region }
    s:=0;
    for y:=-1 to 1 do
    begin
        d:=fi2.pt  + (y+y0)*fi2.sl + x0;
        e:=fpsf.pt + (y+yc)*fpsf.sl + xc;
        f:=tas.pt  + (y+y0)*tas.nxc*2 + x0;
        for x:=-1 to 1 do
        begin
            r:=d[x]+a*e[x]-f[x];
            if r>s then s:=r;
        end;
    end;
{    if ((abs(x0-7658*2)<10) and (abs(y0-5739*2)<10)) or
        ((abs(x0-4934*2)<10) and (abs(y0-7488*2)<10))
        then writeln(stderr,'x=',x0,' y=',y0,' a=',real2str(a),' s=',real2str(s),' b=',real2str(b),' thr=',real2str(thr));}
    s:=(s-b)*thr;
    if s<simg_mina then exit;
    s:=2/s;

    { calculate result }
    n:=0; m:=0;
    xr:=0; yr:=0;
    for y:=ya to yz do
    begin
        y1:=y-y0;
        d:=fi2.pt  + y*fi2.sl;
        e:=fpsf.pt + (yc+y-y0)*fpsf.sl + xc-x0;
        f:=tas.pt  + y*tas.nxc*2;
        x:=round(sqrt(2*sqr(rad)-sqr(y-y0))-0.5);
        xa2:=x0-x; if xa2<xa then xa2:=xa;
        xz2:=x0+x; if xz2>xz then xz2:=xz;
        for x:=xa2 to xz2 do
        begin
            r:=(d[x]+a*e[x]-f[x]-b)*s-1;
            if r>0 then
            begin
                if r>1 then r:=1;
                x1:=x-x0;
                sr:=sqr(x1)+sqr(y1);
                t:=sr*rsr;
                if t>1 then r*=2-t;
                n+=r;
                m+=sr*r;
                xr+=x1*r;
                yr+=y1*r;
            end;
        end;
    end;
    if n>simg_mina then
    begin
        r:=1/n;
        xr*=r; yr*=r;
//      write(stderr,'  xc=',xr:0:1,' yc=',yr:0:1,'  ');
        if xr<-0.5 then xr:=-0.5
        else if xr>0.5 then xr:=0.5;
        if yr<-0.5 then yr:=-0.5
        else if yr>0.5 then yr:=0.5;
        m-=(sqr(xr)+sqr(yr))*n;
    end;
    result:=true;
end;


{******* SImg_StarSubtract_WidthFilter_tests *****************************}
procedure SImg_StarSubtract_WidthFilter_tests(cc:CSImg_Config);
const rad = 20;
      rot = -30;        { in degree }
      w1  = 12;
      w2  = 5;
      wr  = 10;
var fpsf,f1,fr  : TSImg_FPImage;
    tas         : TSImg_FFTImage;
    i,x,y       : longint;
    s,c         : double;
    a,b,r       : simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;

    { f1: constant image }
    f1.pt:=nil;
    SImg_New(cc, f1, 5*rad+1, 7*rad+1, 1);
    for i:=0 to f1.size-1 do
        f1.pt[i]:=1;

    { fpsf: PSF }
    fpsf.pt:=nil;
    SImg_New(cc, fpsf, 6*rad+1, 4*rad+1, 1);
    c:=cos(rot*pi/180);
    s:=sin(rot*pi/180);
    i:=0;
    for y:=-2*rad to 2*rad do
        for x:=-3*rad to 3*rad do
        begin
            fpsf.pt[i]:=exp(-sqr((c*x-s*y)/w1)-sqr((s*x+c*y)/w2));
            i+=1;
        end;
    SImg_WriteImageFile(cc,'corrTest-psf.fits',fpsf);

    { fr: reference PSF }
    fr.pt:=nil;
    SImg_Filter(cc, fpsf,fr, wr);
    SImg_WriteImageFile(cc,'corrTest-r.fits',fr);

    { tas: a*s }
    tas.pt:=nil;
    SImg_New_findbase(cc, tas, fpsf.nx, fpsf.ny);
    SImg_CopyC(cc, fpsf,tas, 0);

    { approx tests  }
    SImg_StarSubtract_WidthFilter_approx(f1  ,fpsf,fr,fr, tas, 3*rad,2*rad, 3*rad,2*rad, 1,10, 2*rad,0.5, a,r,b);
    writeln(stderr,'constant: '+real2str(a)+'  '+real2str(r)+'  '+real2str(b));

    SImg_StarSubtract_WidthFilter_approx(fpsf,fpsf,fr,fr, tas, 3*rad,2*rad, 3*rad,2*rad, 1,10, 2*rad,0.5, a,r,b);
    writeln(stderr,'PSF     : '+real2str(a)+'  '+real2str(r)+'  '+real2str(b));

    SImg_StarSubtract_WidthFilter_approx(fr  ,fpsf,fr,fr, tas, 3*rad,2*rad, 3*rad,2*rad, 1,10, 2*rad,0.5, a,r,b);
    writeln(stderr,'ref. PSF: '+real2str(a)+'  '+real2str(r)+'  '+real2str(b));

    SImg_StarSubtract_WidthFilter_approx(fr  ,fpsf,fr,fr, tas, 3*rad,2*rad, 3*rad,2*rad, 1.5,10, 2*rad,0.5, a,r,b);
    writeln(stderr,'normal. : '+real2str(a)+'  '+real2str(r)+'  '+real2str(b));

    SImg_Duplicate(cc, fpsf,f1, true);
    for i:=0 to f1.size-1 do
        f1.pt[i]+=2+3*fr.pt[i];
    SImg_StarSubtract_WidthFilter_approx(f1  ,fpsf,fr,fr, tas, 3*rad,2*rad, 3*rad,2*rad, 1,10, 2*rad,0.5, a,r,b);
    writeln(stderr,'mix     : '+real2str(a)+'  '+real2str(r)+'  '+real2str(b));

    SImg_Filter(cc, fr,f1, 0.2*wr);
    SImg_StarSubtract_WidthFilter_approx(f1  ,fpsf,fr,fr, tas, 3*rad,2*rad, 3*rad,2*rad, 1,10, 2*rad,0.5, a,r,b);
    writeln(stderr,'wider   : '+real2str(a)+'  '+real2str(r)+'  '+real2str(b));

    SImg_Filter(cc, fpsf,f1, 0.8*wr);
    SImg_StarSubtract_WidthFilter_approx(f1  ,fpsf,fr,fr, tas, 3*rad,2*rad, 3*rad,2*rad, 1,10, 2*rad,0.5, a,r,b);
    writeln(stderr,'narrower: '+real2str(a)+'  '+real2str(r)+'  '+real2str(b));

    { width tests  }
    SImg_StarSubtract_WidthFilter_width(fpsf,fpsf, tas, 3*rad,2*rad, 3*rad,2*rad, 1,0,rad,0.5, s,c);
    writeln(stderr,'width test fpsf: n='+real2str(c,-1)+' m='+real2str(s,-1)+' ar='+real2str(SImg_GetAspectRatio(s,c,1e-3),-4));
    SImg_StarSubtract_WidthFilter_width(fr  ,fpsf, tas, 3*rad,2*rad, 3*rad,2*rad, 1,0,rad,0.5, s,c);
    writeln(stderr,'width test fr  : n='+real2str(c,-1)+' m='+real2str(s,-1)+' ar='+real2str(SImg_GetAspectRatio(s,c,1e-3),-4));

    SImg_Free(cc, fpsf);
    SImg_Free(cc, f1);
    SImg_Free(cc, fr);
    SImg_Free(cc, tas);
end;


{******* SImg_StarSubtract_WidthFilter **********************************}
{   ti,tn  : thread index, thread number
    fi2    : background corrected input (1 color image)
    fpsf   : input PSF s
    r0,r1  : reference PSF in center, at corner
    fa     : approximation factor a
    rm     : mask image
    tas    : a \ast s
    inf    : psf info
    sat    : saturation limit
    nor    : normalization relative to a
    sr0,sr1: search radius in center corners (half the search radius for approximation test
    wthr   : threshold for width/shape test
    rm0,rm1: square root of reference m in center, corner
}
threadprocedure SImg_StarSubtract_WidthFilter_thread(ti:longint; tn:longint; fi2:TSImg_FPImage; fpsf:TSImg_FPImage; fr0:TSImg_FPImage; fr1:TSImg_FPImage; fa:TSImg_FPImage; rm:TSImg_RDImage; tas:TSImg_FFTImage; inf:TSImg_PSFInfo; sat:simgfp; nor:simgfp; sr0:simgfp; sr1:simgfp; wthr:simgfp; rm0:simgfp; rm1:simgfp);
//const breg   = 0.4;             { Brightness normalization at saturation relative to a }
const breg   = 0.25;             { Brightness normalization at saturation relative to a }
var a,b,an,a1,a2        : simgfp;
    rr,sr,rsr,ran       : simgfp;
    m,n                 : double;
    i,ia,iz,x0,y0,x,y   : longint;
    xa,xz,ya,yz         : longint;
    e                   : ^simgfp;
    f                   : ^fftfp;

begin
    ia:=(rm.size*ti) div tn;
    iz:=(rm.size*(ti+1)) div tn - 1;
    rsr:=1/(sqr(fi2.nx div 2)+sqr(fi2.ny div 2));
    ran:=inf.max/sat;
    for i:=ia to iz do
    begin
        if (rm.ptb[i] and 6)=2 then
        begin
            y0:=i div fi2.sl;
            x0:=i-y0*fi2.sl;
            if (x0<=0) or (y0<=0) or (x0>=fi2.nx-1) or (y0>=fi2.ny-1) then continue;
            rr:=(sqr(x0-fi2.nx div 2)+sqr(y0-fi2.nx div 2))*rsr;  { 0 in center, 1 in corner }
            sr:=sr0*(1-rr) + sr1*rr;
            a:=fa.pt[i];
            { normalization for bright pixels }
            if a<=0 then an:=0
            else an:=a*( breg*sqrt(a*ran) + nor);

            { tests }
            if not SImg_StarSubtract_WidthFilter_approx(fi2,fpsf,fr0,fr1, tas, inf.xm,inf.ym, x0,y0, a,sat,sr*2,rr, a1,a2,b) then continue;
{           if ((abs(x0-7658*2)<10) and (abs(y0-5739*2)<10)) or
                ((abs(x0-4934*2)<10) and (abs(y0-7488*2)<10))
                then writeln(stderr,'x=',x0,' y=',y0,' b=',real2str(b),' a1=',real2str(a1),' a2=',real2str(a2),' a=',real2str(a),' an=',real2str(an)); }
            if (a1+an>0) and SImg_StarSubtract_WidthFilter_width(fi2,fpsf, tas, inf.xm,inf.ym, x0,y0, a+an,b, sr,wthr, m,n) then
            begin
{               if ((abs(x0-7658*2)<10) and (abs(y0-5739*2)<10)) or
                    ((abs(x0-4934*2)<10) and (abs(y0-7488*2)<10))
                    then writeln(stderr,'x=',x0,' y=',y0,' b=',real2str(b),' a1=',real2str(a1),' a2=',real2str(a2),' a=',real2str(a),' an=',real2str(an),' m='+real2str(m),' limit='+real2str(sqr(rm0*(1-rr)+rm1*rr)));}
                if m<=sqr(rm0*(1-rr)+rm1*rr) then continue;
            end;

{           if ((abs(x0-7658*2)<10) and (abs(y0-5739*2)<10)) or
               ((abs(x0-4934*2)<10) and (abs(y0-7488*2)<10))
               then writeln('x=',x0,' y=',y0,': rm');}
            { remove source, not totally thread save }
            rm.ptb[i]:=rm.ptb[i] and 249;
            ya:=y0-inf.ym;
            if ya<0 then ya:=0;
            yz:=y0+fpsf.ny-1-inf.ym;
            if yz>=fi2.ny then yz:=fi2.ny-1;
            xa:=x0-inf.xm;
            if xa<0 then xa:=0;
            xz:=x0+fpsf.nx-1-inf.xm;
            if xz>=fi2.nx then xz:=fi2.nx-1;
            for y:=ya to yz do
            begin
                e:=fpsf.pt + (inf.ym+y-y0)*fpsf.sl + inf.xm-x0;
                f:=tas.pt + y*tas.nxc*2;
                for x:=xa to xz do
                    f[x]-=a*e[x];
            end;
            fa.pt[i]:=0;
        end;
    end;
end;

{   fi2    : background corrected input (1 color image)
    fpsf   : input PSF s
    fa     : approximation factor a
    rm     : mask image
    tas    : a \ast s
    inf    : PSF info
    sat    : saturation limit
    tol    : PSF tolerance in center
    tolc   : PSF tolerance in corners
    ff     : filter strength (clipped to valid range of 0..2)
    nor    : normalization relative to a
}
procedure SImg_StarSubtract_WidthFilter(cc:CSImg_Config; fi2,fpsf,fa:TSImg_FPImage; rm:TSImg_RDImage; tas:TSImg_FFTImage; const inf:TSImg_PSFInfo; sat,tol0,tolc,ff,nor:simgfp);
const nmin = 15;               { minimum number of pixels brighter than threshold for width/shape test }
var fr0,fr1     : TSImg_FPImage;
    sr0,sr1     : simgfp;
    i,x,y,j0,j1 : longint;
    r,s,t,xc,yc : double;
    n0,n1,m0,m1 : double;
    wthr        : double;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi2) then cc.error('SImg_StarSubtract_WidthFilter: Invalid image: fi2');
    if fi2.nc<>1 then cc.error('SImg_StarSubtract_WidthFilter: Invalid dimensions: fi2');
    if not SImg_Recover(cc,fpsf) then cc.error('SImg_StarSubtract_WidthFilter: Invalid image: fpsf');
    if (fpsf.nc<>1) or (fpsf.nx<9) or (fpsf.ny<9) then cc.error('SImg_StarSubtract_WidthFilter: Invalid dimensions: fpsf');
    if not SImg_Recover(cc,rm) and (rm.fmt<>SImg_RawByte) then cc.error('SImg_StarSubtract_WidthFilter: Invalid image: rm');
    if rm.size<>fi2.size then cc.error('SImg_StarSubtract_WidthFilter: Invalid dimensions: rm');
    if fa.nc<>1 then cc.error('SImg_StarSubtract_WidthFilter: Invalid dimensions: fa');
    if fa.size<>fi2.size then cc.error('SImg_StarSubtract_WidthFilter: Invalid dimensions: fa');
    if not SImg_Integrity(tas) then cc.error('SImg_StarSubtract_WidthFilter: Invalid image: tas');
    if (tas.nxr<fi2.nx) or (tas.ny<fi2.ny) then cc.error('SImg_StarSubtract_WidthFilter: Invalid dimensions: tas');

    if tol0<1 then tol0:=1;
    if tolc<tol0 then tolc:=tol0;
    if ff<0 then ff:=0;
    if ff>2 then ff:=2;

    { reference PSF's }
    fr0.pt:=nil;  fr1.pt:=nil;
    SImg_Filter(cc, fpsf,fr0, SIG_TO_RAD*sqrt(0.25 + sqr(inf.sig)*(sqr(tol0)-1)) );  { 0.25 = 4/12-1/12 }
    SImg_Filter(cc, fpsf,fr1, SIG_TO_RAD*sqrt(0.25 + sqr(inf.sig)*(sqr(tolc)-1)) );  { 0.25 = 4/12-1/12 }

    { width/shape test: wthr, m0, r0 }
    t:=0;
    for i:=0 to fr0.size-1 do
        if fr0.pt[i]>t then t:=fr0.pt[i];
    wthr:=(0.5-ff*0.2)/0.9; { desired brightness threshold for width/shape test }
    repeat
        wthr*=0.9;
        r:=2/(wthr*t);
        n0:=0; m0:=0;
        xc:=0; yc:=0;
        j0:=0;
        i:=0;
        for y:=0 to fr0.ny-1 do
            for x:=0 to fr0.nx-1 do
            begin
                s:=fr0.pt[i]*r-1;
                if s>0 then
                begin
                    if s>1 then s:=1;
                    n0+=s;
                    xc+=x*s;
                    yc+=y*s;
                    m0+=(sqr(x)+sqr(y))*s;
                    j0+=1;
                end;
                i+=1;
            end;
    until (n0>=nmin) or (wthr<1e-6);
    m0-=(sqr(xc)+sqr(yc))/n0;
    sr0:=sqrt(j0/pi);

    { m1, sr1 }
    t:=0;
    for i:=0 to fr1.size-1 do
        if fr1.pt[i]>t then t:=fr1.pt[i];
    r:=2/(wthr*t);
    n1:=0; m1:=0;
    xc:=0; yc:=0;
    j1:=1;
    i:=0;
    for y:=0 to fr1.ny-1 do
        for x:=0 to fr1.nx-1 do
        begin
            s:=fr1.pt[i]*r-1;
            if s>0 then
            begin
                if s>1 then s:=1;
                n1+=s;
                xc+=x*s;
                yc+=y*s;
                m1+=(sqr(x)+sqr(y))*s;
                j1+=1;
            end;
        i+=1;
        end;
    m1-=(sqr(xc)+sqr(yc))/n1;
    sr1:=sqrt(j1/pi);
    cc.debug('SImg_StarSubtract_WidthFilter: sr0='+real2str(sr0,-1)+' sr1='+real2str(sr1,-2)+' wthr='+real2str(wthr*100,-1)+'%  n0='+real2str(n0,-1)+' m0='+real2str(m0,-1)+'  n1='+real2str(n1,-1)+' m1='+real2str(m1,-1));

    m0:=sqrt(m0);
    m1:=sqrt(m1);
    for i:=0 to cc.threads-1 do
        dothread SImg_StarSubtract_WidthFilter_thread(i,cc.threads, fi2,fpsf,fr0,fr1,fa, rm, tas, inf, sat,nor, sr0,sr1 ,wthr,m0,m1);
    threadwait;

    SImg_Free(cc, fr0);
    SImg_Free(cc, fr1);
end;


{******* SImg_SubtractStars_PSFErr ***************************************}
{ Calculate PSF error which is the weighted error of shifted PSF
    fpsf   : input PSF
    fe     : error function
    sig    : RMS width of PSF
    Gaussian curve used as weight function
}
procedure SImg_StarSubtract_PSFErr(cc:CSImg_Config; fpsf:TSImg_FPImage; var fe:TSImg_FPImage; sig,tol:double);
const fn  = 3; // filter steps
var fj,fk               : TSImg_FPImage;
    r                   : double;
    i,j,x,y,xs,ys       : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fpsf) then cc.error('SImg_StarSubtract_PSFErr: Invalid image: fpsf');
    if fpsf.nc<>1 then cc.error('SImg_StarSubtract_PSFErr: Invalid dimensions: fpsf');
    if tol<1.1 then tol:=1.1;
    fj.pt:=nil;  fk.pt:=nil;
    SImg_Duplicate(cc, fpsf,fj, true);
    { filter in fn steps and use maximum }
    for j:=1 to fn do
    begin
        SImg_Filter(cc, fpsf,fk, SIG_TO_RAD*sig*sqrt(sqr(tol)-1)*j/fn);
        for i:=0 to fj.size-1 do
            if fk.pt[i]>fj.pt[i] then fj.pt[i]:=fk.pt[i];
    end;
    SImg_Free(cc,fk);
    { shift 1 pixel in each direction  }
    for ys:=-1 to 1 do
        for xs:=-1 to 1 do
        begin
            if (ys=0) and (xs=0) then continue;
            j:=fpsf.sl*ys+xs;
            for y:=1 to fpsf.ny-2 do
            begin
                i:=y*fpsf.sl+1;
                for x:=1 to fpsf.nx-2 do
                begin
                    r:=abs(fpsf.pt[i]-fpsf.pt[i+j]);
                    if r>fj.pt[i] then fj.pt[i]:=r;
                    i+=1;
                end;
            end;
        end;
    SImg_Move(cc, fj,fe);
end;



{******* SImg_StarSubtract_PSFCorrectionWeight **************************}
{ Underweight pixels with large positve residual in order to reduce artifacts caused by bright objects near stars
    fr  : (filtered) residual (input)
    fwi : weight (input)
    fwo : weight (output, must be allocated)
}
procedure SImg_StarSubtract_PSFCorrectionWeight(cc:CSImg_Config; fr,fwi,fwo:TSImg_FPImage);
const l    = 10;        { limit relative to average residual, >= 2 }
      itn  = 4;         { maximum number of iterations }
var fw          : TSImg_FPImage;
    i,it        : longint;
    r,s,t,u     : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fr) then cc.error('SImg_StarSubtract_PSFCorrection: Invalid image: fr');
    if not SImg_Recover(cc,fwi) then cc.error('SImg_StarSubtract_PSFCorrection: Invalid image: fwi');
    if (fwi.nc<>1) or (fwi.ny<>fr.ny) or (fwi.nx<>fr.nx) then cc.error('SImg_StarSubtract_PSFCorrectionWeight: Invalid dimensions: fwi');
    if not SImg_Recover(cc,fwo) then cc.error('SImg_StarSubtract_PSFCorrection: Invalid image: fwo');
    if (fwo.nc<>1) or (fwo.ny<>fr.ny) or (fwo.nx<>fr.nx) then cc.error('SImg_StarSubtract_PSFCorrectionWeight: Invalid dimensions: fwo');

//    SImg_WriteImageFile(cc,'starSubtract-corr_w-0.fits',fw);
    fw:=fwi;
    for it:=1 to itn do
    begin
        { avg residual }
        s:=simg_mina; t:=0;
        for i:=0 to fr.size-1 do
        begin
            r:=fw.pt[i];
            if r<=0 then continue;
            s+=abs(fr.pt[i])*r;
            t+=r;
        end;
        s:=t/(s*l);

        { new weight }
        u:=0;
        for i:=0 to fr.size-1 do
        begin
            r:=abs(fr.pt[i])*s;
            if r<=1 then r:=fwi.pt[i]
            else if r>2 then r:=0
            else r:=(2-r)*fwi.pt[i];
            fwo.pt[i]:=r;
            u+=r;
        end;
//      writeln(stderr,real2str(u/t));
        if u>0.98*t then break;
        fw:=fwo;
    end;
//    SImg_WriteImageFile(cc,'starSubtract-corr_w-'+int2str(itn)+'.fits',fw);
end;


{******* SImg_StarSubtract_PSFCorrection ********************************}
{ Calculate PSF correction
    fr  : residual (input and output)
    fw  : weight, must be between 0 and 1
    fs  : deconvoluted stars
    rad : correction radius
    rfr : residual filtering radius for weight calculation
    Residual is approximated by:
        1. W [ (r*w \ast s^0.5^T) / ( ( w \ast s^0.5^T ) * ( \sum s^0.5 ) ) ] \ast s^0.5
        2. W [ (r*w \ast s^T    ) / ( ( w \ast s^T     ) * ( \sum s     ) ) ] \ast s
        3. W [ (r*w \ast s^2^T  ) / ( ( w \ast s^2^T   ) * ( \sum s^2   ) ) ] \ast s^2
        4. W [ (r*w \ast s^0.5^T) / ( ( w \ast s^0.5^T ) * ( \sum s^0.5 ) ) ] \ast s
    where W is a window function.
}
procedure SImg_StarSubtract_PSFCorrection(cc:CSImg_Config; fr,fw,fs:TSImg_FPImage; rad:longint; rfr:double);
//define[starSubtract_correct_wiener]
//ifdef[starSubtract_correct_wiener]
const itn1 = 2;         { approximation iterations per weight iteration }
//else
const itn1 = 3;         { approximation iterations per weight iteration }
//endif
      itn0 = 2;         { weight iterations }
      swn0 = 1;         { underweight and only partially correct brightest swn stars }
      frad = 0.15;      { filter radius factor for smoothing of the correction function }
type star = record
        b   : simgfp;
        i   : longint;
     end;
var ts                  : array[0..2] of TSImg_FFTImage;
    ti                  : array[0..3] of TSImg_FFTImage;
    fw1,fw2,fi,fj,fk    : TSImg_FPImage;
    fr2                 : TSImg_FPImage;
    rr                  : TSImg_RDImage;
//ifdef[debug_starSubtract]
    fc                  : TSImg_FPImage;
//endif
    smax                : array[0..swn0] of star;
    x,y,i,j,k,rad2      : longint;
    l,m                 : longint;
    it0,it1,swn         : longint;
    x0,y0,xa,ya,xz,yz   : longint;
    r,s,t,u,v,w         : double;
    a00,a01,a02,a03,a33 : double;
    a11,a12,a13,a22,a23 : double;
    b0,b1,b2,b3         : double;
    mm,mv               : matrix;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if rad<5 then rad:=5;
    if not SImg_Recover(cc,fr) then cc.error('SImg_StarSubtract_PSFCorrection: Invalid image: fr');
    if (fr.nc<>1) or (fr.ny<3*rad) or (fr.nx<3*rad) then cc.error(cc.name+': Size of PSF correction tiles to small');
    if not SImg_Recover(cc,fw) then cc.error('SImg_StarSubtract_PSFCorrection: Invalid image: fw');
    if (fw.nc<>1) or (fw.ny<>fr.ny) or (fw.nx<>fr.nx) then cc.error('SImg_StarSubtract_PSFCorrection: Invalid dimensions: fw');
    if not SImg_Recover(cc,fs) then cc.error('SImg_StarSubtract_PSFCorrection: Invalid image: fs');
    if (fs.nc<>1) or (fs.ny<>fr.ny) or (fs.nx<>fr.nx) then cc.error('SImg_StarSubtract_PSFCorrection: Invalid dimensions: fs');

    { filter radius and fk=W[1] }
    rad2:=round(rad*(1+2.5*frad)+0.51);
    rr.pt:=nil;
    SImg_New(cc, rr, SImg_RawByte, 2*rad2+1,2*rad2+1,1,false);
    fk.pt:=nil;
    SImg_New(cc, fk, rr.nx,rr.ny, 1);
    u:=1/sqr(rad);
    i:=0;
    for y:=-rad2 to rad2 do
        for x:=-rad2 to rad2 do
        begin
            r:=(sqr(x)+sqr(y))*u;
            if r<1 then
            begin
                r:=1-sqrt(r);
                fk.pt[i]:=1;
            end
            else r:=0;
            rr.ptb[i]:=round(255*r);
            i+=1;
        end;
    SImg_Filter(cc, fk,fk, rad*frad*0.5, rr);
    SImg_Filter(cc, fk,fk, rad*frad    , rr);
//    SImg_WriteImageFile(cc,'starSubtract-rr.fits',rr);
//    SImg_WriteImageFile(cc,'starSubtract-fk.fits',fk);

    { underweight boundaries }
    fw1.pt:=nil;
    SImg_Duplicate(cc, fw,fw1, true);
    k:=rad;
    if k>fr.nx div 4 then k:=fr.nx div 4;
    if k>fr.ny div 4 then k:=fr.ny div 4;
    v:=1/sqr(k);
    for y:=0 to fw1.ny-1 do
    begin
        i:=y*fw1.sl;
        if y+1<k then y0:=y+1
        else if fw1.ny-y<k then y0:=fw1.ny-y
        else y0:=k;
        r:=y0*v;
        for x:=0 to k-2 do
            fw1.pt[i+x]*=(x+1)*r;
        if y0<>k then
        begin
            s:=r*k;
            for x:=k-1 to fw1.nx-k do
                fw1.pt[i+x]*=s;
        end;
        for x:=fw1.nx-k+1 to fw1.nx-1 do
            fw1.pt[i+x]*=(fw1.nx-x)*r;
    end;

    { find the swn0+1 brightest stars which will be not fully corrected }
    for i:=0 to swn0 do
        smax[i].b:=0;
    swn:=0; k:=0;
    for i:=0 to fs.size-1 do
    begin
        r:=fs.pt[i];
        if r>=simg_mina then k+=1;
        if r>smax[swn].b then
        begin
            j:=swn;
            while (j>0) and (r>smax[j-1].b) do
                j-=1;
            for x:=swn downto j+1 do
                smax[x]:=smax[x-1];
            if swn<swn0 then swn+=1;
            smax[j].b:=r;
            smax[j].i:=i;
        end;
    end;
    if 20*swn>k then swn:=round(k*0.05);

    { underweight bright stars }
//    SImg_WriteImageFile(cc,'starSubtract-fw0.fits',fw1);
    v:=smax[swn].b;
    u:=1/sqr(rad);
    for i:=0 to swn-1 do
    begin
        s:=1-v/smax[i].b;
        x0:=smax[i].i mod fw1.nx;
        y0:=smax[i].i div fw1.nx;
        xa:=x0-rad; if xa<0 then xa:=0;
        xz:=x0+rad; if xz>=fw1.nx then xz:=fw1.nx-1;
        ya:=y0-rad; if ya<0 then ya:=0;
        yz:=y0+rad; if yz>=fw1.ny then yz:=fw1.ny-1;
        for y:=ya to yz do
        begin
            j:=y*fw1.sl;
            for x:=xa to xz do
            begin
                r:=1-(sqr(x-x0)+sqr(y-y0))*u;
                if r>0 then fw1.pt[x+j]*=1-r*s;
            end;
        end;
    end;
//    SImg_WriteImageFile(cc,'starSubtract-fw1.fits',fw1);

    { FFT of:
      s^0.5 = s_0  -> ts[0]
      s     = s_1  -> ts[1]
      s^2   = s_2  -> ts[2] }
    ts[0].pt:=nil; ts[1].pt:=nil; ts[2].pt:=nil;
    SImg_New_findbase(cc, ts[0], fr.nx+2*rad2, fr.ny+2*rad2);
    SImg_New(cc, ts[1], ts[0].nxr,ts[0].ny);
    SImg_New(cc, ts[2], ts[0].nxr,ts[0].ny);
    SImg_CopyR(cc, fs,ts[1], 0, -rad2,-rad2);
    t:=smax[swn].b;
    for i:=0 to ts[0].ny*ts[0].nxc*2-1 do
    begin
        r:=ts[1].pt[i];
        if r>t then
        begin
            r:=t;
            ts[1].pt[i]:=t;
        end;
        if r>0 then
        begin
            s:=sqrt(r); ts[0].pt[i]:=s;
            r:=sqr(r);  ts[2].pt[i]:=r;
        end;
    end;
    SImg_FFT_r2c(cc,ts[0]);
    SImg_FFT_r2c(cc,ts[1]);
    SImg_FFT_r2c(cc,ts[2]);
    for i:=0 to 2 do
    begin
        ti[i].pt:=nil;
        SImg_New(cc, ti[i], ts[0].nxr,ts[0].ny);
    end;

    { weight iterations }
    fw2.pt:=nil;
    SImg_New(cc,fw2, fw.nx,fw.ny,1);
    fj.pt:=nil; fi.pt:=nil;
    SImg_New(cc, fj, rr.nx,rr.ny, 3);
    SImg_New(cc, fi, rr.nx,rr.ny, 3);
//ifdef[debug_starSubtract]
    fc.pt:=nil;
    SImg_New(cc, fc, rr.nx,rr.ny, 3);
//endif
    am2(mm, zmatrix(4,4));
    am2(mv, zmatrix(1,4));
    for it0:=1 to itn0 do
    begin
        { underweight unlikely residuals }
        fr2.pt:=nil;
        SImg_Filter(cc, fr,fr2, rfr );
        SImg_StarSubtract_PSFCorrectionWeight(cc, fr2,fw1,fw2);
        SImg_Free(cc,fr2);
//      SImg_WriteImageFile(cc,'starSubtract-fw'+int2str(it0+1)+'.fits',fw2);

        { fj[c=0] = w \ast s^0.5^T
          fj[c=1] = w \ast s^T
          fj[c=2] = w \ast s^2^T   }
        SImg_CopyC(cc, fw2,ti[0], 0);
        SImg_FFT_r2c(cc, ti[0]);
        for j:=0 to 2 do
        begin
            SImg_MulK_C(cc, ti[0],ts[j], ti[1]);
            SImg_FFT_c2r(cc, ti[1]);
            SImg_Copy(cc, ti[1],fj, j);
        end;
//    SImg_WriteImageFile(cc,'starSubtract-fj.fits',fj);

        { reciprocal of fj and regularization }
        v:=sqrt(smax[swn].b);
        for j:=0 to 2 do { regularization }
        begin
            t:=v*(swn+1);
//          cc.info('q25['+int2str(j)+']='+real2str(SImg_GetQuantile(cc,fj,j,0.25)/t));
            s:=ts[j].pt[0];
            t:=sqr(t);
            i:=j;
            for y:=-rad2 to rad2 do
                for x:=-rad2 to rad2 do
                begin
                    r:=fj.pt[i];
                    if r<0 then r:=0;
//ifdef[starSubtract_correct_wiener]
                    fj.pt[i]:=r*s/(sqr(r)+t);
//else
                    fj.pt[i]:=r/((sqr(r)+t)*s);
//endif
                    i+=3;
                end;
            v:=sqr(v);
        end;

        { approximation iterations }
        ti[3].pt:=nil;
        SImg_New(cc, ti[3], ts[0].nxr,ts[0].ny);
        for it1:=1 to itn1 do
        begin
            { ti[0] = r*w }
            t:=0; v:=0;
            for y:=0 to ts[0].ny-1 do
            begin
                j:=y*ts[0].nxc*2;
                x:=0;
                if y<fr.ny then
                begin
                    i:=y*fr.sl;
                    for x:=0 to fr.nx-1 do
                    begin
                        r:=fr.pt[i];
                        s:=fw2.pt[i];
                        ti[0].pt[x+j]:=r*s;
//ifdef[debug_starSubtract]
                        t+=sqr(r)*s;
                        v+=s;
//endif
                        i+=1;
                    end;
                    x:=fr.nx;
                end;
                for x:=x to ts[0].nxr-1 do
                    ti[0].pt[x+j]:=0;
            end;
//ifdef[debug_starSubtract]
//          cc.info(cc.name+': RMS residual after iteration '+int2str(it1-1)+': '+real2str(sqrt(t/v),6)+'   approximation factors: '+real2str(mv.pt[0])+', '+real2str(mv.pt[1])+', '+real2str(mv.pt[2])+', '+real2str(mv.pt[3]));
//endif

            { fi[c=0] = r*w \ast s^0.5^T
              fi[c=1] = r*w \ast s^T
              fi[c=2] = r*w \ast s^2 }
            SImg_FFT_r2c(cc, ti[0]);
            for j:=0 to 2 do
            begin
//ifdef[starSubtract_correct_wiener]
                SImg_FFTWiener(cc, 5, ti[0],ti[1], ts[j]);
//else
                SImg_MulK_C(cc, ti[0],ts[j], ti[1]);
//endif
                SImg_FFT_c2r(cc, ti[1]);
                SImg_Copy(cc, ti[1],fi, j);
            end;
//            SImg_WriteImageFile(cc,'starSubtract-fi-'+int2str(it)+'-0.fits',fi);

            {  fi[c=0] = (r*w \ast s^0.5^T) / ( ( w \ast s^0.5^T ) * ( \sum s^0.5 ) )
               fi[c=1] = (r*w \ast s^T    ) / ( ( w \ast s^T     ) * ( \sum s     ) )
               fi[c=2] = (r*w \ast s^2^T  ) / ( ( w \ast s^2^T   ) * ( \sum s^2   ) ) }
             for i:=0 to fi.size-1 do
                fi.pt[i]*=fj.pt[i];
//          SImg_WriteImageFile(cc,'starSubtract-fi-'+int2str(it)+'-1.fits',fi);

            { offset and window function }
            k:=sqr(rad); l:=sqr(rad2);
            i:=0; j:=0;
            a00:=0; a01:=0; a02:=0;
            for y:=-rad2 to rad2 do
                for x:=-rad2 to rad2 do
                begin
                    m:=sqr(x)+sqr(y);
                    if (m>k) and (m<=l) then
                    begin
                        a00+=fi.pt[i  ];
                        a01+=fi.pt[i+1];
                        a02+=fi.pt[i+2];
                        j+=1;
                    end;
                    i+=3;
                end;
            r:=1/j;
            a00*=r; a01*=r; a02*=r;
            i:=0;
            for y:=-rad2 to rad2 do
                for x:=-rad2 to rad2 do
                begin
                    if sqr(x)+sqr(y)<=k then
                    begin
                        fi.pt[i  ]-=a00;
                        fi.pt[i+1]-=a01;
                        fi.pt[i+2]-=a02;
                    end
                    else begin
                        fi.pt[i  ]:=0;
                        fi.pt[i+1]:=0;
                        fi.pt[i+2]:=0;
                    end;
                    i+=3;
                end;
            SImg_Filter(cc, fi,fi, rad*frad*0.5,rr);
            SImg_Filter(cc, fi,fi, rad*frad    ,rr);

            { ti[j] = W(fi[c=j]) \ast ts[j] }
            for j:=0 to 2 do
            begin
                { ti[j] = fi[c=j] \ast s_j }
                SImg_CopyC(cc, fi,ti[j], j);
                SImg_FFT_r2c(cc, ti[j]);
                if j=0 then
                begin
                    SImg_Mul_C(cc, ti[0],ts[1], ti[3]);
                    SImg_FFT_c2r(cc,ti[3]);
                end;
                SImg_Mul_C(cc, ti[j],ts[j], ti[j]);
                SImg_FFT_c2r(cc,ti[j]);
            end;
//            SImg_WriteImageFile(cc,'starSubtract-fi-'+int2str(it)+'-2.fits',fi);

            { approximation }
            a00:=0; a01:=0; a02:=0; a03:=0;
            a11:=0; a12:=0; a13:=0;
            a22:=0; a23:=0; a33:=0;
            b0:=0; b1:=0; b2:=0; b3:=0;
            for y:=0 to fr.ny-1 do { calculate vector and matrix on x<=y}
            begin
                j:=y*ts[0].nxc*2;
                i:=y*fr.sl;
                for x:=0 to fr.nx-1 do
                begin
                    r:=fr.pt[i];
                    s:=fw2.pt[i];
                    t:=ti[0].pt[j];
                    u:=ti[1].pt[j];
                    v:=ti[2].pt[j];
                    w:=ti[3].pt[j];
                    a00+=sqr(t)*s;
                    a01+=t*u*s;
                    a02+=t*v*s;
                    a03+=t*w*s;
                    a11+=sqr(u)*s;
                    a12+=u*v*s;
                    a13+=u*w*s;
                    a22+=sqr(v)*s;
                    a23+=v*w*s;
                    a33+=sqr(w)*s;
                    b0+=r*t*s;
                    b1+=r*u*s;
                    b2+=r*v*s;
                    b3+=r*w*s;
                    i+=1; j+=1;
                end;
            end;
            mm.pt[ 0]:=a00;  mm.pt[ 1]:=a01;  mm.pt[ 2]:=a02;  mm.pt[ 3]:=a03;
            mm.pt[ 4]:=a01;  mm.pt[ 5]:=a11;  mm.pt[ 6]:=a12;  mm.pt[ 7]:=a13;
            mm.pt[ 8]:=a02;  mm.pt[ 9]:=a12;  mm.pt[10]:=a22;  mm.pt[11]:=a23;
            mm.pt[12]:=a03;  mm.pt[13]:=a13;  mm.pt[14]:=a23;  mm.pt[15]:=a33;
            mv.pt[0]:=b0;   mv.pt[1]:=b1;   mv.pt[2]:=b2;    mv.pt[3]:=b3;

//          printmatrix(mv);
//          printmatrix(mm);
            am(mv, solve(mm,mv));
//          printmatrix(mv);
            r:=mv.pt[0];
            s:=mv.pt[1];
            t:=mv.pt[2];
            u:=mv.pt[3];
            for y:=0 to fr.ny-1 do
            begin
                j:=y*ts[0].nxc*2;
                i:=y*fr.sl;
                for x:=0 to fr.nx-1 do
                begin
                    fr.pt[i]-=r*ti[0].pt[j] + s*ti[1].pt[j] + t*ti[2].pt[j] + u*ti[3].pt[j];
                    j+=1; i+=1;
                end;
            end;
//ifdef[debug_starSubtract]
            for i:=0 to fc.nx*fc.ny-1 do
            begin
                fc.pt[i*3  ]+=r*fi.pt[i*3  ];
                fc.pt[i*3+1]+=s*fi.pt[i*3+1] + u*fi.pt[i*3];
                fc.pt[i*3+2]+=t*fi.pt[i*3+2];
            end;
//endif
        end;
        SImg_Free(cc,ti[3]);
    end;
    SImg_Free(cc,fi);
    SImg_Free(cc,fj);
    SImg_Free(cc,fk);
    SImg_Free(cc,fw1);
    SImg_Free(cc,fw2);
    SImg_Free(cc,rr);
    for i:=0 to 2 do
    begin
        SImg_Free(cc,ts[i]);
        SImg_Free(cc,ti[i]);
    end;
    free(mm);
    free(mv);

//ifdef[debug_starSubtract]
    SImg_New(cc, fi, fc.nx,fc.ny,1, false);
    for j:=0 to 2 do
    begin
        for i:=0 to fi.size-1 do
            fi.pt[i]:=fc.pt[i*3+j];
        SImg_WriteImageFile(cc,'starSubtract-psfcorr-'+int2str(j)+'.fits',fi);
    end;
    SImg_Free(cc,fc);
    SImg_Free(cc,fi);
//endif
end;


{******* SImg_StarSubtract **********************************************}
{ Subtract stars
    fi     : input (1 color image)
    fpsf   : input PSF
    sh     : shared data. sh.fi2 and sh.fw will be freed if not used anymore
    rms    : mask input and output. Possible inputs:
        1. undefined
        2. mask
            bit 0: interpolated pixels (output only)
            bit 1: star (input and output)
            bit 2: save star (input and output )
        3. floating point (SImg_RawFP) containing deconvoluted stars that will be brightness filtered
    fob    : output image: background
    fos    : output image: deconvoluted stars
    fod    : output image: difference between input and fob where saturated regions are extrapolated
    ofob   : true: output fob; false: do not touch fob
    ofos   : true: output fos; false: do not touch fos
    ofod   : true: output fod; false: do not touch fod
    getmask: bit 0: find new stars
             bit 1: filter input candidates, otherwise they will be set to save stars.
    tol    : psf width tolerance as factor in center
    tolc   : psf width tolerance as factor in corners
    ff     : Source filter strength >=0, high values ==> less objects are considered as stars
    nf     : Noise correction factor. Usually 1.
    ithrn  : Interpolation threshold relative to noise >=1, recommended: >4
    ithrr  : Interpolation threshold relative to residual
    cwdt   : Maximum PSF correction width relative to PSF. Valid: 0..1. Recommended: 0.25..1
    tn     : PSF correction tiles into direction of larger dimension; <0.5 to disable PSF correction
}
procedure SImg_StarSubtract(cc:CSImg_Config; fi,fpsf:TSImg_FPImage; sh:TSImg_StarSubtract_shared; var rms:TSImg_RDImage; var fob,fos,fod:TSImg_FPImage; ofob,ofos,ofod:boolean; getmask:longint; tol,tolc: double; ff,ithrn,ithrr,cwdt:double; tn:double);
const wreg = 1e-4;      { weight regularization }
var inf,inf1            : TSImg_PSFInfo;        { info of fpsf and fpsf2 }
    rm                  : TSImg_RDImage;
    fj,fk,fl,fm,fn,fp   : TSImg_FPImage;
    fpsf1,fpsfw,fw      : TSImg_FPImage;
    fk2,fj2,fw2,fk3     : TSImg_FPImage;        { tiles (fk2,fj2,fw2) and mosaic output (fk3) }
    fc11,fc12,fc22      : TSImg_FPImage;        { matrix components }
    rm2                 : TSImg_RDImage;
    ts,tss2,ts2         : TSImg_FFTImage;       { Fourier transforms of s, s*s2 and s2 }
    ti,tj               : TSImg_FFTImage;
    i,j,x,y,y2,it       : longint;
    itn,itm,itf,cradc   : longint;
    o,p,q,r,s,a,b,c     : double;
    sig2,fr,cradi       : double;               { star detection threshold, filter factor, PSF correction radius }
    w1,w2,w3,w4,w5      : double;
    w6,w7,w8,w9         : double;
    mb                  : byte;
    tinf                : TSImg_TileInfo;
    psfcorrect          : boolean;

begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_StarSubtract: fi='+SImg_InfoStr2(fi)+' fpsf='+SImg_InfoStr2(fpsf)+' rms='+SImg_InfoStr2(rms)+select(ofob,' fob='+SImg_InfoStr2(fob),'')+select(ofos,' fos='+SImg_InfoStr2(fos),'')+select(ofod,' fod='+SImg_InfoStr2(fod),'')+
        ' getmask='+int2str(getmask)+' tolc='+real2str(tolc)+' tol='+real2str(tol)+' ff='+real2str(ff)+' ithrn='+real2str(ithrn)+' ithrr='+real2str(ithrr)+' cwdt='+real2str(cwdt,-2)+' tn='+realstr(tn,-1)+
        ' sh.fi2='+SImg_InfoStr2(sh.fi2)+' sh.fw='+SImg_InfoStr2(sh.fw)+' sh.rad='+real2str(sh.rad)+' sh.sig='+real2str(sh.sig)+' sh.offs='+real2str(sh.offs)+' sh.sat='+real2str(sh.sat)+' sh.irad='+real2str(sh.irad));
    if not SImg_Recover(cc,fi) then cc.error('SImg_StarSubtract: Invalid image: fi');
    if fi.nc<>1 then cc.error('SImg_StarSubtract: Invalid dimensions: fi');
    if not SImg_Recover(cc,fpsf) then cc.error('SImg_StarSubtract: Invalid image: fpsf');
    if (fpsf.nc<>1) or (fpsf.nx<9) or (fpsf.ny<9) then cc.error('SImg_StarSubtract: Invalid dimensions: fpsf');
    if rms.pt<>nil then
    begin
        if (not SImg_Recover(cc,rms)) or ((rms.fmt<>SImg_RawByte) and (rms.fmt<>SImg_RawFP)) then cc.error('SImg_StarSubtract: Invalid image: rm');
        if rms.size<>fi.size then cc.error('SImg_StarSubtract: Invalid dimensions: rm');
    end;
    if ofob and (not SImg_OverwriteAlert(fpsf,fob) and not SImg_OverwriteAlert(fi,fob)) then SImg_Free(cc,fob);
    if ofos and (not SImg_OverwriteAlert(fpsf,fos) and not SImg_OverwriteAlert(fi,fos)) then SImg_Free(cc,fos);
    if ofod and (not SImg_OverwriteAlert(fpsf,fod) and not SImg_OverwriteAlert(fi,fod)) then SImg_Free(cc,fod);

    tol:=abs(tol);
    if tol<1.1 then tol:=1.1;
    if tolc<tol then tolc:=tol;
    if cwdt>2 then cwdt:=2;
    if ithrn<1.5 then ithrn:=1.5;
    if ff<0 then ff:=0;

    { extend psf space }
    q:=SImg_GetPSFWidth(cc,fpsf,0.01)*0.5+1;
    fj.pt:=nil;
    it:=round(q+0.51);
    SImg_New(cc, fj, fpsf.nx+2*it,fpsf.ny+2*it,1, true);
    for y:=0 to fpsf.ny-1 do
    begin
        j:=it*(fj.nx+1) + y*fj.nx;
        i:=y*fpsf.nx;
        for x:=0 to fpsf.nx-1 do
            fj.pt[j+x]:=fpsf.pt[i+x];
    end;
    fpsf:=fj;
    fj.pt:=nil;

    { normalize PSF + PSF info }
    r:=0;
    for i:=0 to fpsf.size-1 do
        r+=fpsf.pt[i];
    if r<simg_mina then cc.error('SImg_StarSubtract: Invalid PSF, sum not positive');
    r:=1/r;
    for i:=0 to fpsf.size-1 do
        fpsf.pt[i]*=r;
    SImg_GetPSFInfo(cc,fpsf,inf);
    sig2:=sh.sig*sqrt(2)/sqrt(0.5+sqr(inf.sig));
    if sig2>sh.sig then sig2:=sh.sig;
    fr:=0.1+1.8*ff;
//ifdef[debug_starSubtract]
    cc.info
//else
    cc.debug
//endif
        (cc.name+': sig='+real2str(sh.sig,5)+'  sig2='+real2str(sig2,5)+'  offs='+real2str(sh.offs,5)+'  fr='+real2str(fr,-2)+'  sat/sig='+real2str(sh.sat/sh.sig,5));
//ifdef[debug_starSubtract]
    cc.info('PSF : '+SImg_PSFInfoString(inf));
//else
    cc.info(cc.name+': PSF info: '+SImg_PSFInfoString(inf));
//endif

    { noise filtered PSF }
    fpsf1.pt:=nil;
    SImg_Filter(cc, fpsf,fpsf1, inf.sig);       { inf.sig is noise filter radius for SImg_SubtractBackground}
    SImg_GetPSFInfo(cc,fpsf1,inf1);
    inf.xm:=inf1.xm;    { will be used as center }
    inf.ym:=inf1.ym;
//ifdef[debug_starSubtract]
    cc.info
//else
    cc.debug
//endif
        ('psf1: '+SImg_PSFInfoString(inf1));

    { fpsfw: PSF weight / window function }
    fpsfw.pt:=nil;
    SImg_FBGFilter(cc, fpsf1,fpsfw, q, false, false);
    SImg_GetPSFInfo(cc,fpsfw,inf1);
//ifdef[debug_starSubtract]
    SImg_WriteImageFile(cc,'starSubtract-fpsfw.fits',fpsfw);
    cc.info
//else
    cc.debug
//endif
        ('psfw: '+SImg_PSFInfoString(inf1));
    SImg_Free(cc,fpsf1);

    { input and output star mask, get initial mask, initial solution }
    fj.pt:=nil; rm.pt:=nil;
    if (rms.pt<>nil) and (rms.fmt=SImg_RawFP) then
    begin
        SImg_New(cc, rm, SImg_RawByte, fi.nx,fi.ny,1, false);
        SImg_New(cc, fj, fi.nx,fi.ny,1, false); { create new instance in order to avoid trouble in the calling code }
        if getmask and 2=0 then mb:=6
        else mb:=2;
        r:=sig2/inf.max;
        q:=r*100;
        for i:=0 to fj.size-1 do
        begin
            s:=rms.ptp[i];
            if s<r then
            begin
                fj.pt[i]:=0;
                rm.ptb[i]:=0;
            end
            else begin
                fj.pt[i]:=s;
                if (s<q) and (sh.fi2.pt[i]<sh.sat) then rm.ptb[i]:=mb
                else rm.ptb[i]:=6;
            end;
        end;
        SImg_Free(cc,rms);
        rms:=rm;
    end
    else if (rms.pt<>nil) and (getmask and 2<>0) then { unset save stars which are not verry bright }
    begin
        s:=0.1*sh.sat;
        for i:=0 to rms.size-1 do
        begin
            mb:=rms.ptb[i];
            if (mb and 4<>0) and (sh.fi2.pt[i]<s) then rm.ptb[i]:=(mb and (255-4)) or 2;
        end;
    end;

    if rms.pt=nil then getmask:=1;
    if getmask=2 then sig2*=0.5;

    if getmask<>0 then
    begin
        itm:=3;         { 2nd getmask attempt, 0 to disable }
        itf:=itm+3;     { 1st filtering iteration }
    end
    else begin
        itm:=0;
        itf:=0;
    end;
    itn:=itf+5;         { total number of iterations }

    rm2.pt:=nil;
    if rms.pt=nil then SImg_New(cc, rms, SImg_RawByte, fi.nx,fi.ny,1, true)
    else if (itm>0) and (getmask and 2<>0) then         { make a backup for the 2nd getmask filter attempt }
    begin
        SImg_New(cc, rm2, SImg_RawByte, fi.nx,fi.ny,1, false);
        move(rms.pt^, rm2.pt^, rms.size);
    end;
    rm:=rms;
    if (getmask<>0) or (fj.pt=nil) then SImg_StarSubtract_GetMask(cc, sh.fi2, fpsf,  rm,  fj,  getmask,  inf,  sh.sat, sig2, sh.offs, ff);
    SImg_StarSubtract_GetMask2(cc, sh.fi2, fpsf, rm, sh.sat,sh.sat_thr,ff);
//    SImg_WriteImageFile(cc,'starSubtract-mask0.fits',rm);

    { w \ast s^2^T }
    ts.pt:=nil;
    SImg_New_findbase(cc, ts, fi.nx+fpsf.nx-1, fi.ny+fpsf.ny-1);
    i:=0;
    for y:=0 to fpsf.ny-1 do
    begin
        y2:=y-inf.ym;
        if y2<0 then y2+=ts.ny;
        for x:=0 to fpsf.nx-1 do
        begin
            j:=x-inf.xm;
            if j<0 then j+=ts.nxr;
            j+=y2*ts.nxc*2;
            ts.pt[j]:=sqr(fpsf.pt[i]);
            i+=1;
        end;
    end;
    tj.pt:=nil;
    SImg_New(cc, tj, ts.nxr,ts.ny);
    SImg_Copy(cc, sh.fw, tj, 0);
    SImg_FFT_r2c(cc, tj);
    SImg_FFT_r2c(cc, ts);
    SImg_MulK_C(cc, tj,ts, ts);
    SImg_FFT_c2r(cc, ts);
    fk.pt:=nil;
    SImg_New(cc, fk, fi.nx,fi.ny,1);
    SImg_Copy(cc, ts,fk, 0);

    { ts, tss2, ts2 }
    ts2.pt:=nil; tss2.pt:=nil;
    ti.pt:=nil; { used for Fourier transform of sqr(s)*s2 }
    SImg_Clear(cc, ts);
    SImg_New(cc, ti,  ts.nxr,ts.ny);
    SImg_New(cc, ts2, ts.nxr,ts.ny);
    SImg_New(cc, tss2,ts.nxr,ts.ny);
    i:=0;
    for y:=0 to fpsf.ny-1 do
    begin
        y2:=y-inf.ym;
        if y2<0 then y2+=ts.ny;
        for x:=0 to fpsf.nx-1 do
        begin
            j:=x-inf.xm;
            if j<0 then j+=ts.nxr;
            j+=y2*ts.nxc*2;
            r:=fpsf.pt[i];
            s:=fpsfw.pt[i];
            ts.pt[j]:=r;
            ts2.pt[j]:=s;
            tss2.pt[j]:=r*s;
            ti.pt[j]:=sqr(r)*s;
            i+=1;
        end;
    end;
//    SImg_FFTWrite(cc,'starSubtract-ts.fits',ts);
    SImg_FFT_r2c(cc,ts);
    SImg_FFT_r2c(cc,ts2);
    SImg_FFT_r2c(cc,tss2);
    SImg_FFT_r2c(cc,ti);

    { matrix components }
    fc11.pt:=nil; fc12.pt:=nil; fc22.pt:=nil;
    SImg_MulK_C(cc, tj,ti, ti);
    SImg_FFT_c2r(cc,ti);
    SImg_New(cc, fc11, fi.nx,fi.ny,1);
    SImg_Copy(cc, ti,fc11, 0);

    SImg_MulK_C(cc, tj, tss2, ti);
    SImg_FFT_c2r(cc,ti);
    SImg_New(cc, fc12, fi.nx,fi.ny,1);
    SImg_Copy(cc, ti,fc12, 0);
    SImg_Free(cc, ti);

    SImg_MulK_C(cc, tj, ts2, tj);
    SImg_FFT_c2r(cc,tj);
    SImg_New(cc, fc22, fi.nx,fi.ny,1);
    SImg_Copy(cc, tj,fc22, 0);
    ti:=tj; tj.pt:=nil;

    { a_0 \ast s }
    SImg_CopyC(cc, fj,ti, 0);
    SImg_FFT_r2c(cc,ti);
    SImg_Mul_C(cc, ti, ts, ti);
    SImg_FFT_c2r(cc,ti);

    { iterations }
    tj.pt:=nil;
    for it:=1 to itn do
    begin
        { 2nd getmask attempt }
        if (it>1) and (it=itm+1) then
        begin
            { prepare mask }
            if rm2.pt=nil then SImg_New(cc, rm2, SImg_RawByte, fi.nx,fi.ny,1, true);    { no backup of input stars }
            for i:=0 to rm.size-1 do
                if (rm.ptb[i] and 6)<>0 then rm2.ptb[i]:=6;     { mark already known positions as save stars }
            { prepare input residual }
            fl.pt:=nil;
            SImg_New(cc, fl, fi.nx,fi.ny,1, false);
            q:=1/sh.sat;
            for y:=0 to fl.ny-1 do
            begin
                i:=y*fl.sl;
                j:=y*ti.nxc*2;
                for x:=0 to fl.nx-1 do
                begin
                    s:=sh.fi2.pt[i+x];
                    r:=s-ti.pt[j+x];
                    if (r<0) or (s>sh.sat) then
                    begin
                        rm2.ptb[i+x]:=6;        { disable that region }
                        fl.pt[i+x]:=0;
                    end
                    else fl.pt[i+x]:=r*sqr(sqr(1-s*q));
                end;
            end;
//          SImg_WriteImageFile(cc,'starSubtract-gm2-fi.fits',fl);
            SImg_Free(cc, ti);
            { getmask }
            SImg_StarSubtract_GetMask(cc,  fl, fpsf, rm2, fl,  getmask,  inf,  simg_large, sig2, sh.offs, ff);
            { handle new star candidates }
            SImg_New(cc, ti, ts.nxr,ts.ny);
            for y:=0 to fl.ny-1 do
            begin
                i:=y*fl.sl;
                j:=y*ti.nxc*2;
                for x:=0 to fl.nx-1 do
                begin
                    mb:=rm.ptb[i+x];
                    if (rm2.ptb[i+x] and 6=2) and (mb and (32+6)=0) then        { new star candidate }
                    begin
                        mb:=mb or 2;
                        rm.ptb[i+x]:=mb;
                        fj.pt[i+x]:=fl.pt[i+x];
                    end;
                    if mb and 6<>0 then ti.pt[j+x]:=fj.pt[i+x];
                end;
            end;
            SImg_Free(cc,fl);
            SImg_Free(cc,rm2);
            { a_k*s }
            SImg_FFT_r2c(cc,ti);
            SImg_Mul_C(cc, ti, ts, ti);
            SImg_FFT_c2r(cc,ti);
        end;

        cc.info(cc.name+': Iteration '+int2str(it)+'/'+int2str(itn)+select((getmask<>0) and (it>=itf),' (with width filter)',''));
        { calculate a and b at star centers and filter stars }
        for y:=0 to ti.ny-1 do
        begin
            j:=y*ti.nxc*2;
            x:=0;
            if y<fi.ny then
            begin
                i:=y*sh.fi2.sl;
                for x:=0 to fi.nx-1 do
                    ti.pt[j+x]:=(sh.fi2.pt[i+x]-ti.pt[j+x])*sh.fw.pt[i+x];
                x:=sh.fi2.nx;
            end;
            for x:=x to ts.nxr-1 do
                ti.pt[j+x]:=0;  { weight on borders is 0 }
        end;
{        if it>=itn-2 then
        begin
            fl.pt:=nil;
            SImg_New(cc, fl, fi.nx,fi.ny,1, false);
            SImg_Copy(cc, ti,fl, 0);
            SImg_WriteImageFile(cc,'starSubtract-rw-'+int2str(it)+'.fits',fl);
            SImg_Free(cc,fl);
        end; }

        SImg_FFT_r2c(cc,ti);
        SImg_New(cc, tj, ts.nxr,ts.ny);
        SImg_MulK_C(cc, ti, ts2, tj);   { tj=d_2 }
        SImg_FFT_c2r(cc,tj);
        SImg_MulK_C(cc, ti, tss2, ti);  { ti=d_1 }
        SImg_FFT_c2r(cc,ti);
{       for y:=0 to ti.ny-1 do
        begin
            j:=y*ts.nxc*2;
            for x:=0 to ti.nxr-1 do
            begin
                r:=tc11.pt[j];
                s:=tc12.pt[j];
                b:=r*tc22.pt[j]-sqr(s);
                if (abs(b)<1e-100) or (r<1e-100) or (y>=sh.fi2.ny) or (x>=sh.fi2.nx) then
                begin
                    ti.pt[j]:=0;
                    tj.pt[j]:=0;
                end else
                begin
//                  a:=(tc22.pt[j]*ti.pt[j]-tc12.pt[j]*tj.pt[j])/b;
//                  a:=ti.pt[j]/tc11.pt[j];
//                  b:=(tc11.pt[j]*tj.pt[j]-tc12.pt[j]*ti.pt[j])/b;
                    a:=ti.pt[j];
                    b:=(r*tj.pt[j]-s*a)/b;
                    if b<0 then b:=0;
                    a:=(a-b*s)/r;
                    ti.pt[j]:=a;
                    tj.pt[j]:=b;
                end;
                j+=1;
            end;
        end;
        SImg_FFTWrite(cc,'starSubtract-a.fits',ti);
        SImg_FFTWrite(cc,'starSubtract-b.fits',tj);}
        if it>itm then i:=it-itm
        else i:=it;
        q:=1.5*i/(itn-itm);
        if q>1 then q:=1;
        for y:=0 to ts.ny-1 do
        begin
            j:=y*ts.nxc*2;
            x:=0;
            if (y>0) and (y<fi.ny-1) then { ignore borders }
            begin
                ti.pt[j]:=0;
                tj.pt[j]:=0;
                j+=1;
                i:=y*fj.sl+1;
                for x:=1 to fi.nx-2 do
                begin
                    a:=0;
                    mb:=rm.ptb[i];
                    if mb and 6<>0 then
                    begin
                        r:=fc11.pt[i];
                        s:=fc12.pt[i];
                        b:=r*fc22.pt[i]-sqr(s);
                        if (abs(b)<1e-50) or (abs(r)<1e-50) then rm.ptb[i]:=mb and 249  { disable any star if a and b are undefined }
                        else begin
                            { approximate star residual at y by a(y) \ast s(x-y) + b(y) }
                            a:=ti.pt[j];
                            b:=(r*tj.pt[j]-s*a)/b;
                            if b<sh.offs then b:=sh.offs;
                            a:=(a-b*s)/r;
{                           if ((abs(x-7321)<5) and (abs(y-11436)<5))
                                then writeln(stderr,int2str(x)+','+int2str(y)+': a='+real2str(a+fj.pt[i])+' b='+real2str(b));}
                            c:=fj.pt[i];
                            if a<-c then a:=-c; { positive constraint }
                            if (getmask<>0) and (mb and 4=0) then
                            begin
                                { filter stars }
                                if a>0 then r:=(a+c)*inf.max
                                else r:=(c+q*a)*inf.max;
                                w1:=sh.fw.pt[i-sh.fw.sl-1];  w2:=sh.fw.pt[i-sh.fw.sl];  w3:=sh.fw.pt[i-sh.fw.sl+1];
                                w4:=sh.fw.pt[i         -1];  w5:=sh.fw.pt[i         ];  w6:=sh.fw.pt[i         +1];
                                w7:=sh.fw.pt[i+sh.fw.sl-1];  w8:=sh.fw.pt[i+sh.fw.sl];  w9:=sh.fw.pt[i+sh.fw.sl+1];
                                s:=(  sh.fi2.pt[i-sh.fi2.sl-1]*w1 + sh.fi2.pt[i-sh.fi2.sl]*w2 + sh.fi2.pt[i-sh.fi2.sl+1]*w3
                                    + sh.fi2.pt[i          -1]*w4 + sh.fi2.pt[i          ]*w5 + sh.fi2.pt[i          +1]*w6
                                    + sh.fi2.pt[i+sh.fi2.sl-1]*w7 + sh.fi2.pt[i+sh.fi2.sl]*w8 + sh.fi2.pt[i+sh.fi2.sl+1]*w9
                                    + wreg*(b+r) ) / ( w1+w2+w3+w4+w5+w6+w7+w8+w9+wreg );
                                s-=r;
                                if b>s then s:=b;  { b>offs ==> s>offs }
                                s-=sh.offs;
//                              if ((abs(x-7321)<5) and (abs(y-11436)<5))
{                               if ((abs(x-1401)<5) and (abs(y-491)<5))
                                    then writeln(stderr,int2str(x)+','+int2str(y)+': r='+real2str(r)+'   s='+real2str(s)+' q='+real2str(q)+'   fr='+real2str(fr)+' sig2='+real2str(sig2));}
                                s:=q*(sig2+q*fr*s);
                                if r<s then
                                begin
                                    rm.ptb[i]:=mb and 249;
                                    a:=0;
                                end;
                                fj.pt[i+1]:=q*s;        { constraint for mixing, address can always be used because adjacent star centers and stars on border are not allowed }
                            end;
//                          fj.pt[i+1]:=b;  { address can always be used because adjacent star centers and stars on border are not allowed }
                        end;
                    end;
                    ti.pt[j]:=a;
                    tj.pt[j]:={sqrt}(abs(a));
                    i+=1; j+=1;
                end;
                x:=fi.nx-1;
            end;
            for x:=x to ts.nxr-1 do
            begin
                ti.pt[j]:=0;
                tj.pt[j]:=0;
                j+=1;
            end;
        end;

        { a mixing and iteration: a_k = a_(k-1)  +  a * |a|*(w \ast ^2^T)/((|a| \ast s)*w \ast s^T) }
        SImg_FFT_r2c(cc,tj);
        SImg_Mul_C(cc, tj, ts, tj);
        SImg_FFT_c2r(cc,tj);
        for y:=0 to ts.ny-1 do
        begin
            j:=y*ts.nxc*2;
            x:=0;
            if y<fi.ny then
            begin
                i:=y*fi.sl;
                for x:=0 to fi.nx-1 do
                    tj.pt[j+x]*=sh.fw.pt[i+x];
                x:=fi.nx;
            end;
            for x:=x to ts.nxr-1 do
                tj.pt[j+x]:=0;
        end;
        SImg_FFT_r2c(cc,tj);
        SImg_MulK_C(cc, tj, ts, tj);
        SImg_FFT_c2r(cc,tj);
        i:=0;
        for y:=0 to fj.ny-1 do
        begin
            j:=y*ts.nxc*2;
            for x:=0 to fj.nx-1 do
            begin
                mb:=rm.ptb[i];
                if mb and 6<>0 then
                begin
                    r:=ti.pt[j];
                    q:=fk.pt[i];
//                  if (abs( x-9750)<10) and (abs(y-8673)<10) then cc.info('it='+int2str(it)+'  x='+int2str(x)+' y='+int2str(y)+'  rm='+int2str(rm.ptb[i])+'  a_k = '+real2str(fj.pt[i])+'  +  '+real2str(r) + ' * ' +real2str(abs(r)*q/tj.pt[j]));
                    q:={sqrt}(abs(r))*q/tj.pt[j];
//                  if (q>1) or (fj.pt[i]+r*q<0) then writeln('x=',x,' y=',y,' r=',r,' fj=',fj.pt[i],' rm=',rm.ptb[i],' q=',q);
                    if q>1 then q:=1;
                    r:=fj.pt[i] + r*q;
                    if (mb and 6=2) and (r<fj.pt[i+1]) then
                    begin
                        r:=0;
                        rm.ptb[i]:=mb and 249;
                    end;
                    fj.pt[i]:=r;
                    ti.pt[j]:=r;
                end;
                i+=1; j+=1;
            end;
        end;
        SImg_Free(cc,tj);

        { a_k*s }
        SImg_FFT_r2c(cc,ti);
        SImg_Mul_C(cc, ti, ts, ti);
        SImg_FFT_c2r(cc,ti);

        { width filter }
        if (getmask<>0) and (it>=itf) then
        begin
            r:=1-1.5*(it-itf)/(itn-itf);
            if r<0 then r:=0;
            SImg_StarSubtract_WidthFilter(cc, sh.fi2,fpsf,fj, rm, ti, inf, sh.sat, tol,tolc,ff, r);
        end;

//      SImg_WriteImageFile(cc,'starSubtract-mask'+int2str(it)+'.fits',rm);
    end;
    SImg_Free(cc, fk);
    SImg_Free(cc, tss2);
    SImg_Free(cc, fc11);
    SImg_Free(cc, fc12);
    SImg_Free(cc, fc22);
    SImg_Free(cc, fpsfw);

    if (not ofob) and (not ofod) then
    begin
        SImg_Free(cc, ts);
        SImg_Free(cc, ts2);
        SImg_Free(cc, ti);
        SImg_Free(cc, sh.fw);
        SImg_Free(cc, sh.fi2);
        SImg_Free(cc, fpsf);
        { skip interpolation }
        if not ofos then SImg_Free(cc,fj)
        else begin
            for i:=0 to fj.size-1 do
                if rm.ptb[i] and 6=0 then fj.pt[i]:=0; { makes fj valid }
            SImg_Move(cc,fj,fos);
        end;
        SImg_DecDebugLevel;
        exit;
    end;

    { convoluted stars }
    fp.pt:=nil;
    SImg_New(cc, fp, fi.nx,fi.ny,1, false);
    SImg_Copy(cc, ti,fp, 0);
    SImg_Free(cc,ti);

    { PSF error, fos }
    SImg_StarSubtract_PSFErr(cc,fpsf,fpsf1, inf.sig,tol);
    cradi:=sh.irad + (sh.rad-sh.irad)*cwdt;                { correction interpolation radius used for fbgfilter }
//    cradc:=round(2*cradi*sh.rad/(cradi+sh.rad)+0.49);    { correction radius: sh.irad < cradi <= cradc <= sh.rad }
    cradc:=round(cradi*2);
    if cradc>=sh.rad then cradc:=round(sqrt(sh.rad*cradi*2));
//   cradc:=round(sqrt(2*sh.rad*cradi));
    if cradi<2*sh.irad then
    begin
        if tn>0.4999 then cc.warning(cc.name+': PSF correction radius to small, disabling PSF correction');
        tn:=0;
        cradi:=2*sh.irad;
    end;
//    cc.info(cc.name+': Calculating interpolation and PSF correction masks. Maximum normal interpolation radius: '+real2str(sh.irad,-1)+'. Correction radius: '+int2str(cradc)+'.');
    cc.info(cc.name+': Calculating interpolation and PSF correction masks. Maximum normal interpolation radius: '+real2str(sh.irad,-1)+'. Correction radius: '+real2str(cradi,-1)+'.');
//ifdef[debug_starSubtract]
    SImg_WriteImageFile(cc,'starSubtract-psferr.fits',fpsf1);
//endif
    SImg_Clear(cc,ts);
    SImg_CopyR(cc, fpsf1,ts, 0, -inf.xm,-inf.ym);
    i:=0;
    SImg_Clear(cc,ts2);
    for y:=0 to fj.ny-1 do
    begin
        j:=y*ts.nxc*2;
        for x:=0 to fj.nx-1 do
        begin
            if rm.ptb[i] and 6<>0 then ts2.pt[j]:=fj.pt[i]
            else fj.pt[i]:=0; { makes fj valid }
            i+=1; j+=1;
        end;
    end;
    SImg_FFT_r2c(cc,ts);
    SImg_FFT_r2c(cc,ts2);
    SImg_Mul_C(cc, ts, ts2, ts2);
    SImg_Free(cc,ts);
    SImg_FFT_c2r(cc,ts2);
    SImg_Free(cc,fpsf1);

    { limiting interpolation size, fl: <error PSF> \ast <stars=fj> }
    fl.pt:=nil;  fm.pt:=nil;
    SImg_New(cc, fl, fj.nx,fj.ny,1);
    i:=0;
    for y:=0 to fj.ny-1 do
    begin
        j:=y*ts.nxc*2;
        for x:=0 to fj.nx-1 do
        begin
            r:=ts2.pt[j];
            if r<0 then r:=0; { may happen due to FFT rounding errors }
            fl.pt[i]:=r;
            i+=1; j+=1;
        end;
    end;
    SImg_Free(cc,ts2);
    fw.pt:=nil;
    SImg_FBGFilter(cc, fl,fw, sh.irad, true, false);                            { irad <= cradi <= cradc }
    SImg_FBGFilter(cc, fw,fm, sqrt(sqr(cradi)-sqr(sh.irad)), true, true);       { fm<=fw }

    { denoised residual for interpolation thresholds }
    fn.pt:=nil;
    if ithrr<1e-6 then SImg_New(cc, fn, fi.nx,fi.ny,1, true)
    else begin
        SImg_New(cc, fn, fi.nx,fi.ny,1, false);
        a:=ithrr/ithrn;
        b:=(sh.offs+2*sh.sig)*a;
        for i:=0 to fn.size-1 do
            fn.pt[i]:=(sh.fi2.pt[i]-fp.pt[i])*a-b;
        SImg_FilterW(cc, fn,sh.fw, fn, sh.irad*2);
//      SImg_WriteImageFile(cc,'starSubtract-ress1.fits',fn);
    end;

    { interpolation mask and weight for PSF correction }
    a:=sqr(sh.sig);
    c:=ithrn*sh.sig;
    psfcorrect:=tn>0.4999;
    fk.pt:=nil;
    SImg_New(cc, fk, fj.nx,fj.ny,1, false);
    for i:=0 to fi.size-1 do
    begin
        { interpolation masks }
        mb:=rm.ptb[i] and (255-1-8-16);
        q:=fl.pt[i];                                    { stars convoluted with error PSF }
        p:=q-fm.pt[i];
        r:=fn.pt[i];
        if r>0 then r+=sh.sig
        else r:=sh.sig;
        o:=r*ithrn;
//        if p<0.5*r then p:=0
        if p<0.5*sh.sig then p:=0
        else begin
            mb:=mb or 8;                                { PSF correction interpolation }
            s:=q-fw.pt[i];
            if (p>0.5*(o+r)) or (s>r) then mb:=mb or 16;        { possible extended interpolation }
            if s<o then p:=1
            else begin
                mb:=mb or 1;                            { forced interpolation }
                p:=2-s/o;
            end;
        end;

        { residual + constraints }
        r:=sh.fi2.pt[i];
        s:=fp.pt[i];
        if s<0 then                                     { may happen due to rounding errors }
        begin
            s:=0;
            fp.pt[i]:=0;
        end;
        if r>sh.sat then
        begin
            if mb and 32<>0 then
            begin                                       { no interpolation in extended saturated areas }
                mb:=mb and (255-1-8-16);
                p:=s+sh.sat;
            end
            else begin
                mb:=mb or 1;
                p:=s+sh.offs;
            end;
            if r<p then
            begin
                r:=p;
                sh.fi2.pt[i]:=p;
            end;
            p:=0;
        end;
        rm.ptb[i]:=mb;
        s:=r-s;
        fk.pt[i]:=s;                    { residual }
        if r<sh.offs then fm.pt[i]:=r   { constrainted residual }
        else if s<sh.offs then fm.pt[i]:=sh.offs
        else fm.pt[i]:=s;

        { PSF correction weight }
        if psfcorrect then
        begin
            if p<simg_mina then fw.pt[i]:=0             { correction mask: with smooth boundaries }
            else fw.pt[i]:=p * a/(sqr(o-c)+a);
        end;
    end;
    if not ofod then SImg_Free(cc,fp);
//    cc.info(SImg_InfoStr2(fi)+'   '+SImg_InfoStr2(sh.fi2)+'   '+SImg_InfoStr2(fm));

    { PSF correction }
    SImg_InterpolateMasked(cc, fm, rm , 1+8, SImg_StarSubtract_ipol_fd);        { also lower constraint at mask or 8 or 16}
//    cc.info(SImg_InfoStr2(fi)+'   '+SImg_InfoStr2(sh.fi2)+'   '+SImg_InfoStr2(fm));
    if psfcorrect then
    begin
        SImg_Free(cc,fn);
//        SImg_WriteImageFile(cc,'starSubtract-corr.fits',fk);
//        SImg_WriteImageFile(cc,'starSubtract-corr_w.fits',fw);

        { correction residual }
        for i:=0 to fk.size-1 do
            fk.pt[i]-=fm.pt[i];

        { tiles }
        p:=sqrt(1+sqr(SImg_GetPSFWidth(cc, fpsf, 0.2)*0.5));
        if tn<1 then tn:=1;
        if fi.nx>fi.ny then
        begin
            tinf.nx:=round(tn);
            tinf.ny:=round(fi.ny/fi.nx*tn);
        end
        else begin
            tinf.ny:=round(tn);
            tinf.nx:=round(fi.nx/fi.ny*tn);
        end;
        tinf.bb:=round(sh.rad);
        tinf.bu:=cradc;
        i:=fk.nx div (tinf.nx*2)-1;
        if i<tinf.bb then tinf.bb:=i;
        i:=fk.ny div (tinf.ny*2)-1;
        if i<tinf.bb then tinf.bb:=i;
        fk3.pt:=nil;
        if (tinf.nx>1) or (tinf.ny>1) then SImg_New(cc, fk3, fk.nx,fk.ny,1)
        else fk3:=fk;
        fk2.pt:=nil; fw2.pt:=nil; fj2.pt:=nil;
        for y:=0 to tinf.ny-1 do
            for x:=0 to tinf.nx-1 do
            begin
                cc.info(cc.name+': PSF correction for tile '+int2str(y*tinf.nx+x+1)+' / '+int2str(tinf.nx*tinf.ny));
                SImg_GetTile(cc, tinf, fk,fk2, x,y);
                SImg_GetTile(cc, tinf, fw,fw2, x,y);
                SImg_GetTile(cc, tinf, fj,fj2, x,y);
{                if (x=0) and (y=0) then
                begin
                    SImg_WriteImageFile(cc,'starSubtract-corr-fk2.fits',fk2);
                    SImg_WriteImageFile(cc,'starSubtract-corr-fw2.fits',fw2); }
                SImg_StarSubtract_PSFCorrection(cc, fk2,fw2,fj2, cradc,p);
{                   SImg_WriteImageFile(cc,'starSubtract-corr-fk3.fits',fk2);
                end; }
                if fj.pt<>fj2.pt then SImg_Free(cc,fj2);
                if fw.pt<>fw2.pt then SImg_Free(cc,fw2);
                SImg_BlendTile(cc, tinf, fk2,fk3, x,y);
            end;
        if fk2.pt<>fk.pt then SImg_Free(cc, fk2);
        SImg_Move(cc, fk3,fk);
//    cc.info(SImg_InfoStr2(fi)+'   '+SImg_InfoStr2(sh.fi2)+'   '+SImg_InfoStr2(fm));

        { new residual }
        for i:=0 to fk.size-1 do
            fk.pt[i]+=fm.pt[i]; { residual }
        { new denoised residual }
        if ithrr<1e-6 then SImg_New(cc, fn, fi.nx,fi.ny,1, true)
        else begin
            SImg_New(cc, fn, fi.nx,fi.ny,1, false);
            a:=ithrr/ithrn;
            b:=(sh.offs+2*sh.sig)*a;
            for i:=0 to fn.size-1 do
                fn.pt[i]:=fk.pt[i]*a-b;
            SImg_FilterW(cc, fn,sh.fw, fn, sh.irad*2);
//          SImg_WriteImageFile(cc,'starSubtract-ress2.fits',fn);
//          SImg_WriteImageFile(cc,'starSubtract-fw.fits',sh.fw);
        end;
    end;
    cc.info(cc.name+': Calculating extended interpolation masks.');
    SImg_Free(cc,fw);
    SImg_Free(cc,sh.fw);
    if not ofos then SImg_Free(cc,fj);

// define[starSubtract_noIpol]
//ifndef[starSubtract_noIpol]
//    cc.info(SImg_InfoStr2(fi)+'   '+SImg_InfoStr2(sh.fi2)+'   '+SImg_InfoStr2(fm));
    { constrainted residual and mask for extended interpolation residual }
    for i:=0 to fk.size-1 do
    begin
        { constrainted residual }
        if rm.ptb[i] and 32<>0 then s:=sh.sat
        else begin
            r:=sh.fi2.pt[i];
            s:=fk.pt[i];
            if (r<sh.offs) or ((s>r) and (r<sh.sat)) then s:=r
            else if (s<sh.offs) then s:=sh.offs;
        end;
        fk.pt[i]:=s;
        { mask for extended interpolation residual }
        mb:=rm.ptb[i];
        if mb and 17=16 then    { limit interpolation size to size of correction interpolation }
        begin
            r:=sh.fi2.pt[i];
            if r<=sh.sat then
            begin
                r-=fk.pt[i];                            { star brightness }
                q:=fl.pt[i];                            { stars convoluted with error PSF }
                if q>r then r:=q;
                p:=fn.pt[i]*ithrn;
                if p<0 then p:=0;
                if r<p+sh.sig then rm.ptb[i]:=mb and (255-8);   { forbid interpolation if no isolated area }
            end;
        end;
        fl.pt[i]:=s;
    end;
//    SImg_WriteImageFile(cc,'starSubtract-imask2.fits',rm);

    { noise filtered extended interpolaton residual }
    SImg_InterpolateMasked(cc, fl, rm, 1+16, SImg_StarSubtract_ipol_fd);        { fl will be used for lower constraint }
    SImg_New(cc, fw, fi.nx,fi.ny,1, false);
    for i:=0 to fi.size-1 do
        fw.pt[i]:=fk.pt[i]-fl.pt[i];
    SImg_Filter(cc, fw,fw,inf.sig);

    { final interpolation, upper and lower constraint for result with background }
    a:=sh.sig/sqrt(1+inf.sig);  { takes noise reduction into account }
    b:=ithrr/ithrn;
    for i:=0 to fk.size-1 do
    begin
        mb:=rm.ptb[i];
        o:=sh.fi2.pt[i];
        if (o>=sh.sat) and (mb and 1=0) then
        begin
            r:=fi.pt[i];
            fk.pt[i]:=r;
            fm.pt[i]:=r;
            continue;
        end;
        { extended interpolation }
        r:=fm.pt[i];                    { mask or 8 interpolation, also used as lower residual if extended interpolation was disabled }
        if r<sh.offs then r:=sh.offs;
        s:=(r-sh.offs)*b;               { s >= 0 }
        if mb and 17=16 then
        begin
            p:=fn.pt[i];
            if s>p then p:=s;
            if abs(fw.pt[i])>p*ithrn+a then
            begin
                mb:=mb or 1;
                rm.ptb[i]:=mb;
            end;
        end;
        { residual, constraints and uninterpolated output }
        s:=fk.pt[i];                    { residual }
        if mb and (8+16)<>16 then
        begin
            if mb and (8+16)=(8+16) then        { two interpolation results which are not equal to fk.pt[i]: use the smaller one as lower constraint }
            begin
                q:=fl.pt[i];
                if q<r then r:=q;
                if s<r then s:=r;
            end;
            if s<sh.offs then s:=sh.offs;
            if s>o then s:=o;                   { r <= s <= o }
        end
        else if s<r then rm.ptb[i]:=mb or 9;    { interpolation if extended interpolation was disabled }
        q:=fi.pt[i]-o;                          { background = lower constraint }
        fk.pt[i]:=q+s;
        fm.pt[i]:=q;
    end;
//    SImg_WriteImageFile(cc,'starSubtract-imask3.fits',rm);
    SImg_FillMasked(cc, rm, 8,1);
    it:=0;
    for i:=0 to fk.size-1 do
    begin
        mb:=rm.ptb[i];
        if (mb and 9)=1 then rm.ptb[i]:=mb and 254
        else if mb and 1<>0 then it+=1;
    end;
    SImg_InterpolateMasked(cc, fk, rm,1, SImg_StarSubtract_ipol_fd);
//    SImg_WriteImageFile(cc,'starSubtract-imask4.fits',rm);
//    SImg_WriteImageFile(cc,'starSubtract-out.fits',fk);
    SImg_Free(cc, fl);
    SImg_Free(cc, fw);
    SImg_Free(cc, fn);

    { make sure that fk is smaller than fi }
    j:=0;
    for i:=0 to fk.size-1 do
    begin
        r:=fi.pt[i];
        if (fk.pt[i]>r) and (sh.fi2.pt[i]<sh.sat) then
        begin
            fk.pt[i]:=r;
            mb:=rm.ptb[i];
            if mb and 1<>0 then j+=1;
            rm.ptb[i]:=mb and 254;
        end;
    end;
    cc.info(cc.name+': Interpolation of '+real2str((it-j)/fk.size*100,-1)+'% of all pixels');
    if j>0 then SImg_InterpolateMasked(cc, fk, rm,1, SImg_StarSubtract_ipol_fd);

    j:=0;
    s:=random;
    a:=sqrt(3)*sh.sig;
    for i:=0 to fk.size-1 do
    begin
        mb:=rm.ptb[i];
        q:=fk.pt[i];
        r:=fi.pt[i];
//ifdef[debug_starSubtract]
        rm.ptb[i]:=(mb and 63) or (mb shl 6);
//else
//        rm.ptb[i]:=mb and 7;
        rm.ptb[i]:=mb and 31;
//endif
        if q>r then
        begin
            if sh.fi2.pt[i]<sh.sat then fk.pt[i]:=r;
            continue;
        end;
        if mb and 1=0 then continue;  { neither interpolated nor lower constraint was violated }
        r-=q;
        p:=q-fm.pt[i];
        if p<=0 then continue;
        if p<r then r:=p;
        if r>a then r:=a;
        fk.pt[i]:=q+(2*s-1)*r;
        j+=1;
        if j and 1023<>0 then s:=frac(pi*s+0.876)       { cheap random }
        else s:=random;
    end;
//else   // starSubtract_noIpol
    SImg_Free(cc, fl);
//endif  // starSubtract_noIpol
    SImg_Free(cc, sh.fi2);
    SImg_Free(cc, fpsf);

//    cc.info(SImg_InfoStr2(fi));
    { output difference and extrapolate saturated areas }
    if ofod then
    begin
        { saturation limit for fi }
        a:=fi.pt[0];
        for i:=1 to fi.size-1 do
        begin
            r:=fi.pt[i];
            if r>a then a:=r;
        end;
        a:=1.5/a;
        for i:=0 to fi.size-1 do
        begin
            s:=fi.pt[i];
            r:=s-fk.pt[i];
            s:=s*a-0.2;
            if s>0 then
            begin
                q:=fp.pt[i];
                if s<1 then s:=sqr(s)
                else s:=1;
                r:=r*(1-s)+q*s;
            end;
            fm.pt[i]:=r;
        end;
        SImg_Move(cc,fm,fod);
        SImg_Free(cc,fp);
    end
    else SImg_Free(cc,fm);

    if ofob then SImg_Move(cc,fk,fob)
    else SImg_Free(cc,fk);

    if ofos then SImg_Move(cc,fj,fos)
    else SImg_Free(cc,fj);

    SImg_DecDebugLevel;
end;
